#ifdef COPYRIGHT_INFORMATION
#include "gplv3.h"
#endif
/*
 * Copyright (C) 2002-2012 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
static int sdbh_updateBranch (DBHashTable *, FILE_POINTER);
static unsigned char sdbh_readBranches (DBHashTable *, FILE_POINTER);

static FILE_POINTER *sdbh_locate (DBHashTable *, FILE_POINTER *);
static FILE_POINTER *sdbh_locateFind (DBHashTable *, FILE_POINTER *, int);
static void sdbh_cuenta (unsigned char *, unsigned char, FILE_POINTER);
static FILE_POINTER sdbh_write (char, DBHashTable *, unsigned char);
static int sdbh_read (char, DBHashTable *, unsigned char);
static int sdbh_readheader (DBHashTable *);

#ifdef TURN
static FILE_POINTER sdbh_turnaround (FILE_POINTER);
#endif
static FILE_POINTER sdbh_z (unsigned char, unsigned char);
static unsigned char sdbh_cuantumR (FILE_POINTER, unsigned char, FILE_POINTER *);
static unsigned char sdbh_cuantum (FILE_POINTER, unsigned char, FILE_POINTER *);
static void sdbh_cifra (unsigned char *, unsigned char, unsigned char, FILE_POINTER *);

static int sdbh_barrelong (DBHashTable *, FILE_POINTER, int);
static int sdbh_reversebarrelong (DBHashTable *, FILE_POINTER, int);
static int sdbh_reversebarre (DBHashTable *, FILE_POINTER, int);
static int sdbh_barre (DBHashTable *, FILE_POINTER, int);

static void sdbh_transfer (DBHashTable *);
static DBHashTable *sdbh_open_S (const char *, int);

static int sdbh_size (DBHashTable *, FILE_POINTER);

static char *lock_name(const char *archivo);

/*************************************************************************/

#if 0
// Faster without this.
static int
sync_fd (int fd) {
    /*if (fsync(fd) < 0) {
       // this should not happen, really.
       fprintf(stderr,"fsync(): %s\n", strerror(errno));
       return 0;
       } */
    return 1;
}
#endif

static int
advance_fp (DBHashTable * dbh, FILE_POINTER offset) {
    //if (!sync_fd(dbh->fd)) return 0;
    if(lseek (dbh->fd, (off_t) offset, SEEK_CUR) < 0) {
        fprintf(stderr, "Error: advance_fp failed to advance fp %lld\n", (long long)offset);
        return 0;
    }
    return 1;
}

static int
place_fp_at (DBHashTable * dbh, FILE_POINTER seek) {
    //if (!sync_fd(dbh->fd)) return 0;
    if(lseek (dbh->fd, (off_t) seek, SEEK_SET) != seek) {
        // this should not happen...
        fprintf(stderr, "Error: sdbh_locate() cannot place file pointer at bof=0x%llu\n", (long long unsigned)seek);
        return 0;
    }
    return 1;
}

static FILE_POINTER
place_eof (DBHashTable * dbh) {
    //if (!sync_fd(dbh->fd)) return 0;
    FILE_POINTER eof;
    eof = (FILE_POINTER) lseek (dbh->fd, (off_t) 0L, SEEK_END);
    if(eof < sizeof (dbh_header_t)) {
        fprintf(stderr, "Error: tell_eof() <  sizeof (dbh_header_t)\n");
        eof = -1;
    }
    if(eof < 0) {
        // this should not happen...
        fprintf(stderr, "Error: tell_eof() cannot place file pointer at eof\n");
    }
    return eof;
}

  /* Funciones nivel 0: */
static void
sdbh_operate (DBHashTable * dbh) {
    if(dbh == NULL) {
        TRACE ("Must define function operate(place*) before doing a sweep\n");
    }
}

static char *
sdbh_randomfilename (DBHashTable * dbh, char code) {
    time_t segundos;
    FILE_POINTER divisor;
    char *archivo;
    const char *tmpdir;
    if(!dbh)
        return NULL;
    // determine best tmp directory
    if(dbh->head_info->user_tmpdir == 0) {
        struct stat dbh_st,
          tmp_st;
        stat (dbh->path, &dbh_st);
        stat ("/tmp", &tmp_st);
	if (dbh->tmpdir) free(dbh->tmpdir);dbh->tmpdir = NULL;
        if(dbh_st.st_dev == tmp_st.st_dev) {
            dbh->tmpdir = (char *)malloc (strlen ("/tmp") + 1);
	    if (dbh->tmpdir == NULL) {
		 fprintf(stderr, "malloc: %s\n", strerror(errno));
		 exit(1);
	    }   
	    strcpy (dbh->tmpdir, "/tmp");
        } else {
            dbh->tmpdir = (char *)malloc (strlen (dbh->path) + 1);
	    if (dbh->tmpdir == NULL) {
		 fprintf(stderr, "malloc: %s\n", strerror(errno));
		 exit(1);
	    }   
	    strcpy (dbh->tmpdir, dbh->path);
            if(strrchr (dbh->tmpdir, '/'))
                *strrchr (dbh->tmpdir, '/') = 0;
            else {
                fprintf(stderr, "%s is not an absolute path!", dbh->path);
		//g_error("sdbh_randomfilename\n");
            }
        }
    }
    tmpdir = dbh->tmpdir;
    archivo = (char *)malloc (strlen (tmpdir) + 1 + 1 + 6 + 4 + 1);
    if (archivo == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
		 exit(1);
    }  
    time (&segundos);
    int try=0;
retry:
    srand ((unsigned)segundos);
    divisor = RAND_MAX / 10000;
    if((segundos = rand () / divisor) > 100000L)
        segundos = 50001;
    sprintf (archivo, "%s/%c%ld.tmp", tmpdir, code, (long)segundos);
    // if the file exists, retry with a different random number...
    struct stat st;
    if (stat(archivo, &st)==0){
	if (segundos > 0) segundos--;
	else segundos++;
	if (try++ < 3) goto retry;
	free(archivo);
	return NULL;
    }

    return archivo;
}  /*******************************************************************************/

#if 0
this is unreliable in Linux...
//mode=1 --> write lock
static void
lock_fd (int fd, int mode, const char *archivo) {
    //  return;
    // get the file lock, wait if necesary:
    struct flock lock;

    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // to eof  

    if(mode) {
        lock.l_type = F_WRLCK;
    } else {
        lock.l_type = F_RDLCK;
    }

    int try=0;
retry:
    if(fcntl (fd, F_SETLKW, &lock) < 0) {
	if (try++ < 3) {
	    sleep(1);
	    goto retry;
	}
        fprintf(stderr, "fcntl(F_SETLKW)(%s): %s\n", archivo, strerror (errno));
    }
}

static void
unlock (int fd) {


    // release the lock:
    struct flock lock;

    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;             // to eof  

    lock.l_type = F_UNLCK;
    int try=0;
retry:
    if(fcntl (fd, F_SETLKW, &lock) < 0) {
	if (try++ < 3) {
	    sleep(1);
	    goto retry;
	}
        fprintf(stderr, "fcntl(F_SETLKW): %s\n", strerror (errno));
    }
}
#endif


static int
sdbh_lock_write (DBHashTable * dbh) {
    if(!dbh) return 0;
#ifdef PARALLEL_SAFE
    if (!(dbh->protection_flags & DBH_PARALLEL_SAFE)) return 1;
    if (dbh->lock_p == NULL) return 1;

    NOOP("DBH: locking file for write %s (%d)\n", dbh->path, dbh->fd);
    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
	    dbh->path, strerror (errno));
    }
    struct timespec thread_wait = {
        0, 100000000
    };
    int count=0;
    while (dbh->lock_p->read_lock_count ||
	    (dbh->lock_p->write_lock_count && dbh->lock_p->write_lock != getpid())){
	nanosleep (&thread_wait, NULL);
	if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	    fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
		dbh->path, strerror (errno));
	}
	if (++count % 200 == 0){
	    fprintf(stderr, "Process 0x%x: %d attempts to gain writelock for %s:\n  readlocks=%d writelocks=%d writepid=0x%x\nShared memory lock (may be stale) = %s",
		getpid(),
		count, dbh->path, dbh->lock_p->read_lock_count, 
		dbh->lock_p->write_lock_count, dbh->lock_p->write_lock,
		lock_name(dbh->path));
	}
	if (dbh->lock_attempt_limit && count > dbh->lock_attempt_limit) {
	    fprintf(stderr, "Process 0x%x: removing stale lock after %d attempts to gain access.\n", getpid(), count);
	    // Simply overwrite the read lock and write lock count with zero.
	    dbh->lock_p->read_lock_count = 0;
	    dbh->lock_p->write_lock_count = 0;
	    dbh->lock_p->write_lock = 0;
	    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
		fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
		    dbh->path, strerror (errno));
	    }
	}
#ifdef USE_GLIB 
	    //g_error("stop\n");
#endif
    }
    if (count >= 50){
	    fprintf(stderr, "write lock attained at attempt %d for %s\n",
		count, dbh->path);
    }
	
    dbh->lock_p->write_lock_count++;
    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	    fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
		dbh->path, strerror (errno));
    }
    dbh->lock_p->write_lock = getpid();
    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	    fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
		dbh->path, strerror (errno));
    }
#endif
    //lock_fd (dbh->fd, 1, dbh->path);
	// Flush to disk.
    //	fdatasync(dbh->fd); 
    return 1;
}


static int
sdbh_unlock_write (DBHashTable * dbh) {
    if(!dbh) return 0;
#ifdef PARALLEL_SAFE
    if (!(dbh->protection_flags & DBH_PARALLEL_SAFE)) return 1;

    if (dbh->lock_p == NULL) return 1;

    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
	    dbh->path, strerror (errno));
    }
    if (dbh->lock_p->write_lock_count==0 || dbh->lock_p->write_lock != getpid()){
	fprintf(stderr,"Invalid attempt to unlock write lock for %s: %s\n", 
	    dbh->path, strerror (errno));
	return 1;
    }
    dbh->lock_p->write_lock_count--;
    if (dbh->lock_p->write_lock_count==0){
	dbh->lock_p->write_lock = 0;
    }
    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	    fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
		dbh->path, strerror (errno));
    }
#endif

    return 1;
}

static int
sdbh_lock_read (DBHashTable * dbh) {
    if(!dbh) return 0;
#ifdef PARALLEL_SAFE
    if (!(dbh->protection_flags & DBH_PARALLEL_SAFE)) return 1;

    if (dbh->lock_p == NULL) return 1;

    NOOP("DBH: locking file for write %s (%d)\n", dbh->path, dbh->fd);
    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
	    dbh->path, strerror (errno));
    }
    struct timespec thread_wait = {
        0, 100000000
    };
    // Does the process already have a write lock?
    if (dbh->lock_p->write_lock_count && dbh->lock_p->write_lock == getpid()){
#ifdef USE_GLIB 
	    g_error("stop\n");
#endif
	fprintf(stderr, "this is wrong, process 0x%x already has a write lock on issuing read lock request\n", getpid());
	return 1;
    }
    int count=0;
    while (dbh->lock_p->write_lock_count){
	nanosleep (&thread_wait, NULL);
	if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	    fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
		dbh->path, strerror (errno));
	}
	if (++count % 50 == 0){
	    NOOP("0x%x: %d attempts to gain readlock for %s:\n  readlocks=%d writelocks=%d writepid=0x%x\n",
		getpid(),
		count, dbh->path, dbh->lock_p->read_lock_count, 
		dbh->lock_p->write_lock_count, dbh->lock_p->write_lock);
#ifdef USE_GLIB 
	    //g_error("stop\n");
#endif
	}
    }
    dbh->lock_p->read_lock_count++;
    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	    fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
		dbh->path, strerror (errno));
    }
#endif
    return 1;
}


static int
sdbh_unlock_read (DBHashTable * dbh) {
    if(!dbh) return 0;
#ifdef PARALLEL_SAFE
    if (!(dbh->protection_flags & DBH_PARALLEL_SAFE)) return 1;

    if (dbh->lock_p == NULL) return 1;

    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
	    dbh->path, strerror (errno));
    }
    if (dbh->lock_p->read_lock_count==0){
	NOOP("Invalid attempt to unlock read lock for %s: %s\n", 
	    dbh->path, strerror (errno));
	return 1;
    }
    dbh->lock_p->read_lock_count--;
    if(msync (dbh->lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	    fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
		dbh->path, strerror (errno));
    }
#endif

    return 1;
}

  /* Functions: ************************************************************** */
static unsigned char
sdbh_readBranches (DBHashTable * dbh, FILE_POINTER seek) {
    FILE_POINTER *rama;
    unsigned char ramas;
    rama = dbh->newbranch;
    TRACE ("sdbh_readBranches\n");

    if(!place_fp_at (dbh, seek)){
        return ERROR_VALUE;
    }
    if(read (dbh->fd, &ramas, 1) != 1) {
        TRACE ("sdbh_readBranches error 3.1\n");
        return ERROR_VALUE;
    }
    TRACE ("sdbh_readBranches ramas=%lld\n", (long long)ramas);
    if(!advance_fp (dbh, 1L + sizeof (FILE_POINTER))) {
        return ERROR_VALUE;
    }

    if(read (dbh->fd, rama, sizeof (FILE_POINTER) * ramas) != sizeof (FILE_POINTER) * ramas) {
        TRACE ("sdbh_readBranches error 3.2\n");
        return ERROR_VALUE;
    }
    TRACE ("read:");
    int i;
    for(i = 0; i < ramas; i++)
        TRACE (" rama[%d]=0x%llu", i, (long long unsigned)(rama[i]));
    TRACE ("\n");
#ifdef TURN
    {
        int i;
        for(i = 0; i < ramas; i++)
            rama[i] = sdbh_turnaround (rama[i]);
    }
#endif
    return ramas;
} /***************************************************/

static int
sdbh_updateBranch (DBHashTable * dbh, FILE_POINTER seek) {
    TRACE ("sdbh_updateBranch\n");
    FILE_POINTER *rama;
    unsigned char ramas;
    rama = dbh->newbranch;
    if(!place_fp_at (dbh, seek)) {
        return ERROR_VALUE;
    }
    if(read (dbh->fd, &ramas, 1) != 1) {
        TRACE ("sdbh_updateBranch read error 4.1\n");
        return ERROR_VALUE;
    }

    if(!advance_fp (dbh, 1L + sizeof (FILE_POINTER))){
        return ERROR_VALUE;
    }

#ifdef TURN
    {
        int i;
        for(i = 0; i < ramas; i++)
            rama[i] = sdbh_turnaround (rama[i]);
    }
#endif
    TRACE ("write:");
    int i;
    for(i = 0; i < ramas; i++)
        TRACE (" rama[%d]=0x%llu", i, (long long unsigned)(rama[i]));
    TRACE ("\n");

    if(write (dbh->fd, rama, sizeof (FILE_POINTER) * ramas) != sizeof (FILE_POINTER) * ramas) {
        TRACE ("sdbh_updateBranch write error 4.2\n");
        return ERROR_VALUE;
    }
#ifdef TURN
    {
        int i;
        for(i = 0; i < ramas; i++)
            rama[i] = sdbh_turnaround (rama[i]);
    }
#endif
    return 1;
}  /*****************************************************/

#define BRANCHOFF (dbh->head_info->n_limit-dbh->newbranches)
#define CURRENTSEEK fp[0]
#define LASTSEEK fp[1]
#define CURR_BRANCH fp[2]
#define SWITCHED (dbh->head_info->reservedD)
//FILE_POINTER *fp must have size = 3
static FILE_POINTER *
sdbh_locate (DBHashTable * dbh, FILE_POINTER * fp) {
    TRACE ("sdbh_locate\n");
    FILE_POINTER currentseek,
      lastseek;
    int i;                      /*,offset; */
    lastseek = currentseek = 0;
    CURRENTSEEK = LASTSEEK = CURR_BRANCH = 0;   /* ERROR value assigned */
    currentseek = dbh->head_info->bof;

    if(!place_fp_at (dbh, currentseek)) {
        return NULL;
    }

    if(!sdbh_read (NEW, dbh, 1)) {
        return fp;
    }
  loop:
#define OFFSET (i+(dbh->head_info->n_limit-dbh->newbranches))
    for(i = 0; i < dbh->newbranches; i++) {
        if(*(dbh->key + OFFSET) != *(dbh->newkey + OFFSET))
        {
            lastseek = currentseek;
            if((currentseek = *(dbh->newbranch + i)) == 0) {
                CURR_BRANCH = i;
                break;
            }
            if(!place_fp_at (dbh, currentseek)) {
                return NULL;
            }
            if(!sdbh_read (NEW, dbh, 1)) {
                return fp;
            }
            goto loop;
        }
    }
    CURRENTSEEK = currentseek;
    LASTSEEK = lastseek;
    return fp;
}

static FILE_POINTER *
sdbh_locateFind (DBHashTable * dbh, FILE_POINTER * fp, int keys) {
    TRACE ("sdbh_locateFind\n");
    FILE_POINTER currentseek,
      lastseek;
    int i;
    lastseek = currentseek = 0;
    CURRENTSEEK = LASTSEEK = CURR_BRANCH = 0;   /* ERROR value assigned */
    currentseek = dbh->head_info->bof;

    if(!place_fp_at (dbh, currentseek)) {
        return NULL;
    }

    if(!sdbh_read (NEW, dbh, 1)) {
        return fp;
    }
    if(keys > dbh->head_info->n_limit)
        keys = dbh->head_info->n_limit;
  loop:
    for(i = 0; i < dbh->newbranches - (dbh->head_info->n_limit - keys); i++) {
        /*  offset=i+(dbh->head_info->n_limit-dbh->newbranches); */
        if(*(dbh->key + OFFSET) != *(dbh->newkey + OFFSET)) {
            lastseek = currentseek;
            if((currentseek = *(dbh->newbranch + i)) == 0) {
                CURR_BRANCH = i;
                break;
            }
            if(!place_fp_at (dbh, currentseek)) {
                return NULL;
            }
            if(!sdbh_read (NEW, dbh, 1)) {
                return fp;
            }
            goto loop;
        }
    }
    CURRENTSEEK = currentseek;
    LASTSEEK = lastseek;
    return fp;
}

static char *lock_name(const char *archivo){
    if (!archivo || strlen(archivo)==0){
 	fprintf(stderr,"Cannot determine lock name for %s\n", 
		(archivo)?archivo:"NULL");
#ifdef USE_GLIB 
	    //g_error("stop\n");
#endif
	return NULL;
   }

    char *path = realpath(archivo, NULL);
    if (!path) return NULL;
 
    // Get inode information:
    struct stat st;
    if (stat(path, &st) < 0){
	// This is a valid return value. 
	NOOP(stderr, "Cannot stat %s: %s\n", path, strerror(errno));
	free(path);
	return NULL;
    }

    // Build shm_name
    char buffer[1024];
    memset(buffer, 0, 1024);
    sprintf(buffer, "/%d-%d", (int)st.st_dev, (int)st.st_ino);

    char *shm_name = (char *)malloc((strlen(buffer)+1)*sizeof(char));
    if (!shm_name) {
 	fprintf(stderr,"Cannot malloc lock name for %s\n", path);
		 exit(1);
   }
    memset(shm_name, 0, strlen(buffer)+1);
    sprintf(shm_name, "%s", buffer);
    free(path);
    return shm_name;
}

static void
destroy_shm_lock(char *archivo, dbh_lock_t *lock_p){
#ifdef PARALLEL_SAFE
    if (!lock_p) return;
    char *shm_name=lock_name(archivo);
    if (!shm_name) {
	// Nothing to destroy.
	return;
    }
    munmap (lock_p, sizeof(dbh_lock_t));
    shm_unlink (shm_name);
    free(shm_name);
#endif
}

static void
clear_shm_lock(const char *archivo){
#ifndef PARALLEL_SAFE
    return;
#endif
    char *shm_name=lock_name(archivo);
    if (!shm_name){
	// Nothing to clear.
	return;
    }
    shm_unlink (shm_name);
    free(shm_name);
    return;
}

static dbh_lock_t *
create_shm_lock(const char *archivo){
#ifndef PARALLEL_SAFE
    return NULL;
#else
    char *shm_name=lock_name(archivo);
    if (!shm_name){
	// This would indicate something is wrong...
	fprintf(stderr,"Cannot get lock name for %s\n", 
		archivo);
	return NULL;
    }

    int fd = shm_open (shm_name, O_RDWR, S_IRUSR | S_IWUSR);
    if (fd > 0){
	NOOP("Lockfile exists %s -> %s\n", archivo, shm_name);
    } else {
	fd = shm_open (shm_name, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
    }
    if(fd < 0) {
	fprintf(stderr,"Cannot open shared memory file descriptor for %s (%s): %s\n", 
		shm_name, archivo, strerror (errno));
	free(shm_name);
	return NULL;
    } else {
	if(ftruncate (fd, sizeof(dbh_lock_t)) < 0) {
	    fprintf(stderr,"Cannot ftruncate shared memory item for %s: %s\n", 
		    archivo, strerror (errno));
	    free(shm_name);
	    close(fd);
	    return NULL;
	}
    }
    dbh_lock_t *lock_p = (dbh_lock_t *) mmap (NULL, sizeof(dbh_lock_t),
	PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (lock_p == MAP_FAILED){
	fprintf(stderr,"Cannot mmap shared memory item for %s: %s\n", 
	    archivo, strerror (errno));
	close (fd);
	shm_unlink (shm_name);
	free(shm_name);
	return NULL;
    } 
    if(msync (lock_p,  sizeof(dbh_lock_t), MS_SYNC) < 0){
	fprintf(stderr,"Cannot msync shared memory item for %s: %s\n", 
	    archivo, strerror (errno));	
        munmap (lock_p, sizeof(dbh_lock_t));
	close (fd);
	shm_unlink (shm_name);
	free(shm_name);
	return NULL;
    }
    free(shm_name);
    close (fd);
    return lock_p;
#endif
}

static DBHashTable * 
sdbh_create (const char *path, unsigned char key_length) {
    DBHashTable *dbh;
    dbh = (DBHashTable *) malloc (sizeof (DBHashTable));
    if(dbh == NULL){
	fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
    }
    memset (dbh, 0, sizeof (DBHashTable));

    /////////////////////////////////////////////
    
    unlink (path);
    dbh->fd = open (path, O_CREAT | O_RDWR | O_TRUNC, 0640);
    if(dbh->fd < 0) {
        free (dbh);
        fprintf(stderr, "sdbh_create(%s): %s\n", path, strerror (errno));
        return NULL;
    }

    dbh->head_info = (dbh_header_t *) malloc (sizeof (dbh_header_t));
     if (dbh->head_info == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }  
     
    memset ((void *)(dbh->head_info), 0, sizeof (dbh_header_t));

    dbh->head_info->bof = sizeof (dbh_header_t);
    dbh->head_info->n_limit = key_length;
    dbh->head_info->user_tmpdir = 0;
    dbh->head_info->sweep_erased = 0;
    dbh->head_info->total_space = dbh->head_info->data_space = dbh->head_info->erased_space = 0;
    strncpy (dbh->head_info->version, DBH_FILE_VERSION, 15);
    dbh->head_info->records = 0;
    dbh->operate = sdbh_operate;
    dbh->branch = (FILE_POINTER *) malloc (dbh->head_info->n_limit * sizeof (FILE_POINTER));
     if (dbh->branch == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }      
     dbh->newbranch = (FILE_POINTER *) malloc (dbh->head_info->n_limit * sizeof (FILE_POINTER));
      if (dbh->newbranch == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }        
     dbh->key = (unsigned char *)malloc (key_length);
     if (dbh->key == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }  
     dbh->newkey = (unsigned char *)malloc (key_length);
     if (dbh->newkey == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }          
     dbh->path = (char *)malloc (strlen (path) + 1);
     if (dbh->path == NULL || dbh->newbranch == NULL ||
	    dbh->key == NULL || dbh->newkey == NULL || dbh->path == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	 exit(1);
     }  

    strcpy (dbh->path, path);

    dbh->head_info->record_length = 0;
    dbh->head_info->writeOK = 1;
    sdbh_size (dbh, DEFAULT_DBH_DATASIZE);
    dbh->bytes_userdata = 0;

    dbh_writeheader (dbh);
    TRACE ("DBH: created %s \n", path);
    return dbh;
}

static DBHashTable *
sdbh_open_S (const char *archivo, int mode) {
    TRACE (stderr, "sdbh_open_S: %s\n", archivo);
    DBHashTable *dbh;
    int fd;
    if(mode) {
        fd = open (archivo, O_RDWR);
    } else {
        fd = open (archivo, O_RDONLY);
    }



    if(fd < 0) {
        TRACE ("sdbh_open(%s): %s\n", archivo, strerror (errno));
        return NULL;
    }
    dbh = (DBHashTable *) malloc (sizeof (DBHashTable));
     if (dbh == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }          
    memset (dbh, 0, sizeof (DBHashTable));

    /////////////////////////////////////////////

    dbh->head_info = (dbh_header_t *) malloc (sizeof (dbh_header_t));
     if (dbh->head_info == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }      
     dbh->fd = fd;
    FILE_POINTER eof = place_eof (dbh);
    if(eof < 0) {
        close (fd);
        free (dbh->head_info);
        free (dbh);
        return NULL;
    }

    dbh->path = (char *)malloc (strlen (archivo) + 1);
     if (dbh->path == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
     strcpy (dbh->path, archivo);

    if(!sdbh_readheader (dbh)) {
        close (fd);
        free (dbh->path);
        free (dbh->head_info);
        free (dbh);
        return NULL;
    }
    dbh->operate = sdbh_operate;
    dbh->branch = (FILE_POINTER *) malloc (dbh->head_info->n_limit * sizeof (FILE_POINTER));
     if (dbh->branch  == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
    dbh->newbranch = (FILE_POINTER *) malloc (dbh->head_info->n_limit * sizeof (FILE_POINTER));
     if (dbh->newbranch  == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
    dbh->key = (unsigned char *)malloc (dbh->head_info->n_limit);
      if (dbh->key == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
   dbh->newkey = (unsigned char *)malloc (dbh->head_info->n_limit);
     if (dbh->newkey  == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
    if(mode == WRITE)
        dbh->head_info->writeOK = 1;
    else
        dbh->head_info->writeOK = 0;
    sdbh_size (dbh, (int)dbh->head_info->record_length);
    TRACE ("DBH: opened %s \n", archivo);
    return dbh;
}  /****************************************************************************/

static void
sdbh_cuenta (unsigned char *numero, unsigned char orden, FILE_POINTER cuanto) {
    unsigned char q;
    FILE_POINTER m;
    if(orden == 0)
        numero[0] = 0;
    if(orden == 1)
        numero[0] = (unsigned char)cuanto;
    q = sdbh_cuantumR (cuanto, orden, &m);       /* en que cuantum estamos */
    orden--;
    sdbh_cifra (numero, q, orden, &m);
/* sdbh_cifra(numero,q,orden-1,&m);*/
} /***********************************************************************/

#define RECORD_EXTENT  ( 2 + sizeof(FILE_POINTER)*(*how_many_branches+1) + dbh->head_info->n_limit + the_user_databytes )
static FILE_POINTER
sdbh_write (char newinfo, DBHashTable * dbh, unsigned char write_branches) {
    TRACE ("sdbh_write\n");
    int i;
    void *the_data;
    unsigned char *how_many_branches;
    FILE_POINTER the_user_databytes;
    FILE_POINTER *the_branches;
    unsigned char *the_key;
    if(newinfo) {
        the_branches = dbh->newbranch;
    } else {
        the_branches = dbh->branch;
    }
    the_data = dbh->data;
    how_many_branches = &(dbh->newbranches);
    the_user_databytes = dbh->bytes_userdata;
    the_key = dbh->key;

    if(the_user_databytes > dbh->head_info->record_length) {
        fprintf(stderr,
                 "dbh->bytes_userdata (%lld) is greater than dbh->head_info->record_length (%lld). This is wrong and I stubbornly refuse to write\n",
                 (long long)dbh->bytes_userdata, (long long)dbh->head_info->record_length);
        return ERROR_VALUE;
    }
    if(write (dbh->fd, how_many_branches, 1) != 1){
        return ERROR_VALUE;
    }
    if(write (dbh->fd, &(dbh->flag), 1) != 1){
        return ERROR_VALUE;
    }

    i = write (dbh->fd, &the_user_databytes, sizeof (FILE_POINTER));
    if(i != sizeof (FILE_POINTER)){
        return ERROR_VALUE;
    }

    if(write_branches) {
#ifdef TURN
        for(i = 0; i < *how_many_branches; i++)
            the_branches[i] = sdbh_turnaround (the_branches[i]);
#endif
        i = write (dbh->fd, the_branches, sizeof (FILE_POINTER) * (*how_many_branches));
        if(i != sizeof (FILE_POINTER) * (*how_many_branches)){
        return ERROR_VALUE;
    }

#ifdef TURN
        for(i = 0; i < *how_many_branches; i++)
            the_branches[i] = sdbh_turnaround (the_branches[i]);
#endif
    } else {
        if(!advance_fp (dbh, *how_many_branches * sizeof (FILE_POINTER))){
        return ERROR_VALUE;
    }


    }
    if(write (dbh->fd, the_key, dbh->head_info->n_limit) != dbh->head_info->n_limit) {
        fprintf(stderr, "fwrite: %s\n", strerror (errno));
        return ERROR_VALUE;
    }
    if(the_user_databytes) {
        if(dbh->head_info->reservedC){
            the_user_databytes = write (dbh->fd, the_data, dbh->head_info->record_length);
	} else {
            the_user_databytes = write (dbh->fd, the_data, the_user_databytes);
	}
        return (the_user_databytes);
    } 
    return (ERROR_VALUE);
    
}  /************************************************************************************/

static int
sdbh_read (char newinfo, DBHashTable * dbh, unsigned char read_data) {
    TRACE ("sdbh_read\n");
    unsigned char keylength;
    void *the_data;
    unsigned char *how_many_branches;
    FILE_POINTER *the_user_databytes;
    FILE_POINTER *the_branches;
    unsigned char *the_key;
    if(newinfo) {
        the_branches = dbh->newbranch;
        the_data = dbh->newdata;
        how_many_branches = &(dbh->newbranches);
        the_user_databytes = &(dbh->newbytes_userdata);
        the_key = dbh->newkey;
    } else {
        the_branches = dbh->branch;
        the_data = dbh->data;
        how_many_branches = &(dbh->branches);
        the_user_databytes = &(dbh->bytes_userdata);
        the_key = dbh->key;
    }
    keylength = dbh->head_info->n_limit;

    if(read (dbh->fd, how_many_branches, 1) != 1) {
        TRACE ("fread error 2.1 (on empty dbh file this is OK)\n");
        return ERROR_VALUE;
    }
    if(read (dbh->fd, &(dbh->flag), 1) != 1) {

        TRACE ("fread error 2.2\n");
        return ERROR_VALUE;
    }
    if(read (dbh->fd, the_user_databytes, sizeof (FILE_POINTER)) != sizeof (FILE_POINTER)) {
        TRACE ("fread error 2.3\n");

        return ERROR_VALUE;
    }

    if(*the_user_databytes > dbh->head_info->record_length) {
        fprintf(stderr,
                 "sdbh_read(): the_user_databytes (%ld) is greater than dbh->head_info->record_length (%lld). This is wrong and I stubbornly refuse to read\n",
                 (long)the_user_databytes, (long long)(dbh->head_info->record_length));
        return ERROR_VALUE;
    }
    if(*the_user_databytes == 0) {
        TRACE (stderr, "dbh_RECORD_SIZE() == 0. If this is not intentional, use dbh_set_recordsize() to set record size for %s.\n", dbh->path);
        return ERROR_VALUE;
    }
    if(read (dbh->fd, the_branches, sizeof (FILE_POINTER) * (*how_many_branches)) != sizeof (FILE_POINTER) * (*how_many_branches)) {
        TRACE ("fread error 2.4\n");

        return ERROR_VALUE;
    }
#ifdef TURN
    {
        int i;
        for(i = 0; i < *how_many_branches; i++)
            the_branches[i] = sdbh_turnaround (the_branches[i]);
    }
#endif
    if(read (dbh->fd, the_key, keylength) != keylength) {
        TRACE ("fread error 2.5\n");
        return ERROR_VALUE;
    }
    if(read_data) {
        if(read (dbh->fd, the_data, *the_user_databytes) != *the_user_databytes) {
            TRACE ("fread error 2.51: blocksize=%lld\n", (long long)(*the_user_databytes));

            return ERROR_VALUE;
        }
    }
    return 1;
} /************************************************************************************/

static int
sdbh_readheader (DBHashTable * dbh) {
    TRACE ("sdbh_readheader\n");
    if(!place_fp_at (dbh, 0L)) {
        return 0;
    }
    if(read (dbh->fd, dbh->head_info, sizeof (dbh_header_t)) != sizeof (dbh_header_t)) {
        fprintf(stderr, "Failed to read header for %s: %s\n", dbh->path, strerror (errno));
        return 0;
    }
    if(strncmp (dbh->head_info->version, DBH_FILE_VERSION, 15) != 0) {
        fprintf(stderr,
                 "Failed to read header for %s at sdbh_readheader(): strncmp (\"%s\",\"%s\")\n",
                 dbh->path, dbh->head_info->version, DBH_FILE_VERSION);
        return 0;
    }
    /* volatile values: */
    dbh->head_info->user_tmpdir = 0;

#ifdef TURN
    dbh->head_info->bof = sdbh_turnaround (dbh->head_info->bof);
    dbh->head_info->record_length = sdbh_turnaround (dbh->head_info->record_length);
    dbh->head_info->total_space = sdbh_turnaround (dbh->head_info->total_space);
    dbh->head_info->data_space = sdbh_turnaround (dbh->head_info->data_space);
    dbh->head_info->erased_space = sdbh_turnaround (dbh->head_info->erased_space);
    dbh->head_info->records = sdbh_turnaround (dbh->head_info->records);
#endif
    return 1;
}  /************************************************************************************/

#ifdef TURN
static FILE_POINTER
sdbh_turnaround (FILE_POINTER x) {
    char *where,
      temp;
    where = (char *)&x;
    temp = where[0];
    where[0] = where[3];
    where[3] = temp;
    temp = where[1];
    where[1] = where[2];
    where[2] = temp;
    return x;
} /************************************************************************************/
#endif

static FILE_POINTER
sdbh_z (unsigned char n, unsigned char m) {      /* n=cuantum m=orden */
    FILE_POINTER s = 0;
    unsigned char i;
    if(m < 2)
        return 1;
    if(m == 2)
        return (n + 1);
    m--;
    for(i = n; i > 0; i--)
        s += sdbh_z (i, m);
/* for (i=n;i>0;i--) s += sdbh_z(i,m-1);*/
    return (s + 1);
}  /************************************************************************************/

static unsigned char
sdbh_cuantumR (FILE_POINTER cuanto, unsigned char orden, FILE_POINTER * m) {
    unsigned char i;
    FILE_POINTER ztotal,
      grandtotal = 0;
    for(i = 0; 1; i++) {
        if(i == orden)
            ztotal = grandtotal;
        else
            ztotal = sdbh_z (i, orden);
        grandtotal += ztotal;
        if(cuanto <= grandtotal) {
            if(m)
                *m = grandtotal - cuanto + 1;
            break;
        }
    }
    return i;
}  /************************************************************************************/

static unsigned char
sdbh_cuantum (FILE_POINTER cuanto, unsigned char orden, FILE_POINTER * m) {
    unsigned char i;
    FILE_POINTER ztotal,
      grandtotal = 0;
    for(i = 0; 1; i++) {
        if(i == orden)
            ztotal = grandtotal;
        else
            ztotal = sdbh_z (i, orden);
        grandtotal += ztotal;
        if(cuanto <= grandtotal)
            break;
        if(m)
            *m = cuanto - grandtotal;
    }
    return i;
}  /************************************************************************************/

static void
sdbh_cifra (unsigned char *numero, unsigned char q, unsigned char orden, FILE_POINTER * m) {
    unsigned char subcuantum;
    int t;
    if(orden == 1) {
        t = q - (*m - 1);
        numero[0] = (unsigned char)t;
/*  numero[0]=q - (*m-1);*/
        t = *m - 1;
        numero[1] = (unsigned char)t;
/*  numero[1]= *m-1;*/
        return;
    }
    subcuantum = sdbh_cuantum (*m, orden, m);
    t = q - subcuantum;
    numero[0] = (unsigned char)t;
/* numero[0]=q-subcuantum;*/
    orden--;
    sdbh_cifra (numero + 1, subcuantum, orden, m);
    /* sdbh_cifra(numero+1,subcuantum,orden-1,m); */
    return;
}  /************************************************************************************/

static int
sdbh_barrelong (DBHashTable * dbh, FILE_POINTER startadd, int ramas) {
    unsigned char i;
    if(dbh_load_address (dbh, startadd) == ERROR_VALUE)
        return ERROR_VALUE;
    if(dbh->head_info->dbh_exit)
        return 2;
    if(!ERASED || (ERASED && dbh->head_info->sweep_erased))
        (*(dbh->operate)) (dbh);
    if(ramas) {
        if((ramas = dbh->branches - ramas) < 0)
            ramas = 0;
    }
    for(i = dbh->branches; i > ramas; i--) {
        if(dbh->branch[i - 1])
            sdbh_barrelong (dbh, dbh->branch[i - 1], 0);
        dbh_load_address (dbh, startadd);
    }
    return 1;
}    /**********************************************************************/

static int
sdbh_reversebarrelong (DBHashTable * dbh, FILE_POINTER startadd, int ramas) {
    int i;
    if(dbh_load_address (dbh, startadd) == ERROR_VALUE)
        return ERROR_VALUE;
    if(dbh->head_info->dbh_exit)
        return 2;
    if(!ERASED || (ERASED && dbh->head_info->sweep_erased))
        (*(dbh->operate)) (dbh);
/* if (!ERASED) (*(dbh->operate))(dbh);*/
    if(ramas) {
        if((ramas = dbh->branches - ramas) < 0)
            ramas = 0;
    }
    for(i = ramas; i < dbh->branches; i++) {
        if(dbh->branch[i])
            sdbh_reversebarrelong (dbh, dbh->branch[i], 0);
        dbh_load_address (dbh, startadd);
    }
    return 1;
}   /******************************************************************************/

static int
sdbh_reversebarre (DBHashTable * dbh, FILE_POINTER startadd, int ramas) {
    int i;
    unsigned char oldbranches;
    FILE_POINTER *oldbranch;
    char fullmemory = 0;
    if(dbh_load_address (dbh, startadd) == ERROR_VALUE)
        return ERROR_VALUE;
    if(dbh->head_info->dbh_exit)
        return 2;
    if(!ERASED || (ERASED && dbh->head_info->sweep_erased))
        (*(dbh->operate)) (dbh);
/* if (!ERASED) (*(dbh->operate))(dbh);*/
    oldbranch = (FILE_POINTER *) malloc (dbh->branches * sizeof (FILE_POINTER));
     if (oldbranch == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
    if(oldbranch == NULL)
        fullmemory = 1;
    oldbranches = dbh->branches;
    if(ramas) {
        if((ramas = dbh->branches - ramas) < 0)
            ramas = 0;
    }
    if(!fullmemory) {
        for(i = 0; i < oldbranches; i++)
            oldbranch[i] = dbh->branch[i];
        for(i = ramas; i < oldbranches; i++)
            if(oldbranch[i])
                sdbh_reversebarre (dbh, oldbranch[i], 0);
        free (oldbranch);
    } else
        for(i = ramas; i < oldbranches; i++) {
            if(dbh->branch[i])
                sdbh_reversebarrelong (dbh, dbh->branch[i], 0);
            dbh_load_address (dbh, startadd);
        }
    return 1;
}  /******************************************************************************/

static int
sdbh_barre (DBHashTable * dbh, FILE_POINTER startadd, int ramas) {
    unsigned char i;
    FILE_POINTER *oldbranch;
    char fullmemory = 0;
    if(dbh_load_address (dbh, startadd) == ERROR_VALUE)
        return ERROR_VALUE;
    if(dbh->head_info->dbh_exit)
        return 2;
#if 0
    printf ("barre: currentseek at %lu\n", (long unsigned)startadd);
    printf ("record is erased? -> %d\n", ERASED);
#endif
    if(!ERASED || (ERASED && dbh->head_info->sweep_erased))
        (*(dbh->operate)) (dbh);
/* if (!ERASED) (*(dbh->operate))(dbh);*/
    oldbranch = (FILE_POINTER *) malloc (dbh->branches * sizeof (FILE_POINTER));
      if (oldbranch == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
    if(ramas) {
        if((ramas = dbh->branches - ramas) < 0)
            ramas = 0;
    }
    if(!fullmemory) {
        for(i = 0; i < dbh->branches; i++)
            oldbranch[i] = dbh->branch[i];
        for(i = dbh->branches; i > ramas; i--)
            if(oldbranch[i - 1])
                sdbh_barre (dbh, oldbranch[i - 1], 0);
        free (oldbranch);
    } else
        for(i = dbh->branches; i > ramas; i--) {
            if(dbh->branch[i - 1])
                sdbh_barrelong (dbh, dbh->branch[i - 1], 0);
            dbh_load_address (dbh, startadd);
        }
    return 1;
}   /************************************************************************/

  /* Functions: */
static void
sdbh_transfer (DBHashTable * srcdbh) {
    DBHashTable *dbh_desdbh = srcdbh->dbh_desdbh;
    dbh_desdbh->bytes_userdata = srcdbh->bytes_userdata;
    dbh_update (dbh_desdbh);
/* if ((count++ % 100)==0) printf("\n %ld registros procesados",count);*/
    return;
} /***************************************************************************/

/*****************************************************************************/
/***    old   non static functions ***/
/*****************************************************************************/

static int
sdbh_size (DBHashTable * dbh, FILE_POINTER record_length) {

    /* despues de abrir DBHashTable */
    if(dbh == NULL) {
        fprintf(stderr, "sdbh_size(): dbh != NULL not met\n");
        return 0;
    }
    if(dbh->head_info == NULL) {
        fprintf(stderr, "sdbh_size(): dbh->head_info != NULL not met\n");
        return 0;
    }

    if(record_length < dbh->head_info->record_length) {
        TRACE ("sdbh_size(): may not shrink! current size=%lld requested size=%ld\n",
               (long long)dbh->head_info->record_length, (long)record_length);
        return 0;
    }

    void *data = malloc (record_length);
     if (data == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
    void *newdata = malloc (record_length);
     if (newdata == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
    memset (data, 0, record_length);
    memset (newdata, 0, record_length);
    if(dbh->data != NULL) {
        memcpy (data, dbh->data, dbh->head_info->record_length);
        free (dbh->data);
    }
    if(dbh->newdata != NULL) {
        memcpy (newdata, dbh->newdata, dbh->head_info->record_length);
        free (dbh->newdata);
    }
    dbh->data = data;
    dbh->newdata = newdata;

    if (dbh->head_info->record_length != record_length) {
	dbh->head_info->record_length = record_length;
        if(dbh->head_info->writeOK) dbh_writeheader (dbh);
    }
    return 1;
} /****************************************************************************************/

static int
sdbh_newbarre (DBHashTable * dbh, unsigned char *key1, unsigned char *key2, unsigned char keylength) {
    int i;
    FILE_POINTER pointer;
    if(!dbh) {
        fprintf(stderr, "sdbh_newbarre(): %s\n", strerror (EBADF));

        return ERROR_VALUE;
    }

    dbh->head_info->dbh_exit = 0;
    if(key1 == NULL) {
        sdbh_barre (dbh, dbh->head_info->bof, 0);
        return 1;
    }
    if(key2 != NULL) {
        memcpy ((void *)(dbh->key), (void *)key1, dbh->head_info->n_limit);

        for(i = 0; i < dbh->head_info->n_limit; i++)
            if(key1[i] != key2[i])
                break;
        /* que pedo con ERASED?? (load returns errorvalue on ERASED */
        pointer = dbh_load (dbh);
        if((pointer == ERROR_VALUE) && !(ERASED))
            return ERROR_VALUE;
/*  if ((pointer=dbh_load(dbh)) == ERROR_VALUE) return ERROR_VALUE;*/
        /* aqui hay que considerar que el nodo puede estar marcado como ERASED */
        else {
            pointer = dbh_find (dbh, i);
            sdbh_barre (dbh, pointer, dbh->head_info->n_limit - i);
        }
    } else if(keylength) {
        memcpy ((void *)(dbh->key), (void *)key1, keylength);
        pointer = dbh_find (dbh, keylength);
        if(pointer == ERROR_VALUE)
            return ERROR_VALUE;
        sdbh_barre (dbh, pointer, dbh->head_info->n_limit - keylength);
    }
    return 1;
} /***************************************************************************************/

static int
sdbh_newreversebarre (DBHashTable * dbh, unsigned char *key1, unsigned char *key2, unsigned char keylength) {
    int i;
    FILE_POINTER pointer;
    if(!dbh) {
        fprintf(stderr, "sdbh_newreversebarre(): %s\n", strerror (EBADF));

        return ERROR_VALUE;
    }

    dbh->head_info->dbh_exit = 0;
    if(key1 == NULL) {
        sdbh_reversebarre (dbh, dbh->head_info->bof, 0);
        return 1;
    }
    if(key2 != NULL) {
        memcpy ((void *)(dbh->key), (void *)key1, dbh->head_info->n_limit);
        for(i = 0; i < dbh->head_info->n_limit; i++)
            if(key1[i] != key2[i])
                break;
        /* que pedo con ERASED?? (load returns errorvalue on ERASED */
        pointer = dbh_load (dbh);
        if((pointer == ERROR_VALUE) && !(ERASED))
            return ERROR_VALUE;
/*  if ((pointer=dbh_load(dbh)) == ERROR_VALUE) return ERROR_VALUE;*/
        else {
            pointer = dbh_find (dbh, i);
            sdbh_reversebarre (dbh, pointer, i);
        }
    } else if(keylength != 0) {
        memcpy ((void *)dbh->key, (void *)key1, keylength);

        pointer = dbh_find (dbh, keylength);
        if(pointer == ERROR_VALUE)
            return ERROR_VALUE;
        sdbh_reversebarre (dbh, pointer, keylength);
    }
    return 1;
}  /************************************************************************************/

static DBHashTable *
sdbh_regen (DBHashTable * dbh, int sweep) {
    void *temp, *newtemp;
    char *archivo2, *archivo;
    DBHashTable *newdbh;
    unsigned char *tempkey, *newtempkey;
    void (*fun) (DBHashTable *);

    if(!dbh) {
        fprintf(stderr, "sdbh_regen(): %s\n", strerror (EBADF));
        return NULL;
    }
    if (dbh->protection_flags & DBH_READ_ONLY){
        fprintf(stderr, "DBH table is read only: %s\n", dbh->path);
        return NULL;
    }
    sdbh_lock_write (dbh);

    archivo2 = sdbh_randomfilename (dbh, 's');
    if (!archivo2){
        fprintf(stderr, "Cannot open temporary file: DBHashtable will not be regenerated\n");
	return NULL;
    }

    DBHashTable *dbh_desdbh = sdbh_create (archivo2, dbh->head_info->n_limit);
    if(!dbh_desdbh) {
        fprintf(stderr, "Cannot open for write temporary file: %s\n", archivo2);
	free(archivo2);
        return NULL;
    }
    sdbh_lock_write (dbh_desdbh);

    archivo = (char *)malloc (strlen (dbh->path) + 1);
     if (archivo == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
    strcpy (archivo, dbh->path);
    dbh->dbh_desdbh = dbh_desdbh;

    tempkey = dbh_desdbh->key;
    newtempkey = dbh_desdbh->newkey;
    temp = dbh_desdbh->data;
    newtemp = dbh_desdbh->newdata;
    dbh_desdbh->key = dbh->key;
    dbh_desdbh->newkey = dbh->newkey;
    dbh_desdbh->data = dbh->data;
    dbh_desdbh->newdata = dbh->newdata;

    

    dbh_desdbh->head_info->record_length = dbh->head_info->record_length;
    fun = dbh->operate;

    dbh->operate = sdbh_transfer;
    if(sweep) {
        sdbh_newbarre (dbh, NULL, NULL, 0);
    } else {
        sdbh_newreversebarre (dbh, NULL, NULL, 0);
    }

    dbh_desdbh->data = temp;
    dbh_desdbh->newdata = newtemp;
    dbh_desdbh->key = tempkey;
    dbh_desdbh->newkey = newtempkey;
    // Set thread safe characteristics to new table.
    int flags = dbh->protection_flags;

    sdbh_unlock_write (dbh_desdbh);
    dbh_close (dbh_desdbh);
    sdbh_unlock_write (dbh);
    dbh_close (dbh);

    struct stat st_old,
      st_new;
    stat (archivo, &st_old);
    stat (archivo2, &st_new);
#ifdef CREATE_REGEN_BACKUP
    char *archivobak = (char *)malloc (strlen (archivo) + strlen (".bak") + 1);
      if (archivobak == NULL) {
	 fprintf(stderr, "malloc: %s\n", strerror(errno));
	exit(1);
     }     
   sprintf (archivobak, "%s.bak", archivo);
    if(remove (archivobak) < 0) {
    }
    TRACE ("DBH: rename %s %s\n", archivo, archivobak);
    if(rename (archivo, archivobak) < 0) ;
    free (archivobak);
#endif
    if(st_old.st_dev == st_new.st_dev) {
        TRACE ("DBH: rename %s %s\n", archivo2, archivo);
        if(rename (archivo2, archivo) < 0) {
            fprintf(stderr, "DBH: cannot write to %s: %s\n", archivo, strerror (errno));
        }
    } else {
        TRACE ("DBH: mv %s %s\n", archivo2, archivo);
        pid_t pid = fork ();
        if(!pid) {
            execlp ("mv", "mv", archivo2, archivo, NULL);
            _exit (123);
        } else {
            int status;
            waitpid (pid, &status, 0);
        }
    }

    newdbh = dbh_new (archivo, NULL, flags);
    free (archivo);
    free (archivo2);
    newdbh->operate = fun;
    return newdbh;
}

static void
mark_erased (DBHashTable * dbh) {
    SET_ERASED;
    /* set file pointer at record flag byte */
    if(place_fp_at (dbh, (dbh->reservedB) + 1) < 0) {
        return;
    }
    /* write the flag to the file */
    /* printf("flag for erase=0x%x\n",dbh->flag); */
    if(write (dbh->fd, &(dbh->flag), 1) != 1) {
        fprintf(stderr, "unable to erase: %s\n", strerror (errno));
        return;
    }
    /*printf("erasing %s flag at %lu\n",(char *)dbh->data,
       (long unsigned)(dbh->reservedB+1)); */
    /* update file header information */
    dbh->head_info->data_space -= dbh->bytes_userdata;
    dbh->head_info->erased_space += (dbh->bytes_userdata);
    dbh_writeheader (dbh);
    return;
}

static void
mark_unerased (DBHashTable * dbh) {
    SET_UNERASED;
    /* set file pointer at record flag byte */
    if(place_fp_at (dbh, (dbh->reservedB) + 1) < 0){
        return;
    }
    /* write the flag to the file */
    if(write (dbh->fd, &(dbh->flag), 1) != 1) {
        fprintf(stderr, "unable to unerase: %s\n", strerror (errno));
        return;
    }
    /* update file header information */
    dbh->head_info->data_space += dbh->bytes_userdata;
    dbh->head_info->erased_space -= (dbh->bytes_userdata);
    dbh_writeheader (dbh);
    return;
}
