#ifdef COPYRIGHT_INFORMATION
#include "gplv3.h"
#endif
/*
 * Copyright (C) 2002-2012 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
typedef struct thread_run_t {
    char *wd;
    char **argv;
} thread_run_t;

typedef struct run_data_t {
    widgets_t *widgets_p;
    pid_t pid;
    gchar *command;
    gchar *workdir;
    GtkWidget *button;
} run_data_t;

/***************  threaded ****************/


static
  gboolean
check_sudo (void) {
    return TRUE;
}
#if 0
FIXME: do the sudo check in configure, or else... tough luck
    // 1. check for sudo askpass option
    FILE *pipe;
    gboolean askpass = FALSE;
    pipe = popen ("sudo -L", "r");
    if(pipe) {
#define PAGE_LINE 2048*8
        char line[PAGE_LINE];
        line[PAGE_LINE - 1] = 0;
        while(fgets (line, PAGE_LINE - 1, pipe) && !feof (pipe)) {
            if(strstr (line, "askpass"))
                askpass = TRUE;
        }
        pclose (pipe);
    }
    if(!askpass) {
        g_warning ("Sudo version obsolete. Please update sudo!\n");
    }
    return askpass;
}
#endif

// This is a thread function, must have GDK mutex set for gtk commands...
static
void
run_fork_finished_function (void *user_data) {
    widgets_t *widgets_p = user_data;
    // decrement thread count here

#if 0
    gchar *g = g_strdup_printf ("Cleanup: %d> pid=%d", Tubo_id () - 1, getpid());
    if(rfm_threaded_diagnostics_is_visible (widgets_p)) {
        rfm_threaded_diagnostics (widgets_p, "xffm_tag/command_id", g);
        rfm_threaded_diagnostics (widgets_p, "xffm/stock_no", g_strdup("\n"));
    }
#endif

}

static GMutex *fork_mutex=NULL;
static void
fork_function (void *data) {
    gchar **argv = (char **)data;

    gint i = 0;
    static gchar *sudo_cmd=NULL;
    if (!fork_mutex) rfm_mutex_init(fork_mutex);
    g_mutex_lock(fork_mutex);
    g_free(sudo_cmd);
    sudo_cmd=NULL;
    for(i=0; argv && argv[i] && i < 5; i++) {
	if(!sudo_cmd && 
		(strstr (argv[i], "sudo") ||
		 strstr (argv[i], "ssh")  ||
		 strstr (argv[i], "rsync")  ||
		 strstr (argv[i], "scp"))) {
	    sudo_cmd=g_strdup_printf("<b>%s</b> ", argv[i]);
	    continue;
	} 	
	if (sudo_cmd){
	    if (strchr(argv[i], '&')){
	        gchar **a = g_strsplit(argv[i], "&", -1);
		gchar **p=a;
		for (;p && *p; p++){
		    gchar *space = (strlen(*p))?" ":"";
		    gchar *amp = (*(p+1))?"&amp;":"";
		    gchar *g = g_strconcat(sudo_cmd,  space, "<i>",*p, amp, "</i>", NULL);
		    g_free(sudo_cmd);
		    sudo_cmd=g;
		}
		g_strfreev(a);
	    } else {
		gchar *a = g_strdup(argv[i]);
		if (strlen(a) >13) {
		    a[12] = 0;
		    a[11] = '.';
		    a[10] = '.';
		    a[9] = '.';
		}
		gchar *g = g_strconcat(sudo_cmd,  " <i>",a, "</i>", NULL);
		g_free(a);
		g_free(sudo_cmd);
		sudo_cmd=g;
	    }
	}
    }


    if (i>=MAX_COMMAND_ARGS - 1) {
    	NOOP("%s: (> %d)\n", strerror(E2BIG), MAX_COMMAND_ARGS);
	argv[MAX_COMMAND_ARGS - 1]=NULL;
    }

    if (sudo_cmd) {
	gchar *g = g_strconcat(sudo_cmd,  "\n", NULL);
	g_free(sudo_cmd);
	sudo_cmd = g;
	// This  function makes copies of the strings pointed to by name and value
        // (by contrast with putenv(3))
	setenv("RFM_ASKPASS_COMMAND", sudo_cmd, 1);
	g_free(sudo_cmd);
    } else {
	setenv("RFM_ASKPASS_COMMAND", "", 1);
    }
    g_mutex_unlock(fork_mutex);
    execvp (argv[0], argv);
    g_warning ("CHILD could not execvp: this should not happen\n");
    g_warning ("Do you have %s in your path?\n", argv[0]);
    rfm_threadwait ();
    _exit (123);
}

static gboolean
destroy_run_button (GtkWidget * widget, GdkEvent * event, gpointer data) {
    run_data_t *run_data_p = (run_data_t *) data;
    NOOP ("destroy_run_button now!\n");
    if(!data)
        return TRUE;
    run_data_p->button = NULL;
    NOOP ("run_data_p->button = NULL\n");
    return TRUE;
}

// This is main thread callback
static void
show_run_info (GtkButton * button, gpointer data) {
    if (rfm_global_p && g_thread_self() != rfm_global_p->self){
	g_error("show_run_info() is a main thread function\n");
    }
    run_data_t *run_data_p = data;
    guint button_id = GPOINTER_TO_UINT(g_object_get_data(G_OBJECT(button), "button_id"));
    if (rfm_void(PLUGIN_DIR, "ps", "module_active")){
	if (button_id == 3){
	    NOOP(stderr, "ps popup now... \n");
	    record_entry_t *en = rfm_mk_entry(0);
	    en->type=0; /* remove local-type attributes */
	    en->st = (struct stat *)malloc(sizeof(struct stat));
	    if (!en->st) g_error("malloc: %s\n", strerror(errno));
	    memset(en->st,0,sizeof(struct stat));

	    pid_t child = Tubo_child(run_data_p->pid);
	    en->path = g_strdup_printf("%d:%s", child, run_data_p->command);
	    en->st->st_uid = child;
	    
	    rfm_rational(PLUGIN_DIR, "ps", NULL, en, "private_popup");
	    NOOP(stderr, "popup mapped...\n");
	    return;
	}
	gchar *ps_module = g_find_program_in_path("rodent-ps");
	if (ps_module) {
	    gchar *command=g_strdup_printf("%s %d",
		    ps_module, (gint) run_data_p->pid);
	    GError *error=NULL;
	    if (!g_spawn_command_line_async (command, &error)){
		g_warning("%s: %s", ps_module, error->message);
		g_error_free(error);
		g_free(command);
		// here we default to the terminate dialog...
	    } else {
		g_free(command);
		return;
	    }
	}
    } else { // no rodent-ps
	if (button_id == 3) return;
	
	gchar *text2 =g_strdup_printf ("%s %s: %s\n\n%s %s (%d)?",
		_("Kill (KILL)"), run_data_p->command, 
		strerror(ETIMEDOUT), 
		_("Kill"), run_data_p->command, run_data_p->pid);
	if(rfm_confirm (run_data_p->widgets_p, GTK_MESSAGE_QUESTION, text2, _("No"), _("Yes"))) {
	    // SIGUSR2 will pass on a SIGKILL signal
	    gchar *gg = g_strdup_printf ("%d", (int)(run_data_p->pid));
	    rfm_diagnostics (run_data_p->widgets_p, "xffm/stock_dialog-warning", NULL);
	    rfm_diagnostics (run_data_p->widgets_p, "xffm_tag/command_id",
		    _("Kill (KILL)"), " ", gg, "\n", NULL);
	    g_free (gg);
	    kill (run_data_p->pid, SIGUSR2);
	}
	g_free (text2);
    }
}

static void *
make_run_data_button (gpointer data) {
    run_data_t * run_data_p = data;
    if(run_data_p->widgets_p->button_space) {
	pid_t pid = Tubo_child(run_data_p->pid);
	if (pid < 0) {
	    TRACE("Tubo_child  < 0\n");
	    return NULL;
	}
	gchar *text = g_strdup(_("Left click once to follow this link.\nMiddle click once to select this cell"));
	if (strstr(text, "\n")) *strstr(text, "\n") = 0;
	gchar *tip=g_strdup_printf("%s (%s=%d)\n%s\n%s",  
		run_data_p->command, _("PID"), pid,
	      _("Right clicking pops context menu immediately"),
	      text);
	g_free(text);
	const gchar *icon_id=(rfm_void(PLUGIN_DIR, "ps", "module_active"))?
		"xffm/stock_execute": "xffm/stock_stop";
        run_data_p->button =
            rfm_mk_little_button (icon_id,
                                  (gpointer) show_run_info, run_data_p,
				  tip);
	g_free(tip);
        gtk_box_pack_end (GTK_BOX (run_data_p->widgets_p->button_space), run_data_p->button, FALSE, FALSE, 0);
        g_signal_connect (G_OBJECT (run_data_p->button), "destroy_event", G_CALLBACK (destroy_run_button), run_data_p);
        gtk_widget_show (run_data_p->button);
        NOOP ("DIAGNOSTICS:srun_button made for grandchildPID=%d\n", (int)run_data_p->pid);
    }
    return NULL;
}

static void *
zap_run_button(gpointer data){
	    NOOP(stderr, "zap_run_button...\n");
    if (rfm_global_p && g_thread_self() != rfm_global_p->self){
	g_error("zap_run_button() is a main thread function\n");
    }
    run_data_t *run_data_p = data;
    gtk_widget_hide(GTK_WIDGET (run_data_p->button));
    gtk_widget_destroy (GTK_WIDGET (run_data_p->button));
    g_free (run_data_p->command);
    g_free (run_data_p->workdir);
    g_free (run_data_p);
    return NULL;
}

// This is a thread function...
static gpointer
thread_run_f (gpointer data) {
    run_data_t *run_data_p = (run_data_t *) data;

    /* create little button (note: thread protected within subroutine) */
    rfm_context_function(make_run_data_button, run_data_p);
    NOOP ("grand---thread waitpid for %d on (%s/%s)\n", run_data_p->pid, run_data_p->command, run_data_p->workdir);

    gint status;
    waitpid (run_data_p->pid, &status, 0);

    NOOP ("grand---thread waitpid for %d complete!\n", run_data_p->pid);
    /* remove little button (thread protect gtk here) */

    // Trigger reload to all tabbed views.
    GSList *reload_list = NULL;
    if(rfm_global_p) {
	rfm_rw_lock_reader_lock (&(rfm_global_p->view_list_lock));
        GSList *list = rfm_global_p->window_view_list;
	for (; list && list->data; list = list->next){
	    reload_list=g_slist_prepend(reload_list, list->data);
	}
	rfm_rw_lock_reader_unlock (&(rfm_global_p->view_list_lock));
    }
    GSList *list = reload_list;
    for (; list && list->data; list = list->next){
	  view_t *view_p = list-> data;
	  if (!xfdir_monitor_control_greenlight(&(view_p->widgets)))
	  {
		rodent_trigger_reload(view_p);
	  }
    }

    if(run_data_p->button && GTK_IS_WIDGET(run_data_p->button)){
	// Run has completed.   
	rfm_context_function(zap_run_button, run_data_p);
    }
    return NULL;
}


static void
setup_run_button_thread (widgets_t * widgets_p, const gchar * exec_command, pid_t child) {
    NOOP ("setup_run_button_thread(), grandchildPID=%d\n", (int)child);
    run_data_t *run_data_p = (run_data_t *) malloc (sizeof (run_data_t));
    if (!run_data_p) g_error("malloc: %s", strerror(errno));
    memset (run_data_p, 0, sizeof (run_data_t));

    run_data_p->button = NULL;
    run_data_p->pid = child;
    run_data_p->command = g_strdup (exec_command);
    if (widgets_p->workdir) {
	run_data_p->workdir = g_strdup (widgets_p->workdir);
    } else {
	run_data_p->workdir = g_strdup (g_get_home_dir());
    }
    run_data_p->widgets_p = widgets_p;

    view_t *view_p = widgets_p->view_p;
    rfm_view_thread_create(view_p, thread_run_f, (gpointer) run_data_p, "thread_run_f");
}

static pid_t
thread_run (
	widgets_t * widgets_p, 
	gchar ** argv, 
	gint *stdin_fd,
	void (*stdout_f) (void *stdout_data,
                      void *stream,
                      int childFD),
	void (*stderr_f) (void *stderr_data,
                      void *stream,
                      int childFD),
	void (*tubo_done_f) (void *data)
	) {

    if (!widgets_p) {
	g_warning("thread_run() is being called with NULL widgets_p\n");
    }
    NOOP(stderr, "At thread_run(), widgets_p=0x%x\n",
	    GPOINTER_TO_INT(widgets_p));


    if(widgets_p && widgets_p->workdir && strcmp(g_get_home_dir(),widgets_p->workdir)){
	NOOP( "At chdir(%s)\n", widgets_p->workdir);
	if (chdir (widgets_p->workdir) < 0) {
	    g_warning ("chdir(%s): %s", widgets_p->workdir, strerror (errno));
	}
    } else {
	if (chdir (g_get_home_dir()) < 0) {
	    g_warning ("chdir(%s): %s", g_get_home_dir(), strerror (errno));
	}
    }
    NOOP( "At g_find_program_in_path()\n");
    gchar *g = g_find_program_in_path (argv[0]);
    if(g) g_free (g);
    else if(widgets_p) {
	rfm_threaded_show_text(widgets_p);
	rfm_threaded_diagnostics (widgets_p, "xffm/stock_dialog-error", NULL);
	rfm_threaded_diagnostics (widgets_p, "xffm_tag/stderr", g_strconcat(argv[0], ": ", strerror (ENOENT), "\n", NULL));      
	return -1;
    }


#if DEBUG_TRACE
    if (rfm_global_p) TRACE( "exec Tubo() now: %s\n", (rfm_global_p->self != g_thread_self())?"CHILD fork":"MAIN fork" );
#endif

    pid_t child = Tubo_threads (fork_function,
		  (void *)argv,
		  stdin_fd,
		  (stdout_f)?stdout_f:rfm_operate_stdout,
		  (stderr_f)?stderr_f:rfm_operate_stderr,
		  (tubo_done_f)?tubo_done_f:run_fork_finished_function,
		  (void *)widgets_p,
		  FALSE,
		  TRUE);
    NOOP(stderr, "exec Tubo():%d\n", child);
      
    if (widgets_p){
    //g_free (widgets_p->workdir);
    NOOP(stderr, "exec Tubo():GETWD\n");
	widgets_p->workdir = g_strdup (GETWD);
	NOOP(stderr, "exec Tubo():SETWD\n");
	SETWD();
	// Trigger a monitor reload condition now.
	// This is to get any initial changes to the view before the command completes

	NOOP(stderr, "xfdir_monitor_control_greenlight now\n");
	xfdir_monitor_control_greenlight(widgets_p);
    }
    
    return child;

}

static
pid_t
private_rfm_thread_run_argv (
	widgets_t * widgets_p, 
	gchar ** argv, gboolean interm, 
	gint *stdin_fd,
	void (*stdout_f) (void *stdout_data,
                      void *stream,
                      int childFD),
	void (*stderr_f) (void *stderr_data,
                      void *stream,
                      int childFD),
	void (*tubo_done_f) (void *data)
	) {
    int i = 0;
    gchar *v_argv[MAX_COMMAND_ARGS];
    const gchar *term = NULL;
    NOOP(stderr, "At private_rfm_thread_run_argv()\n");
    if(interm) {
	term = rfm_what_term ();
        const gchar *exec_option = rfm_term_exec_option(term);
        v_argv[i++] = (gchar *)term;
        v_argv[i++] = (gchar *)exec_option;
    }

 
    gchar **p;
    for(p = argv; p && *p && i < MAX_COMMAND_ARGS - 2; p++) {
        v_argv[i++] = *p;
    }
    v_argv[i] = NULL;
    
    if (i==MAX_COMMAND_ARGS - 1) {
	rfm_threaded_show_text(widgets_p);
    	rfm_threaded_diagnostics(widgets_p,"xffm/stock_dialog-error",NULL);
        gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
        rfm_threaded_diagnostics (widgets_p, "xffm_tag/stderr", g_strconcat(strerror(E2BIG)," (> ",max,")","\n", NULL));
        g_free(max);
    }
    int id=Tubo_id ();
    gchar *command = g_strdup (v_argv[0]);
    gchar *qq;
    for(i = 1; v_argv[i]; i++) {
            qq = command;
            command = g_strconcat (qq, " ", v_argv[i], NULL);
            g_free (qq);
    }
    gchar *s = strstr(command, "password=");
    if (s) {
	s = s + strlen("password=");
	for (; s && *s && *s !=' ' && *s !=','; s++) *s = '*';
    }
	

    NOOP (stderr, "private_rfm_thread_run_argv(): thread_run\n");
    pid_t child = thread_run (widgets_p, v_argv, stdin_fd, stdout_f, stderr_f, tubo_done_f);
    gboolean visible;
    if (rfm_global_p) {
	visible = rfm_threaded_diagnostics_is_visible (widgets_p);
    } else {
	visible = rfm_diagnostics_is_visible (widgets_p);
    }
    if(visible) {
        gchar *g = g_strdup_printf ("%d> (%d):", id, child);
        rfm_threaded_diagnostics (widgets_p, "xffm/stock_execute", NULL);
        rfm_threaded_diagnostics (widgets_p, "xffm_tag/command_id", g);
	gchar **ap;
	gint ac;
	if (g_shell_parse_argv (command, &ac, &ap, NULL)){
	    gchar **a = ap;
	    for (;a && *a; a++){
		if (strchr(*a, '*') || strchr(*a, '?') || strchr(*a, ' ')){
		    rfm_threaded_diagnostics (widgets_p, "xffm_tag/command_id", g_strconcat(" \"", *a, "\"", NULL));
		} else {
		    rfm_threaded_diagnostics (widgets_p, "xffm_tag/command_id", g_strconcat(" ", *a, NULL));
		}
	    } 
	    rfm_threaded_diagnostics (widgets_p, NULL, g_strdup("\n"));	
	    g_strfreev(ap);
	} else {
	    rfm_threaded_diagnostics (widgets_p, "xffm_tag/command", g_strconcat(command, "\n", NULL));
	}
    }
    if(widgets_p && child > 0) {
        setup_run_button_thread (widgets_p, command, child);
    }
    g_free (command);
    return child;
}



/*******************************************/

