/* 
 *
 *  Copyright (C) 2012 Edscott Wilson Garcia <edscott@xfce.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define XMLTREE_C
#include "rfm_libs.h"
/* this should be first 2 lines after headers: */
G_MODULE_EXPORT LIBRFM_MODULE 

#include "xmltree.i"


/////////////////////////////////////////////////////////////////////////
//    WINDOW CREATION FUNCTIONALITY (callback)
/////////////////////////////////////////////////////////////////////////

G_MODULE_EXPORT
void *
xmltree (GtkButton * button, gpointer data) {
    GError *error=NULL;
    gchar *keybindings_file = g_build_filename(KEYBINDINGS_FILE, NULL);
    Tag_t *Tag_p;
    if (g_file_test(keybindings_file, G_FILE_TEST_EXISTS)){
	Tag_p =tag_new_from_file(keybindings_file, &error);
	// validate with schema and dump if not valid
	gchar *schema_file = g_build_filename(KEYBINDINGS_SCHEMA, NULL);
	tag_load_schema(Tag_p, schema_file, &error);
	if (!tag_validate(Tag_p)){
	    gchar *text=g_strdup_printf("%s: %s\n\n%s",
		    _("Validate document"), keybindings_file,
		    _("The Document is not valid!"));
	    rfm_confirm(NULL, GTK_MESSAGE_ERROR, text, NULL,NULL);
	    g_free(text);
	    tag_free(Tag_p);
	    Tag_p = tag_new();
	} 
	g_free(schema_file);
    } else {
	Tag_p = tag_new();
    }
    // Add anything missing (may be everything)...
    keybindings (Tag_p);
    gchar *schema_file = g_build_filename(KEYBINDINGS_SCHEMA, NULL);
    Tag_t *schema_Tag_p = tag_new_from_schema_file(schema_file, &error);
    g_free(schema_file);

    GtkTreeModel *schemamodel= GTK_TREE_MODEL(gtk_tree_store_new (H_TREE_COLUMNS,
	    G_TYPE_POINTER,   // tag_item_p or parent's tag_item_p
	    G_TYPE_POINTER,   // attribute_item_p if applicable.
	    GDK_TYPE_PIXBUF, // pretty stuff
	    G_TYPE_STRING,   // tag
	    G_TYPE_STRING,   // attribute (string)
	    G_TYPE_STRING,   // value (string)
	    G_TYPE_STRING,    // prefix (string)
	    GDK_TYPE_COLOR    // color format
	));
	    // G_TYPE_BOXED, // this might be useful later on
	    // G_TYPE_OBJECT,// and this one too...

    // Create treemodels
    GtkTreeModel *model= GTK_TREE_MODEL(gtk_tree_store_new (H_TREE_COLUMNS,
	    G_TYPE_POINTER,   // tag_item_p or parent's tag_item_p
	    G_TYPE_POINTER,   // attribute_item_p if applicable.
	    GDK_TYPE_PIXBUF, // pretty stuff
	    G_TYPE_STRING,   // tag
	    G_TYPE_STRING,   // attribute (string)
	    G_TYPE_STRING,   // value (string)
	    G_TYPE_STRING,    // prefix (string)
	    GDK_TYPE_COLOR    // color format
	));

    if (!GTK_IS_TREE_MODEL(model)){
	g_error("build_treemodel(): cannot create tree model!");
    }
    // Create tag
    populate_tree_model_from_tag(Tag_p, model, &error);
    // Build a graphic treeview.
    // Xml to the right
    GtkTreeView *treeview = build_treeview(model);
    GtkTreePath *ipath = gtk_tree_path_new_from_string("0");
    gtk_tree_view_expand_row(treeview, ipath, FALSE);
    gtk_tree_path_free(ipath);

    g_object_set_data(G_OBJECT(model), "Tag_p", Tag_p);

    gtk_tree_model_foreach(model, switch_colors, NULL);

    // Container for the graphic treeview.
    GtkWidget *treebox = build_treeview_box(treeview);


    GtkWidget *schemabox = NULL;
    // Build schema treeview
    if (schema_Tag_p) {
	//Create schema tag
	populate_tree_model_from_tag(schema_Tag_p, schemamodel, &error);
	// Schema on the left.
	GtkTreeView *schemaview = build_treeview(schemamodel);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW(schemaview), FALSE);
	// Container for the graphic schemaview.
	schemabox = build_treeview_box(schemaview);
    }

    // Create a window for the treeview container.
    GtkWidget *window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    g_object_set_data(G_OBJECT(window), "Tag_p", Tag_p);
    g_object_set_data(G_OBJECT(window), "model", model);
    // Create a shortcut reference.
    g_object_set_data(G_OBJECT(treeview), "parent_window", window);
    g_signal_connect(treeview, "key-press-event", G_CALLBACK(treeview_key), NULL);    
    
    const gchar *window_title = _("Configuration of keybindings");



    gtk_window_set_title (GTK_WINDOW (window), window_title);
    gtk_window_set_default_size (GTK_WINDOW (window), 640, 480);
    GtkWidget *vbox = rfm_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (window), vbox);

    // Add a hpane
    //GtkWidget *hpane = gtk_paned_new(GTK_ORIENTATION_HORIZONTAL);
    GtkWidget *hpane = rfm_hpaned_new();
    gtk_box_pack_start (GTK_BOX (vbox), hpane, TRUE, TRUE, 3);
    //gtk_container_add (GTK_CONTAINER (vbox), hpane);
    if (schemabox) {
	gtk_paned_add1 (GTK_PANED(hpane), schemabox);
	gtk_paned_add2 (GTK_PANED(hpane), treebox);
    } else {
	gtk_paned_add1 (GTK_PANED(hpane), treebox);
    }

    GtkWidget * hbox = rfm_hbox_new (TRUE, 0);
    GtkWidget *b;
    b = rfm_mk_little_button ("xffm/stock_save", 
		      (gpointer)xml_edit_save, window, _("Save"));
    gtk_box_pack_start (GTK_BOX (hbox), b, TRUE, FALSE, 3);
    b = rfm_mk_little_button ("xffm/stock_cancel", 
		      (gpointer)xml_edit_destroy, window, _("Cancel"));
    gtk_box_pack_start (GTK_BOX (hbox), b, TRUE, FALSE, 3);

    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 3);
  

    g_signal_connect(G_OBJECT(window), "destroy-event", G_CALLBACK (on_destroy_k), NULL);
    g_signal_connect(G_OBJECT(window), "delete-event", G_CALLBACK (on_destroy_k), NULL);

 
    gtk_widget_show_all(window);
    gtk_paned_set_position(GTK_PANED(hpane), 0); 

    // Event loop.
    gtk_main();
    return NULL;
}

void *load_custom_keybindings(void *p){
    gchar *keybindings_file = g_build_filename(KEYBINDINGS_FILE, NULL);
    if (g_file_test(keybindings_file, G_FILE_TEST_EXISTS)){
	NOOP(stderr, "load_custom_keybindings now\n");
    }
    g_free(keybindings_file);
    return NULL;
}



/////////////////////////////////////////////////////////////////////////

