#ifdef COPYRIGHT_INFORMATION
#include "gplv3.h"
#endif
///////////////////////////////  Keybinding Callbacks /////////////////////

typedef struct menu_callback_t {
    gint function_id;
    const gchar *string;
    const gchar *icon;
    gpointer function;
    gpointer data;
    guint key; 
    guint mask; 
    gint type;
}menu_callback_t;


enum {
    NULL_TYPE,
    SUBMENU_TYPE,
    MENUITEM_TYPE,
    CHECKITEM_TYPE,
    RADIOITEM_TYPE,
    SEPARATOR_TYPE
};

typedef struct menu_definitions_t{
    gint type;
    const gchar *parent_id;
    const gchar *id;
    menu_callback_t callback;
} menu_definitions_t;


#define TOGGLE_FLAG \
       ((1<<HIDDEN_TOGGLED)|(1<<PREVIEW_TOGGLED)|\
	(1<<CASESORT_TOGGLED)|(1<<NAME_SORT_ACTIVATE)|\
	(1<<TYPE_SORT_ACTIVATE)|(1<<DATE_SORT_ACTIVATE)|\
	(1<<SIZE_SORT_ACTIVATE)|(1<<OWNER_SORT_ACTIVATE)|\
	(1<<GROUP_SORT_ACTIVATE)|(1<<MODE_SORT_ACTIVATE)|\
	(1<<COMPACT_ICONSIZE_TOGGLED)|(1<<TINY_ICONSIZE_TOGGLED)|\
	(1<<NORMAL_ICONSIZE_TOGGLED)|(1<<BIG_ICONSIZE_TOGGLED)|\
	(1<<HUGE_ICONSIZE_TOGGLED))
#define IS_TOGGLE(x) ((x <= HUGE_ICONSIZE_TOGGLED) && ((1<<x) & TOGGLE_FLAG))
#define RADIO_FLAG \
       ((1<<NAME_SORT_ACTIVATE)|\
	(1<<TYPE_SORT_ACTIVATE)|(1<<DATE_SORT_ACTIVATE)|\
	(1<<SIZE_SORT_ACTIVATE)|(1<<OWNER_SORT_ACTIVATE)|\
	(1<<GROUP_SORT_ACTIVATE)|(1<<MODE_SORT_ACTIVATE)|\
	(1<<COMPACT_ICONSIZE_TOGGLED)|(1<<TINY_ICONSIZE_TOGGLED)|\
	(1<<NORMAL_ICONSIZE_TOGGLED)|(1<<BIG_ICONSIZE_TOGGLED)|\
	(1<<HUGE_ICONSIZE_TOGGLED))
#define IS_RADIO(x) ((x <= HUGE_ICONSIZE_TOGGLED) && ((1<<x) & RADIO_FLAG))

// PLEASE NOTE:
// 
// If items in menu_enum_t are moved around, or items added or removed
// you must update KEYBINDINGS_FILE (files.h) to current build.
//
// You have been warned...
//
enum menu_enum_t {
    ACTIVATE, // This should be the first.
    // Toggles (value is use to set a bit flag with shift)
    // NOTE: Checkboxes and toggles should always follow ACTIVATE
    // and order important because of the shift flag defined
    // above.
    HIDDEN_TOGGLED,
    PREVIEW_TOGGLED,
    CASESORT_TOGGLED,
    ASCENDING_ACTIVATE,
    DESCENDING_ACTIVATE,
    // Radio items: sort
    NAME_SORT_ACTIVATE,
    TYPE_SORT_ACTIVATE,
    DATE_SORT_ACTIVATE,
    SIZE_SORT_ACTIVATE,
    OWNER_SORT_ACTIVATE,
    GROUP_SORT_ACTIVATE,
    MODE_SORT_ACTIVATE,
    // Radio items: size
    COMPACT_ICONSIZE_TOGGLED,
    TINY_ICONSIZE_TOGGLED,
    NORMAL_ICONSIZE_TOGGLED,
    BIG_ICONSIZE_TOGGLED,
    HUGE_ICONSIZE_TOGGLED,
    //
    // These are all normal callbacks.
    // Ordering is now irrelevant from this point on.
    PLUS_ICONSIZE_ACTIVATE,
    MINUS_ICONSIZE_ACTIVATE,    
    DEFAULT_ICONSIZE_ACTIVATE,
    DEFAULT_ICONSIZE_ALL_ACTIVATE,

    MENU_ACTIVATE,
    GLOB_ACTIVATE,
    FIND_ACTIVATE,
    HELP_ACTIVATE,
    REMOVE_ACTIVATE,
    REFRESH_ACTIVATE,
    RELOAD_ACTIVATE,
    HOST_ACTIVATE,
    GOUP_ACTIVATE,
    FORWARD_ACTIVATE,
    BACK_ACTIVATE,
    HOME_ACTIVATE,
    JUMP_TO_ACTIVATE,
    PLUGIN_GOTO_ACTIVATE,
    LEVEL_GOTO_ACTIVATE,
    RUN_ACTIVATE,
    OPEN_WITH_ACTIVATE,
    ABOUT_ACTIVATE,
    NEW_WINDOW_ACTIVATE,
    NEW_TAB_ACTIVATE,
    OPEN_IN_TERMINAL_ACTIVATE,
    DIFFERENCES_ACTIVATE,
    SETTINGS_ACTIVATE,
    NEWDIR_ACTIVATE,
    NEWFILE_ACTIVATE,
    AUTOTYPE_ACTIVATE,
    AUTOTYPE_R_ACTIVATE,
    RENAME_ACTIVATE,
    DUPLICATE_ACTIVATE,
    SYMLINK_ACTIVATE,
    DONE_WITH_RENAME,
    TOUCH_ACTIVATE,
    PROPERTIES_ACTIVATE,
    UNSELECT_ALL_ACTIVATE,
    SELECT_ALL_ACTIVATE,
    SELECT_INVERT_ACTIVATE,
    SELECT_BY_FILTER_ACTIVATE,
    UNSELECT_BY_FILTER_ACTIVATE,
    MOUNT_ACTIVATE,
    UNMOUNT_ACTIVATE,
    CUT_ACTIVATE,
    COPY_ACTIVATE,
    PASTE_ACTIVATE,
    CLOSE_ACTIVATE,
    QUIT_ACTIVATE,
    BOOKMARK_ADD_ACTIVATE,
    BOOKMARK_REMOVE_ACTIVATE,
    BOOKMARK_TOGGLED,
    KEYBINDINGS_ACTIVATE,


    ENUM_CALLBACKS
};


/* Use as many different names for as many different
 * autotype menu entries for one file extension type. */
#define AUTO_C_NAME \
    "autotype-0",\
    "autotype-1",\
    "autotype-2",\
    "autotype-3",\
    "autotype-4",\
    "autotype-5",\
    "autotype-6",\
    "autotype-7",\
    "autotype-8",\
    "autotype-9",\
    NULL


#define NEW_FILE_CALLBACK \
    {NEWFILE_ACTIVATE, N_("Create new file in the given directory"),"xffm/stock_new", \
    NULL, NULL,		GDK_KEY_Insert,	GDK_CONTROL_MASK }

#define NEWDIR_CALLBACK \
    {NEWDIR_ACTIVATE, N_("Create a new empty folder inside this folder"),"xffm/actions_folder-new",\
    NULL, NULL,		GDK_KEY_N,	(GDK_CONTROL_MASK|GDK_SHIFT_MASK) }

#define RENAME_CALLBACK \
    {RENAME_ACTIVATE, N_("Rename"),"xffm/stock_save-as",\
    NULL, NULL,		GDK_KEY_R,	(GDK_CONTROL_MASK|GDK_SHIFT_MASK)}

#define DUPLICATE_CALLBACK \
    {DUPLICATE_ACTIVATE, N_("Duplicate"),"xffm/stock_dnd-multiple",\
    NULL, NULL,	GDK_KEY_d,	GDK_CONTROL_MASK}

#define SYMLINK_CALLBACK \
    {SYMLINK_ACTIVATE, N_("Link"),"xffm/stock_convert",\
    NULL, NULL,		GDK_KEY_s,	GDK_CONTROL_MASK}

#define TOUCH_CALLBACK \
    {TOUCH_ACTIVATE, N_("Touch"),"xffm/stock_index",\
    NULL, NULL,		GDK_KEY_T,	(GDK_CONTROL_MASK|GDK_SHIFT_MASK)}
#define PROPERTIES_CALLBACK \
    {PROPERTIES_ACTIVATE, N_("Properties"),"xffm/stock_properties",\
    NULL, NULL,		GDK_KEY_p,	GDK_CONTROL_MASK}

#define NEW_WINDOW_CALLBACK \
    {NEW_WINDOW_ACTIVATE, N_("Open in New Window"),"xffm/places_folder-favorites",\
   NULL, NULL,	GDK_KEY_n,	GDK_CONTROL_MASK}
//Different button icon for the toolbar
#define NEW_TAB_CALLBACK_B \
    {NEW_TAB_ACTIVATE, N_("Open in New Tab"),"xffm/stock_add",\
   NULL, NULL,	0,	0}

#define NEW_TAB_CALLBACK \
    {NEW_TAB_ACTIVATE, N_("Open in New Tab"),"xffm/stock_directory",\
   NULL, NULL,	GDK_KEY_t,	GDK_CONTROL_MASK}

#define SHOW_HIDDEN_CALLBACK \
    {HIDDEN_TOGGLED, N_("Show hidden files"),"xffm/emblem_show-hidden",\
   NULL, NULL,		GDK_KEY_h,	GDK_CONTROL_MASK, CHECKITEM_TYPE}

#define PREVIEW_IMAGES_CALLBACK \
    {PREVIEW_TOGGLED, N_("Show previews of files and directories."),"xffm/emblem_show-previews",\
   NULL, NULL,		GDK_KEY_i,	GDK_CONTROL_MASK, CHECKITEM_TYPE}

#define ABOUT_CALLBACK \
    {ABOUT_ACTIVATE, N_("About"),"xffm/stock_about",\
   NULL, NULL,		GDK_KEY_A,	(GDK_CONTROL_MASK|GDK_SHIFT_MASK)}

#define JUMP_TO_CALLBACK \
    {JUMP_TO_ACTIVATE, N_("Choose Location"),"xffm/stock_jump-to",\
   NULL, NULL,		GDK_KEY_Down,	GDK_MOD5_MASK|GDK_MOD1_MASK}

#define GOTO_HOST_CALLBACK \
    {HOST_ACTIVATE, N_("System"), "xffm/device_computer",\
   NULL, NULL,		GDK_KEY_End,	GDK_MOD5_MASK|GDK_MOD1_MASK}

#define GOTO_HOME_CALLBACK \
    {HOME_ACTIVATE, N_("Home"),"xffm/stock_home",\
   NULL, NULL,		GDK_KEY_Home,	GDK_MOD5_MASK|GDK_MOD1_MASK}

#define GO_UP_CALLBACK \
    {GOUP_ACTIVATE, N_("Go up"),"xffm/stock_go-up",\
   NULL, NULL,		GDK_KEY_Up,	GDK_MOD5_MASK|GDK_MOD1_MASK}

#define GO_BACK_CALLBACK \
    {BACK_ACTIVATE, N_("Go back"),"xffm/stock_go-back",\
    NULL, NULL,		GDK_KEY_Left,	GDK_MOD5_MASK|GDK_MOD1_MASK}

#define GO_FORWARD_CALLBACK \
    {FORWARD_ACTIVATE, N_("Go forward"),"xffm/stock_go-forward",\
    NULL, NULL,		GDK_KEY_Right,	GDK_MOD5_MASK|GDK_MOD1_MASK}

#define CUT_CALLBACK \
    {CUT_ACTIVATE, N_("Cut"),"xffm/stock_cut",\
   NULL, NULL,		GDK_KEY_x,	GDK_CONTROL_MASK}

#define COPY_CALLBACK \
    {COPY_ACTIVATE, N_("Copy"),"xffm/stock_copy",\
   NULL, NULL,		GDK_KEY_c,	GDK_CONTROL_MASK}

#define PASTE_CALLBACK \
    {PASTE_ACTIVATE, N_("Paste"),"xffm/stock_paste",\
   NULL, NULL,		GDK_KEY_v,	GDK_CONTROL_MASK}

#define REMOVE_CALLBACK \
    {REMOVE_ACTIVATE, N_("Delete"),"xffm/stock_delete",\
   NULL, NULL,		GDK_KEY_Delete,	0}

#define AUTOTYPE_CALLBACK \
    {AUTOTYPE_ACTIVATE, "Mimetype command","xffm/stock_execute",\
   NULL, NULL,		0,	0}

#define AUTOTYPE_R_CALLBACK \
    {AUTOTYPE_R_ACTIVATE, "autotype_Prun","xffm/stock_execute",\
   NULL, NULL,	0,	0}

#define OPEN_WITH_CALLBACK \
    {OPEN_WITH_ACTIVATE, N_("Open with"),"xffm/stock_open",\
   NULL, NULL,	GDK_KEY_o,	GDK_CONTROL_MASK}

#define ADD_BOOKMARK_CALLBACK \
    {BOOKMARK_ADD_ACTIVATE, N_("Add bookmark"),"xffm/places_user-bookmarks/composite/stock_yes",\
   NULL, NULL,	0,	0}

#define REMOVE_BOOKMARK_CALLBACK \
    {BOOKMARK_REMOVE_ACTIVATE, N_("Remove bookmark"),"xffm/places_user-bookmarks/composite/stock_no",\
   NULL, NULL,	0,	0}

#define TOGGLE_BOOKMARK_CALLBACK \
    {BOOKMARK_TOGGLED, N_("Toggle Bookmark"),"xffm/places_user-bookmarks",\
   NULL, NULL,	GDK_KEY_F12,	0, CHECKITEM_TYPE}

#define MOUNT_CALLBACK \
    {MOUNT_ACTIVATE, N_("Mount the volume associated with this folder"),"xffm/stock_yes",\
   NULL, NULL,		GDK_KEY_m,	GDK_CONTROL_MASK}

#define UNMOUNT_CALLBACK \
    {UNMOUNT_ACTIVATE, N_("Unmount the volume associated with this folder"),"xffm/stock_no",\
   NULL, NULL,		GDK_KEY_M,	(GDK_CONTROL_MASK|GDK_SHIFT_MASK)}

#define CLOSE_CALLBACK \
    {CLOSE_ACTIVATE, N_("Close"),"xffm/stock_close",\
   NULL, NULL,		GDK_KEY_q,	GDK_CONTROL_MASK}

#define HELP_CALLBACK \
    {HELP_ACTIVATE, N_("Help"),"xffm/stock_help",\
   NULL, NULL,		GDK_KEY_F1,	0}

#define SETTINGS_CALLBACK \
    {SETTINGS_ACTIVATE, N_("Personal settings"),"xffm/stock_preferences",\
   NULL, NULL,		GDK_KEY_F2,	0}

#define TERMINAL_CALLBACK \
    {OPEN_IN_TERMINAL_ACTIVATE, N_("Open terminal"),"xffm/apps_terminal",\
   NULL, NULL,	GDK_KEY_F4,	0}

// This is the same as REFRESH_CALLBACK, except that this one gets displayed in menu:
#define RELOAD_CALLBACK \
    {RELOAD_ACTIVATE, N_("Reload"),"xffm/stock_refresh",\
   NULL, NULL,		GDK_KEY_F5,	0}

#define REFRESH_CALLBACK \
    {REFRESH_ACTIVATE, N_("Reload"),"xffm/stock_refresh",\
   NULL, NULL,		GDK_KEY_r,	GDK_CONTROL_MASK}

#define RUN_CALLBACK \
    {RUN_ACTIVATE, N_("Execute Shell Command"),"xffm/stock_execute",\
    NULL, NULL,		GDK_KEY_F6,	0}

// This is the same as FIND_CALLBACK, except that this one gets displayed in menu:
#define SEARCH_CALLBACK \
    {GLOB_ACTIVATE, N_("Search"),"xffm/stock_find",\
   NULL, NULL,		GDK_KEY_F7,	0}

#define FIND_CALLBACK \
    {FIND_ACTIVATE, N_("Search"),"xffm/stock_find",\
   NULL, NULL,		GDK_KEY_f,	GDK_CONTROL_MASK}

#define DIFFERENCES_CALLBACK \
    {DIFFERENCES_ACTIVATE, N_("Compare Files or Folders"),"xffm/apps_diff-tool",\
   NULL, NULL,	GDK_KEY_F8,	0}

#define CASE_SORT_CALLBACK \
    {CASESORT_TOGGLED, N_("Sort case insensitive"),"xffm/stock_bold",\
   NULL, NULL,		GDK_KEY_F11,	0, CHECKITEM_TYPE}

#define SORT_ASCENDING_CALLBACK \
    {ASCENDING_ACTIVATE, N_("Sort data in ascending order"),"xffm/stock_sort-ascending",\
   NULL, NULL,		GDK_KEY_F2,	GDK_CONTROL_MASK}

#define SORT_DESCENDING_CALLBACK \
    {DESCENDING_ACTIVATE, N_("Sort data in descending order"),"xffm/stock_sort-descending",\
   NULL, NULL,		GDK_KEY_F3,	GDK_CONTROL_MASK}


#define KEYBINDINGS_CALLBACK \
    {KEYBINDINGS_ACTIVATE, N_("Global keybindings support"),"xffm/device_keyboard",\
   NULL, NULL,		GDK_KEY_F1,	GDK_CONTROL_MASK}

#define SELECT_ALL_CALLBACK \
    {SELECT_ALL_ACTIVATE, N_("Select All"),"xffm/stock_file/composite/stock_add",\
    NULL, NULL,	GDK_KEY_a, 	GDK_CONTROL_MASK}

#define SELECT_INVERT_CALLBACK \
    {SELECT_INVERT_ACTIVATE, N_("Invert Selection"),"xffm/actions_flip-horizontal",\
    NULL, \
		NULL,			GDK_KEY_equal, 	GDK_CONTROL_MASK}

#define UNSELECT_ALL_CALLBACK \
    {UNSELECT_ALL_ACTIVATE, N_("Unselect"),"xffm/stock_file/composite/stock_remove",\
    NULL, NULL,	GDK_KEY_u, 	GDK_CONTROL_MASK}

#define SELECT_BYFILTER_CALLBACK \
    {SELECT_BY_FILTER_ACTIVATE, N_("Select Items Matching..."),"xffm/actions_select-items",\
   NULL, \
	NULL,				GDK_KEY_plus, 	GDK_CONTROL_MASK}

#define UNSELECT_BYFILTER_CALLBACK \
    {UNSELECT_BY_FILTER_ACTIVATE, N_("Unselect Items Matching..."),"xffm/actions_select-none",\
    NULL, \
	NULL,				GDK_KEY_minus, 	GDK_CONTROL_MASK}


#define NAME_SORT_CALLBACK \
    {NAME_SORT_ACTIVATE, N_("Sort by name"),NULL,\
   NULL, NULL,	GDK_KEY_F4,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define TYPE_SORT_CALLBACK \
    {TYPE_SORT_ACTIVATE, N_("Default sort order"),NULL,\
   NULL, NULL,		GDK_KEY_F5,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define DATE_SORT_CALLBACK \
    {DATE_SORT_ACTIVATE, N_("Sort by date"),NULL,\
   NULL, NULL,		GDK_KEY_F6,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define SIZE_SORT_CALLBACK \
    {SIZE_SORT_ACTIVATE, N_("Sort by size"),NULL,\
   NULL, NULL,		GDK_KEY_F7,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define OWNER_SORT_CALLBACK \
    {OWNER_SORT_ACTIVATE, N_("User"),NULL,\
   NULL, NULL,		GDK_KEY_F8,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define GROUP_SORT_CALLBACK \
    {GROUP_SORT_ACTIVATE, N_("Group"),NULL,\
   NULL, NULL,		GDK_KEY_F9,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define MODE_SORT_CALLBACK \
    {MODE_SORT_ACTIVATE, N_("By Permissions"),NULL,\
   NULL, NULL,		GDK_KEY_F10,	GDK_CONTROL_MASK, RADIOITEM_TYPE}


#define MENU_CALLBACK \
    {MENU_ACTIVATE, N_("Opens the main menu."),"xffm/places_start-here",\
    NULL, NULL,		GDK_KEY_F3,	0}

#define NULL_CALLBACK  {-1, NULL,NULL, NULL, NULL, 0,0}

////////////////////////////////////////////////////////////////////////////////////


#define PLUS_ICONSIZE_CALLBACK \
    {PLUS_ICONSIZE_ACTIVATE, N_("Enlarge Icons"),"xffm/stock_zoom-in",\
   NULL, NULL, GDK_KEY_KP_Add,	GDK_CONTROL_MASK}

#define MINUS_ICONSIZE_CALLBACK \
    {MINUS_ICONSIZE_ACTIVATE, N_("Shrink Icons"),"xffm/stock_zoom-out",\
   NULL, NULL, GDK_KEY_KP_Subtract,GDK_CONTROL_MASK}

#define DEFAULT_ICONSIZE_CALLBACK \
    {DEFAULT_ICONSIZE_ACTIVATE, N_("Default icon zoom level"),"xffm/stock_zoom-100",\
   NULL, NULL, GDK_KEY_0,	GDK_CONTROL_MASK}

#define DEFAULT_ICONSIZE_ALL_CALLBACK \
    {DEFAULT_ICONSIZE_ALL_ACTIVATE, "Default icon zoom level (all)","xffm/stock_zoom-100/composite/emblem_favorite",\
   NULL, NULL, GDK_KEY_9,	GDK_CONTROL_MASK}

#define COMPACT_ICONSIZE_CALLBACK \
    {COMPACT_ICONSIZE_TOGGLED, N_("View as list"),NULL,\
   NULL, NULL, GDK_KEY_1,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define TINY_ICONSIZE_CALLBACK \
    {TINY_ICONSIZE_TOGGLED, N_("Small Icons"),NULL,\
   NULL, NULL, GDK_KEY_2,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define NORMAL_ICONSIZE_CALLBACK \
    {NORMAL_ICONSIZE_TOGGLED, N_("Normal Icons"),NULL,\
   NULL, NULL, GDK_KEY_3,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define BIG_ICONSIZE_CALLBACK \
    {BIG_ICONSIZE_TOGGLED, N_("Large Icons"),NULL,\
   NULL, NULL, GDK_KEY_4,	GDK_CONTROL_MASK, RADIOITEM_TYPE}

#define HUGE_ICONSIZE_CALLBACK \
    {HUGE_ICONSIZE_TOGGLED, N_("Very large"),NULL,\
   NULL, NULL, GDK_KEY_5,	GDK_CONTROL_MASK, RADIOITEM_TYPE}


//
///////////////////////   button callbacks  /////////////////////////


typedef struct button_callback_t {
    gint id;
    const gchar *button_name;
    menu_callback_t callback;
    const gchar *text;
} button_callback_t;

#define ONE32 ((gint32) 0x01)
#define ONE64 (G_GINT64_CONSTANT(0x01))

#define DEFAULT_TOOLBAR_BUTTONS \
	((ONE64<<NEW_TAB_CALLBACK_B_BUTTON_ID)| \
	(ONE64<<NEW_WINDOW_CALLBACK_BUTTON_ID)| \
	(ONE64<<SORT_DESCENDING_CALLBACK_BUTTON_ID)| \
	(ONE64<<TYPE_SORT_CALLBACK_BUTTON_ID)| \
	(ONE64<<NAME_SORT_CALLBACK_BUTTON_ID)| \
	(ONE64<<DATE_SORT_CALLBACK_BUTTON_ID)| \
	(ONE64<<CASE_SORT_CALLBACK_BUTTON_ID)| \
	(ONE64<<SHOW_HIDDEN_CALLBACK_BUTTON_ID)| \
	(ONE64<<PREVIEW_IMAGES_CALLBACK_BUTTON_ID) \
	)

// If order is changed or items added/removed, then
// you must bump-up configuration file build compatibility
enum button_callback_id_t {
    NEW_WINDOW_CALLBACK_BUTTON_ID,   	// 2
    NEW_FILE_CALLBACK_BUTTON_ID,   	// 0
    NEWDIR_CALLBACK_BUTTON_ID,   	// 1
    ABOUT_CALLBACK_BUTTON_ID,   	// 3 f
    JUMP_TO_CALLBACK_BUTTON_ID,   	// 4
    GOTO_HOST_CALLBACK_BUTTON_ID,   	// 5
    GOTO_HOME_CALLBACK_BUTTON_ID,   	// 6
    GO_UP_CALLBACK_BUTTON_ID,   	// 7 ff
    GO_FORWARD_CALLBACK_BUTTON_ID,   	// 8
    GO_BACK_CALLBACK_BUTTON_ID,   	// 8
    CUT_CALLBACK_BUTTON_ID,   		// 9
    COPY_CALLBACK_BUTTON_ID,   		// 10
    PASTE_CALLBACK_BUTTON_ID,   	// 11 fff
    REMOVE_CALLBACK_BUTTON_ID,   	// 12
    TOGGLE_BOOKMARK_CALLBACK_BUTTON_ID,	// 13
    HELP_CALLBACK_BUTTON_ID,   		// 14
    SETTINGS_CALLBACK_BUTTON_ID,   	// 15 ffff
    TERMINAL_CALLBACK_BUTTON_ID,   	// 16
    REFRESH_CALLBACK_BUTTON_ID,   	// 17
    RUN_CALLBACK_BUTTON_ID,   		// 18
    FIND_CALLBACK_BUTTON_ID,   		// 19 fffff
    DIFFERENCES_CALLBACK_BUTTON_ID,   	// 20
    SORT_ASCENDING_CALLBACK_BUTTON_ID,	// 21
    SORT_DESCENDING_CALLBACK_BUTTON_ID,	// 22
    KEYBINDINGS_CALLBACK_BUTTON_ID,   	// 23 ffffff
    SELECT_ALL_CALLBACK_BUTTON_ID,   	// 24
    SELECT_INVERT_CALLBACK_BUTTON_ID,   // 25
    UNSELECT_ALL_CALLBACK_BUTTON_ID,   	// 26
    SELECT_BYFILTER_CALLBACK_BUTTON_ID,	// 27 fffffff
    UNSELECT_BYFILTER_CALLBACK_BUTTON_ID,   // 28
    TYPE_SORT_CALLBACK_BUTTON_ID,   	//29 
    NAME_SORT_CALLBACK_BUTTON_ID,   	// 30
    DATE_SORT_CALLBACK_BUTTON_ID,   	// 31 ffffffff
    SIZE_SORT_CALLBACK_BUTTON_ID,   	// 32
    OWNER_SORT_CALLBACK_BUTTON_ID,   	// 33
    GROUP_SORT_CALLBACK_BUTTON_ID,   	// 34
    MODE_SORT_CALLBACK_BUTTON_ID,   	// 35 fffffffff
    DEFAULT_ICONSIZE_CALLBACK_BUTTON_ID,// 36
    DEFAULT_ICONSIZE_ALL_CALLBACK_BUTTON_ID,   // 37
    COMPACT_ICONSIZE_CALLBACK_BUTTON_ID,// 38
    TINY_ICONSIZE_CALLBACK_BUTTON_ID,   // 39 ffffffffff
    NORMAL_ICONSIZE_CALLBACK_BUTTON_ID, // 40
    BIG_ICONSIZE_CALLBACK_BUTTON_ID,   	// 41
    HUGE_ICONSIZE_CALLBACK_BUTTON_ID,   // 42
    PLUS_ICONSIZE_CALLBACK_BUTTON_ID,   // 43 fffffffffff
    MINUS_ICONSIZE_CALLBACK_BUTTON_ID,  // 44
    SHOW_HIDDEN_CALLBACK_BUTTON_ID,   	// 45
    PREVIEW_IMAGES_CALLBACK_BUTTON_ID,  // 46 
    CASE_SORT_CALLBACK_BUTTON_ID,   	// 47
    NEW_TAB_CALLBACK_BUTTON_ID,  	// 48
    NEW_TAB_CALLBACK_B_BUTTON_ID,  	// < 64
    BUTTON_OVERFLOW_ID
    };


typedef struct environ_t {
    gchar *env_var;
    gchar *env_string;
    gchar *env_text;
} environ_t;

typedef struct program_options_t{
    const gchar *option;
    const gchar *text;
    const gboolean sensitive;
    gchar **choices;
    gint choice_id;
}program_options_t;


/*
const program_options_t cp_options[]={
    {NULL,NULL}
};
*/
enum {
    /* check buttons: */
    /* general:: */
    RFM_DOUBLE_CLICK_NAVIGATION,/* toggle */
    RFM_USE_GTK_ICON_THEME,     // fallback to default system gtk icon theme 
    RFM_DRAG_DOES_MOVE,         /* toggle */
    RFM_CONTENT_FOLDER_ICONS,
    RFM_ENABLE_TIPS,           /* toggle */
    RFM_ENABLE_LABEL_TIPS,           /* toggle */
    RFM_DIAGNOSTICS_FONT_SIZE,      /* size selector */
    RFM_DEFAULT_ICON_SIZE,      /* size selector */
    TERMINAL_CMD,               /* path selector */
    EDITOR,                 /* path selector */
    RFM_MAXIMUM_COMPLETION_OPTIONS, 
    RFM_MAXIMUM_DIAGNOSTIC_LINES, 
    RFM_LOAD_TIMEOUT, 
    /* desktop:: */
    RFM_ENABLE_DESKTOP,         /* toggle */
    RFM_ENABLE_DESKTOP_DIAGNOSTICS,    /* toggle */
    RFM_NAVIGATE_DESKTOP,       /* toggle */
    RFM_DESKTOP_TOP_MARGIN,     /* spin button*/
    RFM_DESKTOP_BOTTOM_MARGIN,  /* spin button*/
    RFM_DESKTOP_RIGHT_MARGIN,   /* spin button*/
    RFM_DESKTOP_LEFT_MARGIN,    /* spin button*/
    RFM_DESKTOP_DIR,            /* path selector */
    RFM_DESKTOP_IMAGE,          /* path selector */
    RFM_DESKTOP_COLOR,          /* color selector desktop */
    RFM_ICONVIEW_COLOR,         /* color selector iconview */
    RFM_TRANSPARENCY,           /* opacity selector, iconview */

    RFM_PLUGIN_FLAGS,                   /*  */
    RFM_MODULE_FLAGS,                   /*  */
    RFM_CP_FLAGS,                   /*  */
    RFM_MV_FLAGS,                   /*  */
    RFM_LN_FLAGS,                   /*  */
    RFM_RM_FLAGS,                   /*  */
    RFM_SHRED_FLAGS,                   /*  */

    RFM_CP_backup,                   /*  */
    RFM_CP_preserve,                   /*  */
    RFM_CP_no_preserve,                   /*  */
    RFM_CP_reflink,                   /*  */
    RFM_CP_sparse,                   /*  */
    RFM_CP_suffix,                   /*  */

    RFM_MV_backup,                   /*  */
    RFM_MV_suffix,                   /*  */
    
    RFM_LN_backup,                   /*  */
    RFM_LN_suffix,                   /*  */

    RFM_RM_interactive,                   /*  */

    RFM_SHRED_iterations,                   /*  */
    RFM_SHRED_size,                   /*  */

// This should remain:
#ifdef GNU_CP
    VERSION_CONTROL,
#endif

    SMB_USER,                   /* input text */
    SUDO_ASKPASS,
    SSH_ASKPASS,
    RFM_TOOLBAR,
    RFM_PASTEBOARD_SERIAL,
    RFM_BOOKMARK_SERIAL,
    /* what's this for? */
    PWD,                        //XXX
    RFM_MONITOR_ICON_THEME,
    RFM_OPTIONS
};
/* increment version number for MCS_SHM_PLUGIN_FILE on configuration changes
 * (i.e. adding or removing configuration environment variables) */
// MCS_SHM_PLUGIN_FILE depends on CONFIG_BUILD, so that CONFIG_BUILD id 
// should be incremented with each modification to rfm environment variables.
// For rodent Gamma: build is 5190 (see files.h)
// For rodent Delta-unstable: build is 5858 (see files.h)
// For rodent Core-unstable: build is 6203 (see files.h)
// For rodent Core-rc1: build is 6214a (see files.h)

# ifdef RFM_PRIMARY_C
rfm_global_t *rfm_global_p = NULL;
/* the size of env_string must coincide with that of env_vars! 
*/
environ_t environ_v[RFM_OPTIONS + 1] = {        
    // +1 because zero does not count in enum...
    /* general:: */
    { "RFM_DOUBLE_CLICK_NAVIGATION", NULL, N_("Activate items with a double click")}, 
    { "RFM_USE_GTK_ICON_THEME", "", N_("Icon Theme Specification")}, 
    { "RFM_DRAG_DOES_MOVE", "Yes", N_("Drag: move")}, 
    { "RFM_CONTENT_FOLDER_ICONS", N_("Yes"), N_("Emblems")},
    { "RFM_ENABLE_TIPS", N_("Yes"), N_("Enable tooltips")},
    { "RFM_ENABLE_LABEL_TIPS", N_("Yes"), N_("Enable tooltips")},
    { "RFM_DIAGNOSTICS_FONT_SIZE", "9", N_("Text: Font Size")},
    { "RFM_DEFAULT_ICON_SIZE", "Normal", N_("Default Size")},
    { "TERMINAL_CMD", "roxterm", N_("Terminal Emulator")}, 
    { "EDITOR", "gvim -f", N_("Text Editor")},
    { "RFM_MAXIMUM_COMPLETION_OPTIONS", "104", N_("Maximum completion options displayed")},
    { "RFM_LOAD_TIMEOUT", "5", N_("Maximum time (seconds) to wait for a load directory")},
    { "RFM_MAXIMUM_DIAGNOSTIC_LINES", "1000", N_("Maximum lines in lp terminal buffer")},
    /* desktop:: */
    { "RFM_ENABLE_DESKTOP", NULL, N_("Show Desktop Grid")}, 
    { "RFM_ENABLE_DESKTOP_DIAGNOSTICS", N_("Yes"), N_("Console Message Viewer")},
    { "RFM_NAVIGATE_DESKTOP", NULL, N_("Navigation Window")}, 

    { "RFM_DESKTOP_TOP_MARGIN", "20", N_("Top Margin")},
    { "RFM_DESKTOP_BOTTOM_MARGIN", "40", N_("Bottom Margin")},
    { "RFM_DESKTOP_RIGHT_MARGIN", "25", N_("Right margin")},
    { "RFM_DESKTOP_LEFT_MARGIN", "50", N_("Left Margin")},
   
    { "RFM_DESKTOP_DIR", NULL, N_("Desktop path:")},
    { "RFM_DESKTOP_IMAGE", PREFIX"/share/images/juchi-2007.jpg", N_("Background image")}, 
    { "RFM_DESKTOP_COLOR", "#4C1E0C", N_("Background color")}, 
    { "RFM_ICONVIEW_COLOR", "#383C3F", N_("Background color")},
    { "RFM_TRANSPARENCY", NULL, N_("Background transparency:")},

    { "RFM_PLUGIN_FLAGS", "0x0", "plugin bitflags (lite)"}, 
    { "RFM_MODULE_FLAGS", "0x7f", "module bitflags (lite)"}, 
    ///// cp //////
    { "RFM_CP_FLAGS", "0x100460c8", "cp bitflags"}, 
    { "RFM_MV_FLAGS", "0x40a", "mv bitflags"}, 
    { "RFM_LN_FLAGS", "0x1112", "ln bitflags"}, 
    { "RFM_RM_FLAGS", "0x3aa", "rm bitflags"}, 
    { "RFM_SHRED_FLAGS", "0x162", "shred bitflags"}, 

    { "RFM_CP_backup", "existing", "shred bitflags"}, 
    { "RFM_CP_preserve", "all", "shred bitflags"}, 
    { "RFM_CP_no_preserve", "all", "shred bitflags"}, 
    { "RFM_CP_reflink", "auto", "shred bitflags"}, 
    { "RFM_CP_sparse", "auto", "shred bitflags"}, 
    { "RFM_CP_suffix", "~", "shred bitflags"}, 


    { "RFM_MV_backup", "existing", "shred bitflags"}, 
    { "RFM_MV_suffix", "~", "shred bitflags"}, 

    { "RFM_LN_backup", "existing", "shred bitflags"}, 
    { "RFM_LN_suffix", "~", "shred bitflags"}, 

    { "RFM_RM_interactive", "always", "shred bitflags"}, 

    { "RFM_SHRED_iterations", "3", "shred bitflags"}, 
    { "RFM_SHRED_size", "100K", "shred bitflags"}, 

    
#ifdef GNU_CP
    { "VERSION_CONTROL", "existing", "VERSION_CONTROL="},
#endif
    {"SMB_USER", NULL, N_("Samba default remote user")}, 
    // XXX (SMB_USER inactive yet)
    {"SUDO_ASKPASS", PREFIX"/bin/rodent-getpass", N_("Sudo ask password program")}, 
    {"SSH_ASKPASS", PREFIX"/bin/rodent-getpass", N_("Ssh ask passphrase program")}, 
    // XXX (inactive yet)
    {"RFM_TOOLBAR", NULL, N_("Toolbar configuration")},

    {"RFM_PASTEBOARD_SERIAL", "0", N_("Pasteboard serial control")},
    // XXX needs translation fix... (inactive yet)
    {"RFM_BOOKMARK_SERIAL", "0", N_("Bookmark serial control")},
    /* Third party environment */
    { "PWD", NULL, "third party stuff"}, 
    /* Extra stuff, not in settings dialog checkboxes */
    {"RFM_MONITOR_ICON_THEME", "yes", N_("Monitor third party gtk icon theme")},

    { NULL, NULL, NULL}

};



const gchar *font_sizes_v[] = {
    "8",
    "9",
    "10",
    "11",
    "12",
    "14",
    NULL
};

const gchar *icon_sizes_v[] = {
    "Normal",
    "Details",
    "Compact",
    "Big",
    "Huge",
    NULL
};

const gchar *terminals_v[] = {
	"roxterm", 
	"sakura",
	"gnome-terminal", 
	"Eterm", 
	"konsole", 
	"Terminal", 
	"aterm", 
	"xterm", 
	"kterm", 
	"wterm", 
	"multi-aterm", 
	"evilvte",
	"mlterm",
	"xvt",
	"rxvt",
	"urxvt",
	"mrxvt",
	"tilda",
	NULL
};

const gchar *editors_v[] = {
	"gvim -f",  
	"mousepad", 
	"gedit", 
	"kate", 
	"xemacs", 
	"nano",
	"vi",
	"xterm -e nano",
	"xterm -e vi",
	NULL
};

button_callback_t button_callback_v[]={
    {NEW_FILE_CALLBACK_BUTTON_ID, "NEW_FILE_CALLBACK_button", NEW_FILE_CALLBACK, NULL},
    {NEWDIR_CALLBACK_BUTTON_ID, "NEWDIR_CALLBACK_button", NEWDIR_CALLBACK, NULL},
    {NEW_TAB_CALLBACK_BUTTON_ID, "NEW_TAB_CALLBACK_button", NEW_TAB_CALLBACK, NULL},  
    {NEW_WINDOW_CALLBACK_BUTTON_ID, "NEW_WINDOW_CALLBACK_button", NEW_WINDOW_CALLBACK, NULL},
    {ABOUT_CALLBACK_BUTTON_ID, "ABOUT_CALLBACK_button", ABOUT_CALLBACK, NULL},
    {JUMP_TO_CALLBACK_BUTTON_ID, "JUMP_TO_CALLBACK_button", JUMP_TO_CALLBACK, NULL},
    {GOTO_HOST_CALLBACK_BUTTON_ID, "GOTO_HOST_CALLBACK_button", GOTO_HOST_CALLBACK, NULL},
    {GOTO_HOME_CALLBACK_BUTTON_ID, "GOTO_HOME_CALLBACK_button", GOTO_HOME_CALLBACK, NULL},
    {GO_UP_CALLBACK_BUTTON_ID, "GO_UP_CALLBACK_button", GO_UP_CALLBACK, NULL},
    {GO_BACK_CALLBACK_BUTTON_ID, "GO_BACK_CALLBACK_button", GO_BACK_CALLBACK, NULL},
    {GO_FORWARD_CALLBACK_BUTTON_ID, "GO_FORWARD_CALLBACK_button", GO_FORWARD_CALLBACK, NULL},
    {CUT_CALLBACK_BUTTON_ID, "CUT_CALLBACK_button", CUT_CALLBACK, NULL},
    {COPY_CALLBACK_BUTTON_ID, "COPY_CALLBACK_button", COPY_CALLBACK, NULL},
    {PASTE_CALLBACK_BUTTON_ID, "PASTE_CALLBACK_button", PASTE_CALLBACK, NULL},
    {REMOVE_CALLBACK_BUTTON_ID, "REMOVE_CALLBACK_button", REMOVE_CALLBACK, NULL},
    {TOGGLE_BOOKMARK_CALLBACK_BUTTON_ID, "TOGGLE_BOOKMARK_CALLBACK_button", TOGGLE_BOOKMARK_CALLBACK, NULL},
    {HELP_CALLBACK_BUTTON_ID, "HELP_CALLBACK_button", HELP_CALLBACK, NULL},
    {SETTINGS_CALLBACK_BUTTON_ID, "SETTINGS_CALLBACK_button", SETTINGS_CALLBACK, NULL},
    {TERMINAL_CALLBACK_BUTTON_ID, "TERMINAL_CALLBACK_button", TERMINAL_CALLBACK, NULL},
    {REFRESH_CALLBACK_BUTTON_ID, "REFRESH_CALLBACK_button", REFRESH_CALLBACK, NULL},
    {RUN_CALLBACK_BUTTON_ID, "RUN_CALLBACK_button", RUN_CALLBACK, NULL},
    {FIND_CALLBACK_BUTTON_ID, "FIND_CALLBACK_button", FIND_CALLBACK, NULL},
    {DIFFERENCES_CALLBACK_BUTTON_ID, "DIFFERENCES_CALLBACK_button", DIFFERENCES_CALLBACK, NULL},
    {SORT_ASCENDING_CALLBACK_BUTTON_ID, "Ascending_button", SORT_ASCENDING_CALLBACK, NULL},
    {SORT_DESCENDING_CALLBACK_BUTTON_ID, "Descending_button", SORT_DESCENDING_CALLBACK, NULL},
    {KEYBINDINGS_CALLBACK_BUTTON_ID, "KEYBINDINGS_CALLBACK_button", KEYBINDINGS_CALLBACK, NULL},
    {SELECT_ALL_CALLBACK_BUTTON_ID, "SELECT_ALL_CALLBACK_button", SELECT_ALL_CALLBACK, NULL},
    {SELECT_INVERT_CALLBACK_BUTTON_ID, "SELECT_INVERT_CALLBACK_button", SELECT_INVERT_CALLBACK, NULL},
    {UNSELECT_ALL_CALLBACK_BUTTON_ID, "UNSELECT_ALL_CALLBACK_button", UNSELECT_ALL_CALLBACK, NULL},
    {SELECT_BYFILTER_CALLBACK_BUTTON_ID, "SELECT_BYFILTER_CALLBACK_button", SELECT_BYFILTER_CALLBACK, NULL},
    {UNSELECT_BYFILTER_CALLBACK_BUTTON_ID, "UNSELECT_BYFILTER_CALLBACK_button", UNSELECT_BYFILTER_CALLBACK, NULL},
    {TYPE_SORT_CALLBACK_BUTTON_ID, "Type_button", TYPE_SORT_CALLBACK,  N_("Type")},
    {NAME_SORT_CALLBACK_BUTTON_ID, "Name_button", NAME_SORT_CALLBACK, N_("Name")},
    {DATE_SORT_CALLBACK_BUTTON_ID, "Date_button", DATE_SORT_CALLBACK, N_("Date")},
    {SIZE_SORT_CALLBACK_BUTTON_ID, "Size_button", SIZE_SORT_CALLBACK, N_("Size")},
    {OWNER_SORT_CALLBACK_BUTTON_ID, "Owner_button", OWNER_SORT_CALLBACK, N_("Owner")},
    {GROUP_SORT_CALLBACK_BUTTON_ID, "Group_button", GROUP_SORT_CALLBACK, N_("Group")},
    {MODE_SORT_CALLBACK_BUTTON_ID, "Mode_button", MODE_SORT_CALLBACK, N_("Mode")},
    {CASE_SORT_CALLBACK_BUTTON_ID, "case_insensitive_button", CASE_SORT_CALLBACK, NULL},
    {DEFAULT_ICONSIZE_CALLBACK_BUTTON_ID, "DEFAULT_ICONSIZE_CALLBACK_button", DEFAULT_ICONSIZE_CALLBACK, NULL},
    {DEFAULT_ICONSIZE_ALL_CALLBACK_BUTTON_ID, "DEFAULT_ICONSIZE_ALL_CALLBACK_button", DEFAULT_ICONSIZE_ALL_CALLBACK, NULL},
    {COMPACT_ICONSIZE_CALLBACK_BUTTON_ID, "List_button", COMPACT_ICONSIZE_CALLBACK, N_("List")},
    {TINY_ICONSIZE_CALLBACK_BUTTON_ID, "Tiny_button", TINY_ICONSIZE_CALLBACK, N_("Tiny")},
    {NORMAL_ICONSIZE_CALLBACK_BUTTON_ID, "Normal_button", NORMAL_ICONSIZE_CALLBACK, N_("Normal")},
    {BIG_ICONSIZE_CALLBACK_BUTTON_ID, "Big_button", BIG_ICONSIZE_CALLBACK, N_("Big")},
    {HUGE_ICONSIZE_CALLBACK_BUTTON_ID, "Huge_button", HUGE_ICONSIZE_CALLBACK, N_("Huge")},
    {PLUS_ICONSIZE_CALLBACK_BUTTON_ID, "PLUS_ICONSIZE_CALLBACK_button", PLUS_ICONSIZE_CALLBACK, NULL},
    {MINUS_ICONSIZE_CALLBACK_BUTTON_ID, "MINUS_ICONSIZE_CALLBACK_button", MINUS_ICONSIZE_CALLBACK, NULL},
    {SHOW_HIDDEN_CALLBACK_BUTTON_ID, "show_hidden_button", SHOW_HIDDEN_CALLBACK, NULL},
    {PREVIEW_IMAGES_CALLBACK_BUTTON_ID, "show_previews_button", PREVIEW_IMAGES_CALLBACK, NULL},
    {NEW_TAB_CALLBACK_B_BUTTON_ID, "NEW_TAB_CALLBACK_B_button", NEW_TAB_CALLBACK_B, NULL}, // last one < 64 
    {-1, NULL, NULL_CALLBACK, NULL}
};
#ifdef GNU_CP
#define ENABLE_GNU_CP TRUE
#else
#define ENABLE_GNU_CP FALSE
#endif
gchar *cp_v_control[] = {"existing","none","numbered","simple",NULL};
gchar *cp_attributes[] = {"all","mode","ownership","timestamps","context","links","xattr",NULL};
gchar *cp_when[] = {"always","auto","never",NULL};
gchar *cp_suffix[] = {"~",".bak",NULL};
program_options_t cp_options[]={
    {"",N_("Ask the user to get additional parameters"),FALSE,NULL},
    {"-a","-a, --archive",TRUE,NULL},
    {"--attributes-only",NULL,ENABLE_GNU_CP,NULL},
    {"-b",NULL,ENABLE_GNU_CP,NULL},
    {"--backup=",NULL,ENABLE_GNU_CP,cp_v_control,RFM_CP_backup},
    {"--copy-contents","--copy-contents",ENABLE_GNU_CP,NULL},
    {"-d",NULL,ENABLE_GNU_CP,NULL},
    {"-f","-f, --force",TRUE,NULL},
    {"-i","-i, --interactive",ENABLE_GNU_CP,NULL},
    {"-H",NULL,TRUE,NULL},
    {"-l","-l, --link",TRUE,NULL},
    {"-L","-L, --dereference",TRUE,NULL},
    {"-n","-n, --no-clobber",TRUE,NULL},
    {"-P","-P, --no-dereference",TRUE,NULL},
    {"-p",NULL,TRUE,NULL},
    {"--preserve=",NULL,ENABLE_GNU_CP,cp_attributes,RFM_CP_preserve},
    {"--no-preserve=",NULL,ENABLE_GNU_CP,cp_attributes,RFM_CP_no_preserve},
    {"--parents",NULL,ENABLE_GNU_CP,NULL},
    {"-R","-R, -r, --recursive",TRUE,NULL},
    {"--reflink=",NULL,ENABLE_GNU_CP,cp_when,RFM_CP_reflink},
    {"--remove-destination",NULL,ENABLE_GNU_CP,NULL},
    {"--sparse=",NULL,ENABLE_GNU_CP,cp_when,RFM_CP_sparse},
    {"--strip-trailing-slashes",NULL,ENABLE_GNU_CP,NULL},
    {"-s","-s, --symbolic-link",ENABLE_GNU_CP,NULL},
    {"--suffix=","-S, --suffix=",ENABLE_GNU_CP,cp_suffix, RFM_CP_suffix},
    {"--target-directory=",NULL,FALSE,NULL},
    {"-T"," -T, --no-target-directory",ENABLE_GNU_CP,NULL},
    {"-u","-u, --update",ENABLE_GNU_CP,NULL},
    {"-v","-v, --verbose",ENABLE_GNU_CP,NULL},
    {"-x","-x, --one-file-system",ENABLE_GNU_CP,NULL},
    {"--help",NULL,FALSE,NULL},
    {"--version",NULL,FALSE,NULL},
    {NULL,NULL,FALSE,NULL}
};
program_options_t mv_options[]={
    {"",N_("Ask the user to get additional parameters"),FALSE,NULL},
    {"-b",NULL,ENABLE_GNU_CP,NULL},
    {"--backup=",NULL,ENABLE_GNU_CP,cp_v_control,RFM_MV_backup},
    {"-f","-f, --force",TRUE,NULL},
    {"-i","-i, --interactive",ENABLE_GNU_CP,NULL},
    {"--strip-trailing-slashes",NULL,ENABLE_GNU_CP,NULL},
    {"--suffix=","-S, --suffix=",ENABLE_GNU_CP,cp_suffix,RFM_MV_suffix},
    {"--target-directory=",NULL,FALSE,NULL},
    {"-T"," -T, --no-target-directory",ENABLE_GNU_CP,NULL},
    {"-u","-u, --update",ENABLE_GNU_CP,NULL},
    {"-v","-v, --verbose",ENABLE_GNU_CP,NULL},
    {"--help",NULL,FALSE,NULL},
    {"--version",NULL,ENABLE_GNU_CP,NULL},
    {NULL,NULL,FALSE,NULL}
};
program_options_t ln_options[]={
    {"",N_("Ask the user to get additional parameters"),FALSE,NULL},
    {"-b",NULL,ENABLE_GNU_CP,NULL},
    {"--backup=",NULL,ENABLE_GNU_CP,cp_v_control,RFM_LN_backup},
    {"-d","-d, -F, --directory",ENABLE_GNU_CP,NULL},
    {"-f","-f, --force",TRUE,NULL},
    {"-L","-L, --logical",TRUE,NULL},
    {"-n","-n, --no-dereference",TRUE,NULL},
    {"-P","-P, --physical",TRUE,NULL},
    {"-s","-s, --symbolic-link",ENABLE_GNU_CP,NULL},
    {"--suffix=","-S, --suffix=",ENABLE_GNU_CP,cp_suffix,RFM_LN_suffix},
    {"--target-directory=",NULL,FALSE,NULL},
    {"-T"," -T, --no-target-directory",ENABLE_GNU_CP,NULL},
    {"-v","-v, --verbose",ENABLE_GNU_CP,NULL},
    {"--help",NULL,FALSE,NULL},
    {"--version",NULL,ENABLE_GNU_CP,NULL},
    {NULL,NULL,FALSE,NULL}
};
gchar *rm_interactive[]={"never","once","always",NULL};
program_options_t rm_options[]={
    {"",N_("Ask the user to get additional parameters"),FALSE,NULL},
    {"-f","-f, --force",TRUE,NULL},
    {"-i",NULL,ENABLE_GNU_CP,NULL},
    {"-I",NULL,ENABLE_GNU_CP,NULL},
    {"--interactive=",NULL,ENABLE_GNU_CP,rm_interactive,RFM_RM_interactive},
    {"--one-file-system",NULL,ENABLE_GNU_CP,NULL},
    {"--no-preserve-root",NULL,ENABLE_GNU_CP,NULL},
    {"--preserve-root",NULL,ENABLE_GNU_CP,NULL},
    {"-R","-R, -r, --recursive",TRUE,NULL},
    {"-v","-v, --verbose",ENABLE_GNU_CP,NULL},
    {"--help",NULL,FALSE,NULL},
    {"--version",NULL,ENABLE_GNU_CP,NULL},
    {NULL,NULL,FALSE,NULL}
};
gchar *shred_iterations[]={"1","3","5","10","20",NULL};
gchar *shred_size[]={"10K","100K","1M","10M","100M","1G",NULL};
program_options_t shred_options[]={
    {"",N_("Ask the user to get additional parameters"),FALSE,NULL},
    {"-f","-f, --force",TRUE,NULL},
    // this could use a spin button
    {"--iterations=","-n, --iterations=",TRUE,shred_iterations,RFM_SHRED_iterations},
    {"--random-source=",NULL,FALSE,NULL},// this needs a fileselector button.
    {"--size=","-s, --size=",ENABLE_GNU_CP,shred_size,RFM_SHRED_size},// this could use a spinbutton

    {"-u","-u, --remove",ENABLE_GNU_CP,NULL},
    {"-v","-v, --verbose",ENABLE_GNU_CP,NULL},
    {"-x","-x, --exact",ENABLE_GNU_CP,NULL},
    {"-z","-z, --zero",ENABLE_GNU_CP,NULL},
    {"--help",NULL,FALSE,NULL},
    {"--version",NULL,ENABLE_GNU_CP,NULL},
    {NULL,NULL,FALSE,NULL}
};

program_options_t lite_plugin_options[]={
    {"fstab",N_("Mount local disks and devices"),TRUE,NULL},
    {"ps",N_("View current processes and monitor system state"),TRUE,NULL},
    {"dotdesktop",N_("Next-generation application launcher."),TRUE,NULL},

    {"submodule-label",N_("Windows networks (SMB)"),TRUE,NULL},
    {"smb",N_("SMB Browser"),TRUE,NULL},
    {"submodule-indent",NULL,TRUE,NULL},
	{"workgroup",N_("Windows workgroup"),TRUE,NULL},
	{"shares",N_("Windows Shares"),TRUE,NULL},
    {"submodule-unindent",NULL,TRUE,NULL},

    {"submodule-label",N_("FUSE Volume"),TRUE,NULL},
    {"fuse",N_("Mount user-space filesystems (FUSE)"),TRUE,NULL},
    {"submodule-indent",NULL,TRUE,NULL},
	{"nfs",N_("NFS Network Volume"),TRUE,NULL},
	{"sftp",N_("Secure FTP (SSH)"),TRUE,NULL},
	{"ecryptfs",N_("Encrypted filesystem"),TRUE,NULL},
	{"obex",N_("Bluetooth Transfer"),TRUE,NULL},
	{"ftp",N_("FTP Client"),TRUE,NULL},
	{"cifs",N_("CIFS Volume"),TRUE,NULL},
    {"submodule-unindent",NULL,TRUE,NULL},
    {NULL,NULL,FALSE,NULL}
};
program_options_t lite_module_options[]={
    {"callbacks",N_("Callbacks"),FALSE,NULL},
    {"settings",N_("Settings"),FALSE,NULL},
    {"run",N_("Run program and return its output"),FALSE,NULL},
    {"properties",N_("Properties"),TRUE,NULL},
    {"completion",N_("Text completion"),TRUE,NULL},
    {"combobox",N_("History of combo url."),TRUE,NULL},
    {"mime",N_("Mime Type"),FALSE,NULL},
    {"mimemagic",N_("Use MIME type magic"),TRUE,NULL},
    {"mimezip",N_("ZIP archive plugin"),TRUE,NULL},
    {"icons",N_("Icon Themes"),TRUE,NULL},
     {NULL,NULL,FALSE,NULL}
};

# else
extern rfm_global_t *rfm_global_p;
extern environ_t environ_v[];
extern const gchar *terminals_v[];
extern const gchar *editors_v[];
extern const gchar *icon_sizes_v[];
extern const gchar *font_sizes_v[];
extern button_callback_t button_callback_v[];
extern program_options_t cp_options[];
extern program_options_t mv_options[];
extern program_options_t ln_options[];
extern program_options_t rm_options[];
extern program_options_t shred_options[];
extern program_options_t lite_plugin_options[];
extern program_options_t lite_module_options[];
extern gchar *cp_v_control[];
extern gchar *cp_attributes[];
extern gchar *cp_when[];
extern gchar *cp_suffix[];
extern gchar *rm_interactive[];
extern gchar *shred_iterations[];
extern gchar *shred_size[];
# endif

