/*
 * Copyright (C) 2002-2012 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rfm_libs.h"
#include "rfm_modules.h"

#include "primary-icons.i"


gboolean
rfm_is_dark_background(view_t *view_p){
    gboolean is_dark = FALSE;
    // Get background color
    GdkColor gdk_color;
    const gchar *bkg_var=(view_p->flags.type == DESKVIEW_TYPE)?
	"RFM_DESKTOP_COLOR": "RFM_ICONVIEW_COLOR";
    // figure out if light or dark background to use
    // dark or light color set
    if(getenv (bkg_var) && strlen (getenv (bkg_var))) {
	if (!gdk_color_parse (getenv (bkg_var), &gdk_color)){
	    g_warning("cannot parse background color %s", getenv (bkg_var));
	} else {
	    gint sum = (gdk_color.red) +
		      (gdk_color.green) +
		      (gdk_color.blue);
	    is_dark = (sum < 0xffff * 3 / 2);
	}
    } 
    return is_dark;
}

#define NUM_COLORS 8
GdkColor *
rfm_get_gdk_color (view_t *view_p, int p) {
    if(p < 0 || p >= NUM_COLORS) {
        g_warning ("rodent_select_pen: pen %d out of range.", p);
        return NULL;
    }
    GdkColor pen_colors[] = {
/* light background */
        {0, 65535, 65535, 65535},       /*white */
        {1, 0, 0, 0},           /*black */
        {2, 65535, 0, 0},       /*red */
        {3, 0, 30000, 0},       /*green */
        {4, 0, 0, 32000},       /*blue */
        {5, 32000, 32000, 0},   /*brown */
        {6, 65535, 0, 65535},   /*magenta */
        {7, 48000, 48000, 48000},       /*grey */
/* dark background */
        {1, 0, 0, 0},           /*black */
        {0, 65535, 65535, 65535},       /*white */
        {12, 65535, 20000, 20000},      /*red */
        {13, 20000, 60000, 20000},      /*green */
        {14, 40000, 40000, 65535},      /*blue */
        {15, 32000, 32000, 0},  /*brown */
        {16, 65535, 0, 65535},  /*magenta */
        {17, 48000, 48000, 48000}       /*grey */
    };

    GdkColor *gdk_color = (GdkColor *) malloc(sizeof(GdkColor));
    memset(gdk_color, 0, sizeof(GdkColor));
        
    // if background color requested, return now 
    const gchar *bkg_var=(view_p->flags.type == DESKVIEW_TYPE)?
	"RFM_DESKTOP_COLOR": "RFM_ICONVIEW_COLOR";
    if (p == BACKGROUND_COLOR && getenv (bkg_var) && strlen (getenv (bkg_var))) {
	if (!gdk_color_parse (getenv (bkg_var), gdk_color)){
	    g_warning("cannot parse background color %s", getenv (bkg_var));
	} else {
	    NOOP ("DESK: selecting BACKGROUND_COLOR %s\n", getenv (bkg_var));
	    return gdk_color;
	}
    }

    gint offset = (rfm_is_dark_background(view_p))? NUM_COLORS : 0;
    // return color entry
    memcpy(gdk_color,  pen_colors + p + offset, sizeof(GdkColor));
    return gdk_color;
}


typedef struct mime_winner_thread_t {
    gchar *path;
    struct stat st;
    gchar *mimetype;
} mime_winner_thread_t;

typedef struct  mime_winner_dbh_t {
    off_t st_sum;
    gchar mimetype[80];
} mime_winner_dbh_t;

static off_t
st_sum (struct stat *st) {
    off_t sum;
    gint build=atoi(BUILD);
    if(!st){
        sum = 0;
    } else {
        sum = st->st_mtime + st->st_size + st->st_mode + st->st_nlink + st->st_uid + st->st_gid + build;
    }
    return sum;
}

static
gchar *
get_mime_winner_from_cache(const gchar *path, struct stat *st){
    // return NULL; // stress test
    DBHashTable *winner_cache;
    gchar *mimetype=NULL;
    gchar *g = g_build_filename (MIME_COUNT_DBH_FILE, NULL);
    // Here we use an external mutex instead of DBH_THREAD_SAFE
    g_mutex_lock(winner_mutex);
    if((winner_cache = dbh_new (g, NULL, DBH_READ_ONLY|DBH_PARALLEL_SAFE)) != NULL) {
	dbh_set_parallel_lock_attempt_limit(winner_cache, 50);
	GString *gs = g_string_new (path);
	sprintf ((char *)DBH_KEY (winner_cache), "%10u", g_string_hash (gs));
	NOOP(stderr, "%s -> %10u\n", path, g_string_hash (gs));
	g_string_free (gs, TRUE);
	if (dbh_load (winner_cache) != 0){
	    mime_winner_dbh_t *mime_winner_dbh_p =
		(mime_winner_dbh_t *)winner_cache->data;
	    if (st_sum(st) == mime_winner_dbh_p->st_sum) {
		mimetype=g_strdup(mime_winner_dbh_p->mimetype);
	    } else {
		NOOP(stderr, "st mismatch for %s %d != %d\n",
			path, (int)st_sum(st), (int)mime_winner_dbh_p->st_sum);
	    }
	} else {
	    NOOP(stderr, "Could not load record for %s\n", path);
	}
	dbh_close (winner_cache);
    } else {
	NOOP(stderr, "Cannot open %s\n", g);
    }
    g_mutex_unlock(winner_mutex);
    g_free(g);
    /*if (mimetype && strcmp(mimetype, "unknown")==0) {
	g_free(mimetype);
	mimetype = NULL;
    }*/
    return mimetype;
}
static
gpointer
save_mime_winner_to_cache(gpointer data){
    mime_winner_thread_t *mime_winner_thread_p=data;

    // save to dbh
    DBHashTable *winner_cache;
    gchar *cache_dir = g_build_filename (RFM_CACHE_DIR, NULL);
    if(g_mkdir_with_parents (cache_dir, 0700) < 0) {
        g_free (cache_dir);
	NOOP(stderr, "g_mkdir_with_parents\n");
        return NULL;
    }
    g_free (cache_dir);

    gchar *g = g_build_filename (MIME_COUNT_DBH_FILE, NULL);
    g_mutex_lock(winner_mutex);
    if((winner_cache = dbh_new (g, NULL, DBH_PARALLEL_SAFE)) == NULL) {
	unsigned char keylength = 11;
	winner_cache = dbh_new (g, &keylength, DBH_PARALLEL_SAFE|DBH_CREATE);
    }
    dbh_set_parallel_lock_attempt_limit(winner_cache, 50);
    NOOP(stderr, "saving cache value for %s: %s\n",
	    mime_winner_thread_p->path, mime_winner_thread_p->mimetype);
    if(winner_cache == NULL) {
	    g_warning("could not create %s", g);
    } else {
	GString *gs = g_string_new (mime_winner_thread_p->path);
	memset(DBH_KEY(winner_cache), 0, 11);
	sprintf ((char *)DBH_KEY (winner_cache), "%10u", g_string_hash (gs));
	NOOP(stderr, "S: %s -> %10u\n", mime_winner_thread_p->path, g_string_hash (gs));
	g_string_free (gs, TRUE);
	dbh_set_recordsize(winner_cache, sizeof(mime_winner_dbh_t));
	mime_winner_dbh_t *mime_winner_dbh_p =
	    (mime_winner_dbh_t *)winner_cache->data;
	memset(mime_winner_dbh_p->mimetype, 0, 80);
	strncpy(mime_winner_dbh_p->mimetype, mime_winner_thread_p->mimetype, 79);
	mime_winner_dbh_p->st_sum = st_sum(&(mime_winner_thread_p->st));
	if (dbh_update (winner_cache) == 0){
	    g_warning("could not update %s", g);
	}
	NOOP(stderr, "saved cache value for %s at %s\n", 
		mime_winner_thread_p->path,g );
	dbh_close (winner_cache);
    }
    g_mutex_unlock(winner_mutex);
    g_free(g);

// cleanup
    g_free(mime_winner_thread_p->path);
    g_free(mime_winner_thread_p->mimetype);
    g_free(mime_winner_thread_p);
    return NULL;
}

static
gchar *
get_mayor_mime_type(const gchar *path){
    NOOP(stderr, "get_mayor_mime_type(%s)\n",path);
    struct stat st;
    if (stat(path, &st) < 0){
	return NULL;
    }
#if 10
    gchar *cache_value=get_mime_winner_from_cache(path, &st);
    if (cache_value) {
	NOOP(stderr, "cache value! %s got mimetype: %s\n", path, cache_value);
	return cache_value;
    } else {
	NOOP(stderr, "no cache value for %s\n", path);
    }
#endif

    // If cache out of date or empty, get the new value

    DIR *directory;
    directory = opendir (path);
    NOOP(stderr, "path=%s directory=0x%x\n", path,GPOINTER_TO_INT(directory));
    if (!directory){
	return NULL;
    }
    // This is a stack hash table:
    // thus, no need to mutex protect.
    GHashTable *type_hash=g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
    struct dirent *d;
    // count first
    gint total = 0;
    while((d = readdir (directory)) != NULL) {
	total++;
	if (total > 200) break;
    }
    closedir (directory);
    if (total > 200){
	return g_strdup("xffm/stock_dnd-multiple");
    }


    directory = opendir (path);

    //
    while((d = readdir (directory)) != NULL) {
        if(*(d->d_name)=='.') continue;
        gchar *fullpath = g_strdup_printf ("%s/%s", path, d->d_name);
	gchar *hash_key=MIME_type(fullpath,NULL);
	// Don't do mime magic so that hash key may be NULL
	// and do things faster (this is a ballpark figure anyways)
	
	if (!hash_key) {
	    hash_key=MIME_magic(fullpath);
	    if (!hash_key) continue;
	}
	
	void *p = g_hash_table_lookup (type_hash, hash_key);
	gint count;
	if (p==NULL) {
	    count=1;
	    g_hash_table_insert (type_hash, hash_key, GINT_TO_POINTER(count));
	} else {
	    count=GPOINTER_TO_INT(p)+1;
	    g_hash_table_replace (type_hash, hash_key, GINT_TO_POINTER(count));
	}
	NOOP(stderr, "get_mayor_mime_type: %s (%s) %s==%d\n",
		path,d->d_name,hash_key,count);

	g_free(fullpath);
    }
    closedir (directory);

    GList *tmp;
    GList *hash_keys = g_hash_table_get_keys (type_hash);
    gint max=0;
    gchar *winner=NULL;
    for (tmp=hash_keys; tmp && tmp->data; tmp=tmp->next){
	gint value=GPOINTER_TO_INT(g_hash_table_lookup (type_hash, (gchar *)tmp->data));
	NOOP(stderr, "hashkey=%s value=%d\n", (gchar *)tmp->data, value);
	if (value > max) {
	    max=value;
	    winner = tmp->data;
	}
    }
    g_list_free(hash_keys);
    if (!winner) return NULL;
    if (strstr(winner, "No Read Permission")){
	return NULL;
    } 
    winner=g_strdup(winner);

    g_hash_table_remove_all (type_hash);
 
    g_hash_table_destroy(type_hash);
    NOOP(stderr, "%s ---> %s\n", path, winner); 

    //
    mime_winner_thread_t *mime_winner_thread_p=
	(mime_winner_thread_t *)malloc(sizeof(mime_winner_thread_t));
    if (!mime_winner_thread_p){
	g_error("cannot allocate mime_winner_thread_p\n");
    }


    mime_winner_thread_p->mimetype = g_strdup(winner);
    mime_winner_thread_p->path = g_strdup(path);
    memcpy(&(mime_winner_thread_p->st), &st, sizeof(struct stat));
    rfm_view_thread_create(NULL, save_mime_winner_to_cache, mime_winner_thread_p, "save_mime_winner_to_cache");
    if (strcmp(winner, "unknown") == 0) return NULL;
    return winner;
}

gpointer
rfm_preview_thread_f (gpointer data) {
    fireup_thread_manager_t *fireup_thread_manager_p = data;
    preview_manager_t *preview_manager_p = &(fireup_thread_manager_p->preview_manager_v);
    view_t *view_p = preview_manager_p->view_p;
    g_mutex_lock(view_p->mutexes.population_serial);
    if (preview_manager_p->population_serial != view_p->flags.population_serial){
	TRACE( "rfm_preview_thread_f(): user requested abort\n");
	g_mutex_unlock(view_p->mutexes.population_serial);
	return NULL;
    }
    g_mutex_unlock(view_p->mutexes.population_serial);

    gchar *path=fireup_thread_manager_p->path;
    gint type=fireup_thread_manager_p->type;
    population_t *population_p = fireup_thread_manager_p->population_p;

    GdkPixbuf *pixbuf=NULL;
    NOOP ("read_lock by rfm_preview_thread_f (0x%lx) %s\n",
	    (long unsigned)g_thread_self (), path);
     
    gint size = ICON_SIZE(view_p); 
    gboolean do_content_icons = (
	    IS_LOCAL_TYPE(view_p->en->type) && 
	    getenv("RFM_CONTENT_FOLDER_ICONS") && 
	    strlen(getenv("RFM_CONTENT_FOLDER_ICONS")) &&
	    IS_LOCAL_TYPE(type) 
	    );

    gchar *content_icon=NULL;
    if (do_content_icons  
	    && g_file_test(path, G_FILE_TEST_IS_DIR)){
	    gchar *inserted_icon=get_mayor_mime_type(path);
	    if (inserted_icon) {
		const gchar *icon_id=resolve_folder_icon_type (type);
		content_icon=g_strdup_printf("%s/composite2/%s", icon_id, inserted_icon);
		pixbuf = get_pixbuf (content_icon, size, TRUE);
		NOOP(stderr, "composite %s (%s)\n", content_icon,  path);
		g_free(inserted_icon);
	    }
    } else {
	// This means thumbnails will not be hashed, nor will hash be consulted...
        pixbuf = get_pixbuf (path, size, FALSE);
    }



    if (pixbuf && GDK_IS_PIXBUF (pixbuf)) {
	///// READLOCK ON //////////
	if (!rfm_population_read_lock (view_p)) {
	    NOOP(stderr, "unable to get read lock at rfm_preview_thread_f: %s\n",
		    fireup_thread_manager_p->path);
	    g_object_unref(G_OBJECT(pixbuf));
	} else {
	    population_p->pixbuf = pixbuf;
	    population_p->pixbufW = gdk_pixbuf_get_width ((const GdkPixbuf *)pixbuf);
	    population_p->pixbufH = gdk_pixbuf_get_height ((const GdkPixbuf *)pixbuf);
	    
	    if (content_icon) {
		population_p->content_icon = TRUE;
	    }

	    GdkRectangle item_rect;
	    if (rfm_get_population_rect(view_p, population_p, &item_rect)){
		// This will always be true...
		rfm_expose_rect (view_p, &item_rect);
	    }
	    // tag population_p as already previewed
	    population_p->flags |= POPULATION_THUMBNAILED;
	    rfm_population_read_unlock (view_p);
	    ///// READLOCK OFF //////////

	}
    }
    NOOP ("rfm_preview_thread_f (0x%lx) rfm_population_read_unlock!\n", 
		GPOINTER_TO_INT(g_thread_self ()));
    g_free(path);
    g_free(fireup_thread_manager_p);
    return NULL;
}

void *rodent_thread_preview_manager(void *data){
    thread_preview_manager(data);
    return NULL;
}

void
rfm_fireup_previews (view_t * view_p) {
    NOOP ("PREVIEWS: rfm_fireup_previews\n");
    if (view_p->module) return;
    // This function is entered with a write lock on population
    //
    // winner mutex is to access the mimetype winner cache
    // The mimetype winner will be displayed as an icon
    // on the top of the parent folder.
    if (!winner_mutex){
	rfm_mutex_init(winner_mutex);
    }
    if(!view_p->population_pp) {
        return;
    }
    // fireup preview threaded thread manager with private memory block
    preview_manager_t *preview_manager_p = 
	(preview_manager_t *) malloc (sizeof (preview_manager_t));
    memset (preview_manager_p, 0, sizeof (preview_manager_t));
    preview_manager_p->view_p = view_p;
	g_mutex_lock(view_p->mutexes.population_serial);
    preview_manager_p->population_serial = view_p->flags.population_serial;
	g_mutex_unlock(view_p->mutexes.population_serial);
    preview_manager_p->diagnostics = view_p->widgets.diagnostics;
    preview_manager_p->diagnostics_window = view_p->widgets.diagnostics_window;

    rfm_threadqueue_push(rfm_global_p->thread_queue, PREVIEW_MANAGER, view_p, preview_manager_p);

    //rfm_view_thread_create(view_p, thread_preview_manager, (gpointer) (preview_manager_p), "thread_preview_manager");
    return;
}


gboolean
rfm_is_image (record_entry_t * en) {
    if(!en || !en->path) return FALSE;
    if (!IS_SREG(en->type)) return FALSE;
    static GSList *pix_formats = NULL;
    GSList *l;
    gboolean result = FALSE;
    NOOP ("MIME call from  rfm_is_image\n");

    if(!en->mimetype){
	en->mimetype = MIME_type (en->path, en->st);
    }
    // Only do mime magic if extension mime type is not defined...
    // This speeds things up presto for rodent loads.
    // Furthermore, if an extension is specified, then assume that
    // magic is not wanted (why the extension otherwise?)
    gboolean want_magic = IS_LOCAL_TYPE(en->type)
	&& (!en->mimetype || strcmp(en->mimetype, _("unknown"))==0)
	&& !en->mimemagic && en->st
	&& strchr(en->path, '.') && strchr(en->path, '.') != en->path;
    if(want_magic){
	NOOP(stderr, "rfm_is_image %s\n", en->path);
	en->mimemagic = rfm_rational(MODULE_DIR, "mime", en, "mime_magic", "mime_function");
    }
    if (!en->mimemagic) en->mimemagic = g_strdup(_("unknown"));

    /* check for image support types */

    if(!pix_formats)
        pix_formats = gdk_pixbuf_get_formats ();
    for(l = pix_formats; l; l = l->next) {
        gchar **pix_mimetypes;
        int i;
        GdkPixbufFormat *fmt = l->data;
        pix_mimetypes = gdk_pixbuf_format_get_mime_types (fmt);
        for(i = 0; pix_mimetypes[i]; i++) {
            NOOP("allowable pix_format=%s\n", pix_mimetypes[i]);
            if(en->mimetype && 
		    !g_ascii_strcasecmp (pix_mimetypes[i], en->mimetype)) {
                g_strfreev (pix_mimetypes);
                return TRUE;
            }
            if(en->mimemagic && 
		    !g_ascii_strcasecmp (pix_mimetypes[i], en->mimemagic)) {
                g_strfreev (pix_mimetypes);
                return TRUE;
            }        
	}
        g_strfreev (pix_mimetypes);
        if(result) {
            return TRUE;
        }
    }
    /* let's keep this one in memory to reduce overhead */
    /*g_slist_free(pix_formats); */

    NOOP ("%s mimetype = %s, magic=%s is not image\n", en->path, en->mimetype, en->mimemagic);

    return result;
}

GdkPixbuf *
rfm_create_background_pixbuf (const char *file, int width, int height) {

    if (width < 0 || height < 0) {
        rfm_get_drawable_geometry (gdk_x11_get_default_root_xwindow (),
	    NULL, NULL, &width, &height, NULL);
    }
  
    GdkPixbuf *tgt;
    double factor,
      x_factor,
      y_factor;
    int w,
      h;
    if(!file) return NULL;
    NOOP ("getting pixbuf %s", file);
    GdkPixbuf *src = NULL;

    if (g_file_test(file, G_FILE_TEST_EXISTS)){
      src = rfm_pixbuf_new_from_file(file);
      if(!src) return NULL;
    }

    if(!src)
	return NULL;

    w = gdk_pixbuf_get_width (src);
    h = gdk_pixbuf_get_height (src);
    NOOP (" pixbuf is %d,%d width,height is %d,%d", w, h, width, height);
    x_factor = (double)width / w;
    y_factor = (double)height / h;
    factor = (x_factor < y_factor) ? x_factor : y_factor;
    NOOP ("factors %lf,%lf --> %lf, distorsion=%lf", x_factor, y_factor, factor, fabs (x_factor - y_factor) / factor);
    if(fabs (x_factor - y_factor) / factor < 0.20) {
      NOOP ("scaling pixbuf %s to %d,%d", file, width, height);
      tgt = rfm_pixbuf_scale_stretch(src, width, height, GDK_INTERP_BILINEAR);
    } else {
          NOOP ("factor=%lf scaling pixbuf %s to %lf,%lf", factor, file, factor * w, factor * h);
      tgt = rfm_pixbuf_scale_stretch(src, factor * w, factor * h, GDK_INTERP_BILINEAR);
    }
    g_object_unref (G_OBJECT (src));
    NOOP ("got pixbuf %s", file);
    return tgt;
}


Pixmap
rfm_create_background_pixmap (char *file) {
    /* create Pixmap */

    gint root_x;
    gint root_y;
    gint root_w;
    gint root_h;
    gint root_d;

    rfm_get_drawable_geometry(gdk_x11_get_default_root_xwindow (),
	    &root_x, &root_y, &root_w, &root_h, &root_d);
    Display *display=gdk_x11_display_get_xdisplay(gdk_display_get_default());
    Pixmap xpixmap = XCreatePixmap (display, GDK_ROOT_WINDOW (), root_w, root_h, root_d);

    // The X way
    // create a graphic context for the pixmap
    GC graphic_context = XCreateGC(display, xpixmap, 0, 0);;
    // get default colormap
    static Colormap colormap = None;
    if (colormap == None){
	colormap = DefaultColormap(display, 0);
    }
    
    // parse the color specification
    XColor background_color;
    const gchar *background=NULL;
    if(getenv ("RFM_DESKTOP_COLOR") && strlen (getenv ("RFM_DESKTOP_COLOR"))) {
	background=getenv ("RFM_DESKTOP_COLOR");
        if (!XParseColor(display, 
	    colormap, background, &background_color)) {
	    g_warning("cannot parse background color: %s", background);
	    background=NULL;
	}
    }
    if (!background) {
	background="#000000";
        if (!XParseColor(display, 
	    colormap, background, &background_color)) {
	    g_error("cannot parse default background color: %s", background);
	}
    }
    XAllocColor(display, colormap, &background_color);
    XSetForeground(display, graphic_context, background_color.pixel);
    XFillRectangle (display, xpixmap, graphic_context, 0, 0, root_w, root_h);


// cairo way
#ifndef CAIRO_HAS_XLIB_SURFACE
 g_warning ("CAIRO_HAS_XLIB_SURFACE is not defined!"); 
#else
    GdkPixbuf *pixbuf = NULL;
    if (file) {
	pixbuf = rfm_create_background_pixbuf (file, root_w, root_h);
	if(!GDK_IS_PIXBUF(pixbuf)) {
	    g_warning("cannot create pixbuf from %s", file);
	}
    }

    if (pixbuf && GDK_IS_PIXBUF(pixbuf)) {
	Visual *visual = gdk_x11_visual_get_xvisual(gdk_visual_get_system());
	cairo_surface_t *pixmap_surface = 
	    cairo_xlib_surface_create (
		    display, // Display *
		    xpixmap, // Drawable 
		    visual, // Visual *
		    root_w, root_h);
	if(cairo_surface_status (pixmap_surface) != CAIRO_STATUS_SUCCESS) {
	    g_error ("cairo_surface_status(surface) != CAIRO_STATUS_SUCCESS");
	}
	cairo_surface_reference(pixmap_surface);
	cairo_t *pixmap_context = cairo_create (pixmap_surface);
	cairo_reference(pixmap_context);

	gdk_cairo_set_source_pixbuf(pixmap_context, pixbuf,
		       (root_w - gdk_pixbuf_get_width (pixbuf)) / 2, 
		       (root_h - gdk_pixbuf_get_height (pixbuf)) / 2);
	cairo_paint (pixmap_context); 
	cairo_surface_destroy(pixmap_surface);
	cairo_destroy(pixmap_context);
	g_object_unref(pixbuf);
    }
#endif
    XFreeGC(display, graphic_context);
    return xpixmap;
}
/******************************************************************/

gchar *
rfm_get_thumbnail_path (const gchar * file, gint size) {
    gchar *cache_dir;
    gchar *thumbnail_path = NULL;
    GString *gs;
    gchar key[11];

    cache_dir = g_build_filename (RFM_THUMBNAIL_DIR, NULL);
    if(g_mkdir_with_parents (cache_dir, 0700) < 0) {
        g_free (cache_dir);
        return NULL;
    }

    /* thumbnails are not subject to thumbnailization: */
    gchar *dirname = g_path_get_dirname (file);
    if(strncmp (cache_dir, dirname, strlen (cache_dir)) == 0) {
        NOOP ("thumbnails cannot be thumbnailed:%s\n", file);
        g_free (cache_dir);
        g_free (dirname);
        return NULL;
    }

    gs = g_string_new (dirname);
    sprintf (key, "%10u", g_string_hash (gs));
    g_strstrip (key);
    g_string_free (gs, TRUE);
    g_free (dirname);

    gchar *thumbnail_dir = g_build_filename (cache_dir, key, NULL);
    if(g_mkdir_with_parents (thumbnail_dir, 0700) < 0) {
        g_free (thumbnail_dir);
        return NULL;
    }

    gchar *filename = g_path_get_basename (file);

    gs = g_string_new (file);
    sprintf (key, "%10u", g_string_hash (gs));
    g_strstrip (key);
    g_string_free (gs, TRUE);
    g_free (filename);

    filename = g_strdup_printf ("%s-%d.png", key, size);
    thumbnail_path = g_build_filename (thumbnail_dir, filename, NULL);
    g_free (filename);
    g_free (cache_dir);
    g_free (thumbnail_dir);
    NOOP ("thread: %s ->thumbnail_path=%s\n", file, thumbnail_path);

    return thumbnail_path;
}

GdkPixbuf *
rfm_get_from_pixbuf_hash (const gchar * key, int size) {
    pixbuf_t *pixbuf_p = find_in_pixbuf_hash (key, size);
    if(pixbuf_p) return pixbuf_p->pixbuf;
    return NULL;
}

GdkPixbuf *
rfm_get_pixbuf (const gchar * key, int size) {
    return get_pixbuf(key, size, TRUE);
}


void
rfm_change_icontheme(void){
    if(!pixbuf_hash) return;
	
    GtkWidget *window = rfm_global_p->window;

    gchar *rodent_theme = g_strdup_printf("%s/%s/%s/%s", PACKAGE_DATA_DIR, "icons", "Rodent", "index.theme");
    if (!g_file_test(rodent_theme, G_FILE_TEST_EXISTS)){
	g_free(rodent_theme);
	return;
    }
    g_free(rodent_theme);
    // This will cause the persistent X crash on changing option in dialog
    //gdk_flush(); 
    //Display *display=gdk_x11_display_get_xdisplay(gdk_display_get_default());
    //XSync(display, False);   
    // This will also cause the crash:
    // gdk_flush(); 
    // These instructions require a barrier lock, apparently.

    g_mutex_lock (pixbuf_hash_mutex);
    g_hash_table_foreach (pixbuf_hash, update_pixbuf_f, NULL);
    g_mutex_unlock (pixbuf_hash_mutex);
    GdkRectangle allocation;
    gtk_widget_get_allocation (window, &allocation);
    allocation.x = 0;
    allocation.y = 0;

    widgets_t *widgets_p = rfm_get_widget("widgets_p");
    rfm_expose_rect (widgets_p->view_p, &allocation);
}


static void *
pixbuf_scale_simple(void *data){
    void **arg = data;
    GdkPixbuf *in_src = arg[0];
    gint width = GPOINTER_TO_INT(arg[1]);
    gint height = GPOINTER_TO_INT(arg[2]);
    gint type = GPOINTER_TO_INT(arg[3]);
    g_free(arg);
    GdkPixbuf *src = NULL;
    gint ph = gdk_pixbuf_get_height (in_src);
    gint pw = gdk_pixbuf_get_width (in_src);

    if (ph > pw) {
	    gint w = height * pw / ph;
	    src = gdk_pixbuf_scale_simple (in_src, w, height, type);
    } else {
	    gint h = width * ph / pw;
	    src = gdk_pixbuf_scale_simple (in_src, width, h, type);
    }
    return src;
}

static gboolean
pixbuf_save(void *data){
    void **arg = data;
    GdkPixbuf *tgt = arg[0];
    gchar *path = arg[1];
    g_free(arg);
    GError *error = NULL;
    gdk_pixbuf_save (tgt, path, "png", &error,
                     "tEXt::Software", "Rodent", NULL);
    if (error){
	g_warning("pixbuf_save(): %s\n", error->message);
	g_error_free(error);
    }
    g_free(path);
    return FALSE;
}

static void *
pixbuf_from_file(void *data){
    gchar *path = data;
    GError *error = NULL;
    GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file (path, &error);
    if(error) {
	    g_warning ("pixbuf_from_file() %s:%s\n", error->message, path);
	    g_error_free (error);
    }
    g_free (path);
    return pixbuf;
}


GdkPixbuf *
rfm_pixbuf_scale_simple (GdkPixbuf *in_pixbuf,  gint size, gint type){
    void **arg = (void **)malloc(4*sizeof(void *));
    if (!arg) g_error("malloc: %s\n", strerror(errno));
    arg[0] = in_pixbuf;
    arg[1] = GINT_TO_POINTER(size);
    arg[2] = GINT_TO_POINTER(size);
    arg[3] = GINT_TO_POINTER(type);
    GdkPixbuf *pixbuf;
    if (rfm_global_p && rfm_global_p->self != g_thread_self()){
	pixbuf = rfm_context_function(pixbuf_scale_simple, arg);
    } else {
	pixbuf = pixbuf_scale_simple(arg);
    }
    return pixbuf;
}

GdkPixbuf *
rfm_pixbuf_scale_stretch (GdkPixbuf *in_pixbuf,  gint width, gint height, gint type){
    void **arg = (void **)malloc(4*sizeof(void *));
    if (!arg) g_error("malloc: %s\n", strerror(errno));
    arg[0] = in_pixbuf;
    arg[1] = GINT_TO_POINTER(width);
    arg[2] = GINT_TO_POINTER(height);
    arg[3] = GINT_TO_POINTER(type);
    GdkPixbuf *pixbuf;
    if (rfm_global_p && rfm_global_p->self != g_thread_self()){
	pixbuf = rfm_context_function(pixbuf_scale_simple, arg);
    } else {
	pixbuf = pixbuf_scale_simple(arg);
    }
    return pixbuf;
}

	

void 
rfm_pixbuf_save(GdkPixbuf *tgt, const gchar *path){
    if (!tgt || !path) return ;
    void **arg =(void **)malloc(2*sizeof(void *));
    if (!arg) g_error("malloc: %s\n", strerror(errno));
    arg[0] = tgt;
    arg[1] = g_strdup(path);
    if (rfm_global_p && rfm_global_p->self != g_thread_self()){
	// don't wait here
	g_main_context_invoke(NULL, pixbuf_save, arg);
    } else {
	pixbuf_save(arg);
    }
    return;
}


GdkPixbuf *
rfm_pixbuf_new_from_file (const gchar *path){
    if (!path) return NULL;
    GdkPixbuf *pixbuf;
    if (rfm_global_p && rfm_global_p->self != g_thread_self()){
	pixbuf = rfm_context_function(pixbuf_from_file, g_strdup(path));
    } else {
	pixbuf = pixbuf_from_file(g_strdup(path));
    }


    return pixbuf;
}

static void *
insert_pixbuf_tag (void *data){
    void **arg = data;
    const GdkPixbuf *tag = arg[0];
    GdkPixbuf *composite_pixbuf = arg[1];
    gchar *where = arg[2];
    double *scale_factor = arg[3];
    gint overall_alpha = GPOINTER_TO_INT(arg[4]);
    g_free(arg);


    gint width = gdk_pixbuf_get_width (composite_pixbuf);
    gint height = gdk_pixbuf_get_height (composite_pixbuf);
    // "SW"
    gint dest_x = 0;
    gint dest_y = 0;
    gint dest_width = width;
    gint dest_height = height;
    gdouble scale_x = 1.0 / (*scale_factor);
    gdouble scale_y = 1.0 / (*scale_factor);

    // default SW
#define WW ((gdouble)width / (*scale_factor) / 4)
#define HH ((gdouble)height / (*scale_factor) / 8)
    gdouble offset_x = 0.0;
    gdouble offset_y = height - ((gdouble)height) / (*scale_factor);
    if(strcmp (where, "SE") == 0) {
        offset_x = width - ((gdouble)width) / (*scale_factor);
    } else if(strcmp (where, "NW") == 0) {
        offset_y = 0.0;
    } else if(strcmp (where, "NE") == 0) {
        offset_x = width - ((gdouble)width) / (*scale_factor);
        offset_y = 0.0;
    } else if(strcmp (where, "C") == 0) {
        offset_x = ((gdouble)width) / 2 - ((gdouble)width) / (*scale_factor) / 2 - WW;
        offset_y = ((gdouble)height) / 2 - ((gdouble)height) / (*scale_factor) / 2 + HH;
    } else if(strcmp (where, "SC") == 0) {
        offset_x = ((gdouble)width) / 2 - ((gdouble)width) / (*scale_factor) / 2 - WW;
    } else if (strcmp (where, "NC") == 0) {
        offset_y = 0.0;
        offset_x = ((gdouble)width) / 2 - ((gdouble)width) / (*scale_factor) / 2 - WW;
    }

    gdk_pixbuf_composite (tag, composite_pixbuf,
                          dest_x, dest_y,
                          dest_width, dest_height,
			  offset_x, offset_y,
			  scale_x, scale_y,
			  GDK_INTERP_BILINEAR, overall_alpha);

    g_free(scale_factor);
    g_free(where);
    return NULL;
}

gboolean
rfm_insert_pixbuf_tag (GdkPixbuf *tag, GdkPixbuf *composite_pixbuf,
	const gchar *where, gdouble scale_factor, gint overall_alpha){
    if (!tag || !composite_pixbuf || !GDK_IS_PIXBUF(composite_pixbuf) || !where) {
	return FALSE;
    }
    void **arg = (void **)malloc(5*sizeof(void *));
    if (!arg) g_error("malloc: %s\n", strerror(errno));
    arg[0] = tag;
    arg[1] = composite_pixbuf;
    arg[2] = g_strdup(where);
    arg[3] = (gdouble *)malloc(sizeof(gdouble));
    if (!arg[3]) g_error("malloc: %s\n", strerror(errno));
    memcpy(arg[3], &scale_factor, sizeof(gdouble));
    arg[4] = GINT_TO_POINTER(overall_alpha);
    
    if (rfm_global_p && rfm_global_p->self != g_thread_self()){
	rfm_context_function(insert_pixbuf_tag, arg);
    } else {
	insert_pixbuf_tag(arg);
    }
    return TRUE;
}

static void *
pixbuf_from_gtkid(void *data){
    void **arg = data;
    gchar *id = arg[0];
    gint size = GPOINTER_TO_INT(arg[1]);
    GdkPixbuf *pixbuf = NULL;
    //hack: GtkWidget *image=gtk_image_new_from_file("whatever");
    GtkWidget *image=gtk_image_new();
#ifdef USE_GTK2
    pixbuf = gtk_widget_render_icon (image, id, GTK_ICON_SIZE_DIALOG,NULL);
#else
    pixbuf = gtk_widget_render_icon_pixbuf (image, id, GTK_ICON_SIZE_DIALOG);
#endif
    //if (pixbuf && size > SMALL_ICON_SIZE) 
    {
	GdkPixbuf *new_pixbuf = 
	    gdk_pixbuf_scale_simple (pixbuf, size, size, GDK_INTERP_BILINEAR);
	g_object_unref(pixbuf);
	pixbuf=new_pixbuf;
    }

    //g_object_unref(G_OBJECT(image));
    g_free(id);
    g_free(arg);
    return pixbuf;	    
}

GdkPixbuf *
rfm_pixbuf_from_gtkid(const gchar * id, gint size){
    if (!id) return NULL;
    GdkPixbuf *pixbuf = NULL;
    void **arg = (void **)malloc(2*sizeof(void *));
    if (!arg) g_error("malloc: %s\n", strerror(errno));
    arg[0] = g_strdup(id);
    arg[1] = GINT_TO_POINTER(size);
    
    if (rfm_global_p && rfm_global_p->self != g_thread_self()){
	pixbuf = rfm_context_function(pixbuf_from_gtkid, arg);
    } else {
	pixbuf = pixbuf_from_gtkid(arg);
    }
    return pixbuf;
}

static void *
duplicate_pixbuf(void *data){
    GdkPixbuf *pixbuf = data;
    GdkPixbuf *new_pixbuf = gdk_pixbuf_copy (pixbuf);
    return new_pixbuf;
}
    


GdkPixbuf *
rfm_pixbuf_duplicate(GdkPixbuf *pixbuf){
    if (!pixbuf) return NULL;
    GdkPixbuf *new_pixbuf = NULL;
    if (rfm_global_p && rfm_global_p->self != g_thread_self()){
	new_pixbuf = rfm_context_function(duplicate_pixbuf, pixbuf);
    } else {
	new_pixbuf = duplicate_pixbuf(pixbuf);
    }
    return new_pixbuf;
}



////////////////   module wraparounds:
//

gint
rfm_svg_supported (void) {
    void *value = rfm_void (MODULE_DIR, "icons", "svg_supported");
    return GPOINTER_TO_INT (value);
}

///////////////////////  primary_icon_resolve.i   /////////////////////


static const gchar *
get_type_icon_id(record_entry_t *en){
    if(IS_BROKEN_LNK (en->type)) return ("xffm/stock_missing-image");
    if(IS_UP_TYPE (en->type)) {
	// When up item is not a directory.
        return "xffm/stock_go-up";
    } 
    if (IS_SDIR (en->type)){
	// No access directory_:
        if(IS_NOACCESS_TYPE (en->type)) {
	    return "xffm/stock_directory/composite2/emblem_no-access";
        } 
	// Invariant under write or no write flag:
	if(strcmp (en->path, g_get_home_dir ()) == 0) {
	    return "xffm/stock_home";
	} else if(strcmp (en->path, "/") == 0){
		return "xffm/stock_directory";
	} else if(
		(strstr (en->path, "cdrom") ||
		 strstr (en->path, "cdrw") ||
		 strstr (en->path, "dvd"))
		 &&  FSTAB_is_in_fstab(en->path))
	{
	    return "xffm/stock_cdrom";
	} 
	else if(IS_NOWRITE_TYPE (en->type)) {
	    // No write directory
	    return "xffm/stock_directory";
        } else {
	    // Write OK
	    return "xffm/stock_directory/composite/emblem_write-ok";
	}
    } 
    if (IS_SCHR(en->type)) return ("inode/chardevice");
    if (IS_SBLK(en->type)) return ("xffm/stock_harddisk");
    if (IS_SFIFO(en->type)) return ("inode/fifo");
    if (IS_SSOCK(en->type)) return ("inode/socket");
    if (IS_PARTITION_TYPE(en->type)) return ("inode/mount-point");
//    if (IS_SLNK(en->type)) return g_strdup("inode/symlink");
//    if ((en->type)) return g_strdup("inode/virtual-disk");
    return NULL;

}

const gchar *
get_plain_icon_id (record_entry_t * en) {
    if(IS_UP_TYPE (en->type)) {
        return ("xffm/stock_go-up");
    }
    const gchar *id="xffm/stock_file";

    if ( IS_SFIFO(en->type)) return "inode/fifo";
    if ( IS_SCHR(en->type)) return "inode/chardevice";
    if ( IS_SBLK(en->type)) return "xffm/stock_harddisk";
    if ( IS_SSOCK(en->type)) return "inode/socket";  

    if (IS_SDIR(en->type)) {
	    id="xffm/stock_directory";
    } else if (en->st && en->st->st_mode & (S_ISUID|S_ISGID|S_ISVTX)) { 
	id="xffm/generic_executable/composite/stock_run";
    } else if (en->st && en->st->st_mode & (S_IXUSR|S_IXGRP|S_IXOTH)) { 
	id="xffm/generic_executable";
    } else if (en->st && !rfm_read_ok(en->st)) {
	id="xffm/stock_file/composite2/emblem_no-read";
    } else {
	id="xffm/stock_file";
    } 
    
    return id;
}

static gchar *greenball_id(view_t *view_p, record_entry_t *en, gchar *icon_id){
    if (!icon_id || !en) return NULL;
    /* block controlled in fstab plugin: */
    gboolean valid_mount_view = 
	(view_p && (!view_p->en || view_p->en->module ||  IS_LOCAL_TYPE(view_p->en->type)));
    gboolean valid_mount_point =
	(en->path && (IS_SDIR(en->type) || IS_PARTITION_TYPE (en->type)));
    if (valid_mount_point && valid_mount_view){
	gint mounted = GPOINTER_TO_INT(rfm_natural(PLUGIN_DIR, "fstab", en, "entry_is_mounted"));
	if (mounted<0) {
	    gchar *g = g_strconcat (icon_id, "/composite/xffm/emblem_no-access", NULL);
	    g_free(icon_id);
	    icon_id = g;
	} else	if (mounted>0) {
	    gchar *g = g_strconcat (icon_id, "/composite/stock_yes", NULL);
	    g_free(icon_id);
	    icon_id = g;
	} else if (rfm_natural(PLUGIN_DIR, "fstab", en->path, "is_in_fstab")){
	    gchar *g = g_strconcat (icon_id, "/composite/stock_no", NULL);
	    g_free(icon_id);
	    icon_id = g;
	}
    } 
    return icon_id;
}

static gboolean
save_icon_id_to_cache(record_entry_t *en, const gchar *id){
    if (IS_UP_TYPE(en->type) || IS_DUMMY_TYPE(en->type)) return FALSE;
    const gchar *path = en->path;
    DBHashTable *icon_id_cache;
    gchar *g = g_build_filename (ICON_ID_DBH_FILE, NULL);
    if (!g_file_test(g, G_FILE_TEST_EXISTS)){
	unsigned char key_length=11;
	icon_id_cache = dbh_new (g, &key_length, DBH_CREATE|DBH_PARALLEL_SAFE|DBH_THREAD_SAFE);
    } else {
	icon_id_cache = dbh_new (g, NULL, DBH_PARALLEL_SAFE|DBH_THREAD_SAFE);
    }
    if (!icon_id_cache){
	g_warning("cannot open %s for write.\n", g);
	g_free(g);
	return FALSE;
    }
    dbh_set_parallel_lock_attempt_limit(icon_id_cache, 50);
    GString *gs = g_string_new (path);
    gchar *key = g_strdup_printf ("%10u", g_string_hash (gs));
    dbh_set_key  (icon_id_cache,  (unsigned char *)key);
    dbh_set_data (icon_id_cache, (void *)id, strlen(id)+1);
    g_string_free (gs, TRUE);
    g_free(key);
    gboolean retval = TRUE;
    if (dbh_update (icon_id_cache) == 0){
	g_warning("cannot write record %s to cache: %s.\n", path, g);
	retval = FALSE;
    }
    dbh_close(icon_id_cache);
    g_free(g);
    return retval;
}
	
static gchar *
get_icon_id_from_cache(record_entry_t *en){
    //return NULL;
    if (IS_UP_TYPE(en->type) || IS_DUMMY_TYPE(en->type)) return NULL;
    DBHashTable *icon_id_cache;
    gchar *cache_value=NULL;
    gchar *g = g_build_filename (ICON_ID_DBH_FILE, NULL);
    if((icon_id_cache = dbh_new (g, NULL, DBH_READ_ONLY|DBH_PARALLEL_SAFE|DBH_THREAD_SAFE)) != NULL) {
	dbh_set_parallel_lock_attempt_limit(icon_id_cache, 50);
	GString *gs = g_string_new (en->path);
	sprintf ((char *)DBH_KEY (icon_id_cache), "%10u", g_string_hash (gs));
	g_string_free (gs, TRUE);
	if (dbh_load (icon_id_cache) != 0){
	    cache_value = g_strdup((gchar *)icon_id_cache->data);
	}
	dbh_close(icon_id_cache);
    }
    g_free(g);
    return cache_value;
}

gchar *
rfm_get_entry_icon_id(view_t *view_p, record_entry_t *en, gboolean magic_icon){
    if (en==NULL) return g_strdup("xffm/device_computer");
    const gchar *id = NULL;
	
    // Modules may override icon resolution.
    if (en->module){
	id = rfm_natural (PLUGIN_DIR, en->module, en, "item_icon_id");
        // if module does not resolve icon, proceed as normal.
        if(id) {
	    NOOP("rfm_get_entry_icon_id(): %s: %s\n", en->path, id);
	    return greenball_id(view_p, en, g_strdup(id));
        }
	NOOP("module: %s does not resolve icon for %s\n", en->module, en->path);
    }


    if (!IS_LOCAL_TYPE(en->type) || en->module){
	// Remote files or module items will never be done magic.
	magic_icon = FALSE;	
    } 
    if (!magic_icon ) {
	// valid view? directory type icon?
	if (view_p->en && IS_LOCAL_TYPE(view_p->en->type) && IS_SDIR(en->type)){
	    id = get_plain_icon_id(en);	    
	    return greenball_id(view_p, en, g_strdup(id));
	}
	gchar *cache_value=get_icon_id_from_cache(en);
	if (cache_value) return cache_value;
	id = get_plain_icon_id (en);
	// remote file but valid view entry?
	// (this does not do the trick...)
	//if (IS_SDIR(en->type)) return greenball_id(view_p, en, g_strdup(id));
	return g_strdup(id);
    }

    // Try to resolve from readdir type information first.
    id = get_type_icon_id(en);
    if (id) {
	// greenball resolution
	save_icon_id_to_cache(en, id);
	return  greenball_id(view_p, en, g_strdup(id));
    }

    // We need the stat record now.
    if (!en->st ) g_error("WTF: %s no stat record\n", en->path);

    if (!en->st->st_blksize && !en->st->st_mtime){
	DBG("mime magic will not work if stat record is not initialized.\n");
	return g_strdup("xffm/stock_file/composite2/stock_dialog-question");
    }

    // Try to get a valid extension mimetype first.
    g_free(en->mimetype);
    en->mimetype = MIME_type(en->path, en->st);
    if (!en->mimetype) {
	    en->mimetype = g_strdup(_("unknown"));
    }
    
    // Only in the event that a valid mime type is not found, 
    // and magic option is set, do magic.
    if (magic_icon && strcmp(en->mimetype, _("unknown"))==0) {
	// Does the user have read permission for the file?
	// Mime magic will not work otherwise...
	// We need a valid stat record for this.
	g_free(en->mimemagic);
	if (rfm_read_ok(en->st)) {
	    // Magic at this step as a last resort
	    en->mimemagic = rfm_rational(MODULE_DIR, "mime", en, "mime_magic", "mime_function");
	    if (!en->mimemagic) en->mimemagic = g_strdup(_("unknown"));
	} else {
	    en->mimemagic = g_strdup(_("No Read Permission"));
	}
    }

    // Resolve icon id from mimetype or mimemagic.
    if (en->mimetype && strcmp(en->mimetype, _("unknown"))){
	id = en->mimetype;
    } 
    if (!id) {
	id = en->mimemagic;
    }
    // Override of particular mimetypes
    if (!id || strcmp(id, _("unknown"))==0) {
	return g_strdup("xffm/stock_file");
    }
    if (strcmp(id, _("No Read Permission"))==0){
	return g_strdup("xffm/stock_file/composite/emblem_no-read");
    }
    // Override dotdesktop icons with module icon determination.
    if (strcmp("application/x-desktop", id)==0) {
	const gchar *icon_name=rfm_natural(PLUGIN_DIR, "dotdesktop", en, "item_icon_id");
	if (icon_name) {
	    save_icon_id_to_cache(en, icon_name);
	    return g_strdup(icon_name);
	}
    }
    save_icon_id_to_cache(en, id);
    
#if 0
    // This would need fine tuning...
    if (!IS_LOCAL_TYPE(en->type)){
	gchar *g = g_strconcat(icon_id, "/composite/emblem_shared", NULL);
	g_free(icon_id);
	icon_id = g;
    }
#endif
    // Greenball resolution 
    return greenball_id(view_p, en, g_strdup(id));
}
