
/*
 * Copyright (C) 2002-2012 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rodent_libs.h"
#include "rodent_expose.i"

#if 10
void
rodent_expose_all (view_t * view_p) {
    TRACE("rodent_expose_all: %d,%d %d,%d\n",
	    rfm_global_p->allocation->x,
	    rfm_global_p->allocation->y,
	    rfm_global_p->allocation->width,
	    rfm_global_p->allocation->height
	    );
    rfm_expose_rect(view_p, rfm_global_p->allocation);
}
#endif

static
population_t *find_grid_element(view_t * view_p, gdouble x, gdouble y) {
    if(!view_p->population_pp)  return NULL;

    // Margin considerations:
    x -= MARGIN_LEFT(view_p);
    y -= MARGIN_TOP(view_p); 
    if(x < 0 || y < 0)  return NULL;

    // rectangle x,y coordinates
    gint row = y / CELLHEIGHT(view_p);
    gint column = x / CELLWIDTH(view_p);
    gint grid_rows = GRID_ROWS(view_p);
    gint grid_columns = GRID_COLUMNS(view_p);
    if(column >= grid_columns || row >= grid_rows) return NULL;

    gint element;
    if (view_p->flags.type == ICONVIEW_TYPE) {
	element= row * grid_columns + column;
    } else { // transposed
	element= column * grid_rows + row;
    }
    if(element >= MAX_ELEMENTS(view_p)) return NULL;

    population_t *population_p = *(view_p->population_pp + element);

    if (population_p && 
		(population_p->row != row || population_p->column != column)) {
	// This is a sanity hack test. Should never happen in a perfect world.
	NOOP ("THIS IS WRONG: row, column: current (%d,%d) correct (%d,%d) element=%d columns=%d\n",
	    population_p->row, population_p->column,
	    row, column, element,
	    grid_columns);
	population_p->row = row;
	population_p->column = column;
    }

    return population_p;
}


const 
population_t *
rodent_find_in_population (view_t * view_p, gdouble x, gdouble y) {


    population_t *population_p = find_grid_element(view_p, x, y);
    if(!population_p) return NULL;

#if 0
    // this should work, but doesn't...
    GdkRectangle icon_rect;
    if (!rfm_get_population_rect(view_p, population_p, &icon_rect)) {
	g_warning("rodent_find_in_population: !rfm_get_population_rect\n");
	return FALSE;
    }
    if (x >= icon_rect.x && x < icon_rect.x + icon_rect.width &&
	y >= icon_rect.y && y < icon_rect.y + icon_rect.height) {
	return population_p;
    }
    return NULL;
#else
    gint icon_size = ICON_SIZE(view_p);
    gint cellwidth = CELLWIDTH(view_p);
    gint icon_width = icon_size;
    gint icon_height = icon_size;
    if (population_p->flags & POPULATION_THUMBNAILED) {
	icon_width = population_p->pixbufW;
	icon_height = population_p->pixbufH;
    }
    gint x_spacing = (icon_size >= SMALL_ICON_SIZE)?(cellwidth - icon_width) / 2: 0;
    if(x_spacing < 0) x_spacing = 0;
    gint X = (gint)(x-MARGIN_LEFT(view_p)) % cellwidth;

    if(X < x_spacing || X > x_spacing + icon_width) return NULL;


    gint Y = (gint)(y-MARGIN_TOP(view_p)) % CELLHEIGHT(view_p);
    gint y_spacing = (icon_size - icon_height)/2;

    if(Y < y_spacing || Y > y_spacing + icon_height) return NULL;
    return population_p;
#endif
}

const
population_t *
rodent_find_in_labels (view_t * view_p, gdouble x, gdouble y) {

	
    // Find grid element.
    population_t *population_p = find_grid_element(view_p, x, y);
    if(!population_p) return NULL;
;

    GdkRectangle label_rect;
    if (!rfm_get_population_label_rect(view_p, population_p, &label_rect))return NULL;
    if(x >= label_rect.x && x < label_rect.x + label_rect.width 
	    &&
       y >= label_rect.y && y < label_rect.y + label_rect.height)
    {
	return population_p;
    }
    return NULL;
}

    

// Find elements (or icons, specified by the respective gboolean)
// within an expose rectangle. The icons restriction is to do rubberband
// selecting considering only the icons to select.

GSList *rodent_find_items_in_rectangle(view_t * view_p, GdkRectangle *rect, gboolean icons) {

    if(!rect->width || !rect->height) {
	NOOP(stderr, "rodent_find_items_in_rectangle(): !rect->width || !rect->height\n");
	return NULL;
    }
    if (!rfm_population_try_read_lock(view_p)) {
	NOOP(stderr, "rodent_find_items_in_rectangle(): unable to get readlock\n");
	return NULL;
    }
    if(!view_p->population_pp){
	rfm_population_read_unlock (view_p);
        return NULL;
    }

    GSList *list=NULL;

    gint cellwidth = CELLWIDTH(view_p);
    gint cellheight = CELLHEIGHT(view_p);
    NOOP(stderr, "x: %d-%d, y:%d-%d w:%d H: %d\n",
	    rect->x, rect->x+rect->width,
	    rect->y, rect->y+rect->height,
	    cellwidth, cellheight);

    gint low_X = rect->x;
    gint low_Y = rect->y;
	// Margin considerations:
	low_X -= MARGIN_LEFT(view_p);
	low_Y -= MARGIN_TOP(view_p);
    gint high_X = low_X + rect->width - 1;
    gint high_Y = low_Y + rect->height - 1;

    if(low_X < 0){ low_X = 0; }
    if(low_Y < 0){ low_Y = 0; }

    gint low_column  = low_X / cellwidth;
    gint low_row     = low_Y / cellheight;
    gint high_column = high_X / cellwidth;
    gint high_row    = high_Y / cellheight;

   // if ( low_X % cellwidth) low_column--;
   // if ( low_Y % cellheight) low_row--;

   
    gint grid_rows = GRID_ROWS(view_p);
    gint grid_columns = GRID_COLUMNS(view_p);
    if (low_column < 0) low_column = 0;
    if (low_row < 0) low_row = 0;

    if (high_column > grid_columns - 1) {
	NOOP(stderr, "rodent_find_items_in_rectangle(): logic error, high_column (%d) > grid_columns - 1 (%d)\n",

		high_column, grid_columns - 1);
	high_column = grid_columns - 1;
    }
    if (high_row > grid_rows - 1) {
	NOOP(stderr, "rodent_find_items_in_rectangle(): logic error, high_row (%d) > grid_rows - 1 (%d)\n",
		high_row, grid_rows - 1);
	high_row = grid_rows - 1;
    }

    NOOP(stderr, "rodent_find_items_in_rectangle(): rows = %d-%d, columns = %d-%d cellwidth=%d  cellheight=%d\n", 
	    low_row, high_row, low_column, high_column,
	    cellwidth, cellheight);

    // Just check if the element corresponding to the given
    // row and column is within the selected rows and columns.
    gint column;
    gint row;
    gint max_elements = MAX_ELEMENTS(view_p);

    for (column=low_column; column<=high_column; column++){
	for (row=low_row; row<=high_row; row++) {
	    gint element;
	    if (view_p->flags.type == ICONVIEW_TYPE) {
		element = row * grid_columns + column;
	    } else { // transposed
		element = column * grid_rows + row;
	    }
	    if (element >= max_elements) {
		// This may occur when bottom row is not full.
		continue;
	    }
	    population_t *population_p = view_p->population_pp[element];
	    if (!population_p) continue;
	    NOOP("element: %d/%d (%s)\n",
		    element, view_p->max_elements-1,
		    (population_p->en)?population_p->en->path:NULL);
	    // Ok, we item is within the rows and columns, but are we only interested
	    // in the icons? 
	    gboolean found = FALSE;
	    if (!icons){
		found = TRUE;
	    } else 
	    {
		// This is just to consider the icons in the selection
		// process (for rubberband selection, mainly).
		GdkRectangle icon_rect;
		if (!rfm_get_population_icon_rect(view_p, population_p, &icon_rect)) continue;
		gint low_x = icon_rect.x;
		gint high_x = icon_rect.x + icon_rect.width;
		gint low_y = icon_rect.y;
		gint high_y = icon_rect.y + icon_rect.height;
		gboolean X_in=FALSE;
		gboolean Y_in=FALSE;
		X_in = (low_x >= low_X+MARGIN_LEFT(view_p) && 
			low_x < high_X+MARGIN_LEFT(view_p)) ||
		       (high_x >= low_X+MARGIN_LEFT(view_p) &&
			high_x < high_X+MARGIN_LEFT(view_p));
		Y_in = (low_y >= low_Y+MARGIN_TOP(view_p) && 
			low_y < high_Y)+MARGIN_TOP(view_p) ||
		       (high_y >= low_Y+MARGIN_TOP(view_p) && 
			high_y < high_Y+MARGIN_TOP(view_p));
		if(X_in && Y_in){
		    found = TRUE;
		}
	    }
	    if (found) {
		NOOP(stderr, "rodent_find_items_in_rectangle(): found item %d/%d\n", element, max_elements);
		if (population_p->flags & LABEL_SATURATED){
		    list=g_slist_append(list, population_p);
		    //list=g_slist_prepend(list, population_p);
		} else list=g_slist_prepend(list, population_p);	
		//} else list=g_slist_append(list, population_p);	
	    }
	}
    }
    rfm_population_read_unlock(view_p);
    return list;
}

void * 
rodent_clean_paper(gpointer data){
    widgets_t *widgets_p = data;
    cairo_t *gdk_context =  
	gdk_cairo_create(gtk_widget_get_window(widgets_p->paper));
    view_t *view_p = widgets_p->view_p;
    GdkRectangle allocation;
    gtk_widget_get_allocation(widgets_p->paper, &allocation);


    
    rodent_clean_rectangle (view_p, 0, 0, allocation.width, allocation.height, gdk_context);
	cairo_destroy(gdk_context);
    
    return NULL;
}

void 
rodent_threaded_clean_paper(widgets_t *widgets_p)
{
    if (rfm_global_p->self ==g_thread_self()){
	rodent_clean_paper (widgets_p);
	return;
    }
    rfm_context_function(rodent_clean_paper, widgets_p);
    return;
}
    
    
static void *
queue_an_expose(void *data){
    rfm_threadwait();
    void **arg = data;
    view_t *view_p = arg[0];
    void *area_p = arg[1];
    g_free(arg);
    g_thread_yield();
    rfm_expose_rect (view_p, (GdkRectangle *)area_p);
    g_free(area_p);
    return NULL;
}

static gboolean noexpose=TRUE;
//#define BLINK
#ifdef BLINK
//blink() does not work, main thread is blocked...

static gboolean
blink(gpointer data){
    static gint count=0;

    const gchar *icon_id = "xffm/status_image-loading";
    static GdkPixbuf *pixbuf = NULL;
    static GdkPixbuf *unpixbuf = NULL;
    if (!pixbuf) {
	pixbuf = rfm_get_pixbuf(icon_id, BIG_ICON_SIZE);
	gint width = gdk_pixbuf_get_width(pixbuf);
	gint height = gdk_pixbuf_get_height(pixbuf);
	unpixbuf =  gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8,
	    width,height);
        gdk_pixbuf_fill(unpixbuf, 0x0);
	gdk_pixbuf_copy_area (pixbuf, 0, 0, width, height, unpixbuf, 0,0);
	gdk_pixbuf_saturate_and_pixelate (unpixbuf, unpixbuf, 0.0, TRUE);
	g_object_ref(pixbuf);
	g_object_ref(unpixbuf);
    }

    // 1. figure out where repeated expose-all is coming from on first call,
    //    and when vpane is reconfigured, sometimes...
    if (noexpose){
	widgets_t *widgets_p = data;
	gint x = rfm_global_p->allocation->width / 2 - BIG_ICON_SIZE / 2;
	gint y = rfm_global_p->allocation->height / 2 - BIG_ICON_SIZE / 2;
	if (widgets_p->vpane){
	    gdouble page=gtk_adjustment_get_page_size (
		gtk_scrolled_window_get_vadjustment (
		    widgets_p->scrolled_window));
	    gdouble value = gtk_adjustment_get_value (
		gtk_scrolled_window_get_vadjustment (
		    widgets_p->scrolled_window));
	    y = floor(value + page)/2 - BIG_ICON_SIZE / 2;
	    if (y < 0) y=0;
	}
	cairo_t *gdk_context = gdk_cairo_create(gtk_widget_get_window(widgets_p->paper));
	gdk_cairo_set_source_pixbuf(gdk_context, 
		(count%2)?unpixbuf:pixbuf,
			x,y);
	count++;
	//cairo_paint_with_alpha(gdk_context, 0.450);
	cairo_paint (gdk_context); 
	    cairo_destroy(gdk_context);
	//gdk_flush();  
    }
    return noexpose;
}
#endif
// Function to set/unset hold on signal generated expose callback
void rodent_set_expose_hold(view_t *view_p, gboolean expose_hold){
    noexpose = expose_hold;
    // Here we could draw an animated pixbuf in position zero...
    // http://openbooks.sourceforge.net/books/wga/graphics-gdk-pixbuf.html
    // Basically, with a main loop timeout, for each time get a 
    // progressive bitmask and copy the pixmap (with the progressive bitmask)
    // to the drawable. 
    // Probably easiest, don't even use an animated pixbuf. Just toggle between
    // pixmap and unsaturated... 
    if (noexpose){
	// might be called by a thread or main thread.
	rodent_threaded_clean_paper(&(view_p->widgets));	
#ifdef BLINK
	// Add timeout
	g_timeout_add(500, blink, &(view_p->widgets));
#endif
    }
}


#ifdef USE_GTK2
void
rodent_expose (GtkWidget * widget, GdkEventExpose * event, gpointer data) {
    NOOP ("GTK2, rodent_expose(): x,y=%d,%d w,h=%d,%d\n", event->area.x, event->area.y, event->area.width, event->area.height);
    view_t *view_p = (view_t *) data;
    if(!view_p || !view_p->widgets.paper)  return;

    gboolean layout_ready = rfm_layout_is_setup(view_p);

#if 10
    if (noexpose || !layout_ready || !rfm_population_try_read_lock(view_p)) {
	if (view_p->flags.type != DESKVIEW_TYPE){
	    // Since the expose could not proceed, keep generating
	    // expose signal until it works. Expose signal is generated
	    // in its own thread to avoid any deadlock potential.

	    // Clean the area anyways, for that we don't need a readlock.
	    cairo_t *gdk_context =  
		gdk_cairo_create(gtk_widget_get_window(view_p->widgets.paper));
	    rodent_clean_rectangle (view_p, 
		    event->area.x, event->area.y, 
		    event->area.width, event->area.height,
		    gdk_context);
		cairo_destroy(gdk_context);
	}

	TRACE("rodent_expose(): dumping an expose signal now (write lock is on)\n");
	void **arg = (void **)malloc(sizeof(void *));
	if (!arg) g_error("malloc: %s\n", strerror(errno));
	GdkRectangle *area_p = (GdkRectangle *)malloc(sizeof(GdkRectangle));
	if (!area_p) g_error("malloc: %s\n", strerror(errno));
	memcpy(area_p, &(event->area), sizeof(GdkRectangle));
	arg[0] = view_p;
	arg[1] = area_p;
	rfm_view_thread_create(view_p, queue_an_expose, arg, "queue_an_expose");
	return;
    }
#endif

    // check for background pixbuf (deskview)
    // this check will create background image when changed
    background_test(view_p);

    NOOP (stderr, " on_expose area= %d, %d width=%d, height=%d\n", 
	    event->area.x, event->area.y, event->area.width, event->area.height);   
    // Find items which need to be redrawn.
    // take into consideration margin displacement

    GSList *render_list=rodent_find_items_in_rectangle(view_p, &(event->area), FALSE);

    // render_list may be NULL if the area to render does not
    // have any icons.
    GdkRectangle rect;
    rect.x = event->area.x; 
    rect.y = event->area.y; 
    rect.width = event->area.width;
    rect.height = event->area.height;
    rect.x=get_minimum_x(view_p, render_list, event->area.x);
    rect.y=get_minimum_y(view_p, render_list, event->area.y);
    rect.width = event->area.width + (event->area.x - rect.x);
    rect.height = event->area.height + (event->area.y - rect.y);

    NOOP("*** rect= %d,%d width=%d, height=%d\n", rect.x, rect.y, rect.width, rect.height);
    expose_slist_t *expose_slist_p = 
	mk_expose_slist_p(view_p, render_list, &rect);
    expose_slist_f(expose_slist_p);
    // Remove readlock now.
    rfm_population_read_unlock (view_p);
    // Thread cleanup.
    free_expose_slist_p(expose_slist_p);
    //expose_slist(view_p, render_list, &rect);
}

#else

// This function is a callback with an implied GDK mutex.
void
rodent_draw (GtkWidget * widget, cairo_t *cr, gpointer data) {
    g_mutex_lock(rfm_global_p->status_mutex);
    gint status = rfm_global_p->status;
    g_mutex_unlock(rfm_global_p->status_mutex);
    if (status == STATUS_EXIT) return;

    view_t *view_p = (view_t *) data;
    if(!view_p || !view_p->widgets.paper)  return;
    g_mutex_lock(view_p->mutexes.status_mutex);
    status = view_p->flags.status;
    g_mutex_unlock(view_p->mutexes.status_mutex);
    if (status == STATUS_EXIT) return;
    gboolean layout_ready = rfm_layout_is_setup(view_p);

    NOOP( "rodent_draw...\n");
    
    double x1, y1, x2, y2;
    cairo_clip_extents (cr, &x1, &y1, &x2, &y2);
    GdkRectangle area;
    area.x = x1;
    area.y = y1;
    area.width = x2 - x1;
    area.height = y2 - y1;

    NOOP ("rodent_draw(): x,y=%lf,%lf x2,y2=%lf,%lf\n", 
	    x1, y1, x2, y2);
    NOOP (stderr, "rodent_draw(): area x,y=%d,%d w,h=%d,%d\n", area.x, area.y, area.width, area.height);
    
#if 10
    // Put a read lock on population.
    // This presents a race condition problem:
    //   if we lose the race, no expose gets done (this happens)
    //   if we do a lock-wait instead, deadlock occurs...
    //
    if (noexpose || !layout_ready || !rfm_population_try_read_lock(view_p)) {
	NOOP(stderr, ">>><<< rodent_draw(): dumping an expose signal now\n");
	if (view_p->flags.type != DESKVIEW_TYPE){
	    // Clean the area anyways, for that we don't need a readlock.
	    cairo_t *gdk_context =  
		gdk_cairo_create(gtk_widget_get_window(view_p->widgets.paper)); 

	    rodent_clean_rectangle (view_p, 
		    area.x, area.y, 
		    area.width, area.height,
		    gdk_context);
		cairo_destroy(gdk_context);
	}
	void **arg = (void **)malloc(2*sizeof(void *));
	if (!arg) g_error("malloc: %s\n", strerror(errno));
	void *area_p = malloc(sizeof(GdkRectangle));
	if (!area_p) g_error("malloc: %s\n", strerror(errno));
	memcpy(area_p, &area, sizeof(GdkRectangle));
	arg[0] = view_p;
	arg[1] = area_p;
	rfm_view_thread_create(view_p, queue_an_expose, arg, "queue_an_expose");
	return;
    }
#endif
    // check for background pixbuf (deskview)
    // this check will create background image when changed
    background_test(view_p);

    NOOP(stderr, "on_expose area= %d, %d width=%d, height=%d\n", 
	    area.x, area.y, area.width, area.height);   
    // Find items which need to be redrawn.
    // take into consideration margin displacement

    GSList *render_list=rodent_find_items_in_rectangle(view_p, &(area), FALSE);

    // render_list may be NULL if the area to render does not
    // have any icons.
    GdkRectangle rect;
    rect.x = area.x; 
    rect.y = area.y; 
    rect.width = area.width;
    rect.height = area.height;
    rect.x=get_minimum_x(view_p, render_list, area.x);
    rect.y=get_minimum_y(view_p, render_list, area.y);
    rect.width = area.width + (area.x - rect.x);
    rect.height = area.height + (area.y - rect.y);

    NOOP(stderr, "*** rect= %d,%d width=%d, height=%d\n", rect.x, rect.y, rect.width, rect.height);
    expose_slist_t *expose_slist_p = 
	mk_expose_slist_p(view_p, render_list, &rect);
    expose_slist_f(expose_slist_p);
    // Remove readlock now.
    rfm_population_read_unlock (view_p);
    // Thread cleanup.
    free_expose_slist_p(expose_slist_p);
    //expose_slist(view_p, render_list, &rect);
}

#endif


void
rodent_recalc_population_geometry (view_t * view_p) {
   if (!view_p || view_p->population_pp==NULL || *view_p->population_pp==NULL || MAX_ELEMENTS(view_p) == 0){
       NOOP("rodent_recalc_population_geometry: skipping ...\n");
       return;
   }
    gint grid_rows = GRID_ROWS(view_p);
    gint grid_columns = GRID_COLUMNS(view_p);
   NOOP("rodent_recalc_population_geometry(): grid_columns=%d grid_rows=%d max_elements=%d\n", 
	    grid_columns, grid_rows, MAX_ELEMENTS(view_p));

   gint row;
   gint column;
   gint max_elements = MAX_ELEMENTS(view_p);
   for(column = 0; column < grid_columns; column++) {
       for(row = 0; row < grid_rows; row++) {
	   gint element;
	   if (view_p->flags.type == ICONVIEW_TYPE){
	       element = row * grid_columns + column;
	   } else { //transposed...
		element = column * grid_rows + row;
	   }
	   if (element >= max_elements) break;
	   if (view_p->population_pp[element] == NULL) break;
	   // if element is repeated as first column of next row, skip it 
	   if(element == (row + 1) * grid_columns){
	       NOOP("skipping %s\n",view_p->population_pp[element]->en->path);
	       continue;
	   }                
	   view_p->population_pp[element]->column = column;
	   view_p->population_pp[element]->row = row;

       }
   }
   return;
}

void  
rodent_redraw_item(view_t *view_p, const population_t *population_p){
    widgets_t *widgets_p = &(view_p->widgets);
    cairo_t *gdk_context = gdk_cairo_create(gtk_widget_get_window(widgets_p->paper));
    render_item(view_p, population_p, gdk_context);
	cairo_destroy(gdk_context);
    gdk_flush();
    return ;
}

static void *
redraw_items_f(void *data){
    view_t *view_p = data;
    gint grid_columns = GRID_COLUMNS(view_p);
    gint grid_rows = GRID_ROWS(view_p);
    gint low_column = 0;
    gint low_row = 0;
    gint cellheight = CELLHEIGHT(view_p);
    gint cellwidth = CELLWIDTH(view_p);
    if (view_p->flags.type == ICONVIEW_TYPE) {
	double pos = gtk_adjustment_get_value (
		gtk_scrolled_window_get_vadjustment(view_p->widgets.scrolled_window));
	double r = floor(pos / cellheight);
	low_row = r;
    }
    gdouble c = rfm_global_p->allocation->height;
    c = ceil(c / cellheight);
    gint high_row = low_row + c; 

    gint column;
    gint max_elements = MAX_ELEMENTS(view_p);
    population_t **population_pp = view_p->population_pp;
    for (column=low_column; column < grid_columns; column++){
	gint row;
	for (row=low_row; row < high_row; row++) {
	    gint element;
	    if (view_p->flags.type == ICONVIEW_TYPE) {
		element = row * grid_columns + column;
	    } else { // transposed
		element = column * grid_rows + row;
	    }
	    if (element >= max_elements) {
		GdkRectangle rect;
		rect.x = column * cellwidth;
		rect.y = row * cellheight;
		rect.width = cellwidth;
		rect.height = cellheight;
		rfm_expose_rect(view_p, &rect);
		// This may occur when bottom row is not full.
		continue;
	    }
	    rodent_redraw_item(view_p, population_pp[element]);
	}
    }
    return NULL;
}

void 
rodent_redraw_items(view_t *view_p){
    while (!rfm_population_read_lock (view_p)) rfm_threadwait();
    rfm_context_function(redraw_items_f, view_p);
    rfm_population_read_unlock (view_p);
}

