#ifdef COPYRIGHT_INFORMATION
#include "gplv3.h"
#endif
/*
 * Copyright (C) 2002-2012 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
   
static gboolean
environment_condition(const gchar *environment_string, const gchar *current_value){
    const gchar *environment_value=getenv (environment_string);
    if (!environment_value && !current_value) return FALSE;
    else if (!environment_value && current_value) return TRUE;
    else if (environment_value && !current_value) return TRUE;
    else if (strcmp(environment_value, current_value))  return TRUE;
    return FALSE;
}
   
static gchar *
environment_dup(const gchar *environment_string){
    if (getenv(environment_string)) return g_strdup(getenv(environment_string));
    return NULL;
}

// XXX background image works for the iconview too.
//     Just create a different environment variable
//     to do configuration separate from deskview...
static gboolean
background_test(view_t *view_p) {
    gboolean background_condition=(view_p->flags.type == ICONVIEW_TYPE)? FALSE:
	environment_condition("RFM_DESKTOP_IMAGE",view_p->eyecandy.desktop_bg_file);
    gboolean deskcolor_condition=(view_p->flags.type == ICONVIEW_TYPE)? FALSE:
	environment_condition("RFM_DESKTOP_COLOR",view_p->eyecandy.desktop_color);
    gboolean gridcolor_condition=(view_p->flags.type == DESKVIEW_TYPE)? FALSE:
	environment_condition("RFM_ICONVIEW_COLOR",view_p->eyecandy.iconview_color);
	
    if (deskcolor_condition &&
	    getenv("RFM_DESKTOP_IMAGE") &&
	    strlen(getenv("RFM_DESKTOP_IMAGE"))) background_condition = TRUE;

    /*if (background_condition){
	TRACE( "background_condition = %s\n",getenv("RFM_DESKTOP_IMAGE"));
	if  (getenv("RFM_DESKTOP_IMAGE") && strlen(getenv("RFM_DESKTOP_IMAGE"))) {
	    g_free(view_p->eyecandy.desktop_bg_file);
	    view_p->eyecandy.desktop_bg_file = g_strdup(getenv("RFM_DESKTOP_IMAGE"));
	}
    }*/
    if (deskcolor_condition){
	g_free (view_p->eyecandy.desktop_color);
	view_p->eyecandy.desktop_color = environment_dup("RFM_DESKTOP_COLOR");
    }
    if (gridcolor_condition){
	g_free (view_p->eyecandy.iconview_color);
	view_p->eyecandy.iconview_color = environment_dup("RFM_ICONVIEW_COLOR");
    }
    return (gridcolor_condition | deskcolor_condition | background_condition);
}

void
rodent_clean_rectangle (view_t * view_p, int x, int y, int w, int h, cairo_t *gdk_context) {
    if (w == 0 || h == 0) return;
    NOOP(stderr, "rodent_clean_rectangle(%d, %d, %d, %d)\n", x, y, w, h);

    if (view_p->eyecandy.bg_surface) { // background surface is available
	cairo_set_source_surface (gdk_context, view_p->eyecandy.bg_surface, 0,0);
	cairo_rectangle(gdk_context, x, y, w, h);
	cairo_fill(gdk_context);
    } else {
	// select background color
	GdkColor *color=rfm_get_gdk_color (view_p, BACKGROUND_COLOR);


	cairo_set_source_rgb(gdk_context, 
		(double)(color->red)/0xffff,
		(double)(color->green)/0xffff, 
		(double)(color->blue)/0xffff);
	g_free(color);
	cairo_rectangle(gdk_context, x, y, w, h);
	cairo_fill(gdk_context);
    }
}


#define RED 0
#define GREEN 1
#define CYAN 2
#define MAGENTA 3
#define GREY 4
#define BGCOLOR 5
#define RUBBERBAND_COLOR 6
#define RUBBERBAND_COLOR_DEFINITION 0.35, 0.65, 0.65

static void 
saturated_rectangle(view_t * view_p, gint x, gint y, gint width, gint height, cairo_t *gdk_context, gint color) {
	// saturated rectangle (thin margin)
	gboolean is_dark_bg=rfm_is_dark_background(view_p);
	
	switch (color){
	    case RUBBERBAND_COLOR:
		// No outline here... (use same color as fill color)
		cairo_set_source_rgb(gdk_context, RUBBERBAND_COLOR_DEFINITION);
		break;
	    case BGCOLOR:
		// No outline here... (use same color as fill color)
		if (is_dark_bg){
		    cairo_set_source_rgb(gdk_context, 0.05, 0.05, 0.05);
		} else {
		    cairo_set_source_rgb(gdk_context, 0.95, 0.95, 0.95);
		}
		break;
	    case RED:
		// No outline here... (use same color as fill color)
		if (is_dark_bg){
		    cairo_set_source_rgb(gdk_context, 0.950, 0.30, 0.30);
		} else {
		    cairo_set_source_rgb(gdk_context, 0.850, 0.0, 0.0);
		}
		break;
	    case GREEN:
		cairo_set_source_rgb(gdk_context, 0.0, 1.0, 0.0);
		break;
	    case CYAN:
		cairo_set_source_rgb(gdk_context, 0.0, 1.0, 1.0);
		break;
	    case MAGENTA:
		cairo_set_source_rgb(gdk_context, 1.0, 0.0, 1.0);
	    case GREY:
		cairo_set_source_rgb(gdk_context, 0.30, 0.30, 0.30);
		break;
	}
	cairo_rectangle(gdk_context, x, y, width, height);
	cairo_fill(gdk_context);
	
	rodent_clean_rectangle (view_p, x + 1, y + 1, width-2, height-2, gdk_context);
	if (color == RED) {
	    if (is_dark_bg){
		cairo_set_source_rgb(gdk_context, 0.75, 0.75, 0.95);
	    } else {
		cairo_set_source_rgb(gdk_context, 0.3, 0.3, 0.7);
	    }
	} else if (color == BGCOLOR) {
	    if (is_dark_bg){
		cairo_set_source_rgb(gdk_context, 0.95, 0.95, 0.95);
	    } else {
		cairo_set_source_rgb(gdk_context, 0.05, 0.05, 0.05);
	    }
	}  else if (color == RUBBERBAND_COLOR) {
	    // lighter blue background 
	    cairo_set_source_rgb(gdk_context, RUBBERBAND_COLOR_DEFINITION);
	} else {
	    // blue background 
	    cairo_set_source_rgb(gdk_context, 0.45, 0.65, 0.95);
	}
//	cairo_set_source_rgba(gdk_context, 0.0, 0.25, 0.85, 0.45);
	cairo_rectangle(gdk_context,  x + 1, y + 1, width-2, height-2);
	cairo_fill(gdk_context);
}

#define DETAILS_EXPOSE 0x01
#define COMPACT_EXPOSE 0x02
#define NORMAL_EXPOSE 0x04
#define SATURATED_LABEL_EXPOSE 0x08
#define SATURATED_ICON_EXPOSE 0x010
#define SATURATED_EXPOSE (SATURATED_LABEL_EXPOSE | SATURATED_ICON_EXPOSE)
#define SELECTED_EXPOSE 0x020
#define RUBBERBAND_EXPOSE 0x040
#define DARK_BACKGROUND_EXPOSE 0x080

static void
erase_text (view_t * view_p, const population_t * population_p, cairo_t *gdk_context, gint expose_type) {
    GdkRectangle label_rect;
    if (!rfm_get_population_label_rect(view_p, population_p, &label_rect)) return;

    GdkRectangle item_rect;
    if (!rfm_get_population_rect(view_p, population_p, &item_rect)) return;

    gint cellwidth = CELLWIDTH(view_p);
    gint cellheight = CELLHEIGHT(view_p);
    gint width = cellwidth;
    //gint height = label_rect.height;


    // Clean rectangle sets up the clip area for expose event...
    gint icon_size = ICON_SIZE(view_p);
    rodent_clean_rectangle (view_p, 
		item_rect.x,//label_rect.x, 
		label_rect.y,
		width,
		cellheight - icon_size - TEXTSPACING, // full_height
		gdk_context);

    if ((expose_type & SELECTED_EXPOSE) && (expose_type & RUBBERBAND_EXPOSE)) {
	saturated_rectangle(view_p, 
		item_rect.x, 
		label_rect.y,
		cellwidth, 
		cellheight - icon_size - TEXTSPACING, // full_height
		gdk_context, RUBBERBAND_COLOR);
    } else if (expose_type & SATURATED_LABEL_EXPOSE) {
	GdkRectangle rect;
	if (rfm_get_population_label_rect_full(view_p, population_p, &rect)){
	    rodent_clean_rectangle (view_p, 
		label_rect.x, 
		label_rect.y,
		label_rect.width,
		label_rect.height,
		gdk_context);
	    saturated_rectangle(view_p, 
		    rect.x, 
		    rect.y,
		    rect.width, 
		    rect.height, 
		    gdk_context, RED);
	}
    }
}


static void
draw_mask(cairo_t *gdk_context, PangoLayout 	*layout, gint X, gint Y, gint expose_type) {
    if (expose_type & DARK_BACKGROUND_EXPOSE) {
	cairo_set_source_rgb(gdk_context, 0.0, 0.0, 0.0);
    } else {
	cairo_set_source_rgb(gdk_context, 0.99, 0.99, 0.99);
    }
    cairo_move_to (gdk_context,  X+1, Y+1);
    pango_cairo_show_layout (gdk_context, layout);
    cairo_move_to (gdk_context,  X-1, Y-1);
    pango_cairo_show_layout (gdk_context, layout);
    cairo_move_to (gdk_context,  X+1, Y-1);
    pango_cairo_show_layout (gdk_context, layout);
    cairo_move_to (gdk_context,  X-1, Y+1);
    pango_cairo_show_layout (gdk_context, layout);
    cairo_move_to (gdk_context,  X-1, Y);
    pango_cairo_show_layout (gdk_context, layout);
    cairo_move_to (gdk_context,  X+1, Y);
    pango_cairo_show_layout (gdk_context, layout);
    cairo_move_to (gdk_context,  X, Y+1);
    pango_cairo_show_layout (gdk_context, layout);
    cairo_move_to (gdk_context,  X, Y-1);
    pango_cairo_show_layout (gdk_context, layout);
}

static void
redraw_text (view_t * view_p, const population_t * population_p, cairo_t *gdk_context) {

    // Figure out what we are dealing with from the beginning...
    gint expose_type = 0;
    gint icon_size = ICON_SIZE(view_p);
    // Three basic types of view: details, compact and normal
    if (rfm_layout_is_details_size(view_p)){
	expose_type |= DETAILS_EXPOSE;
    } else if (icon_size == TINY_ICON_SIZE){
	expose_type |= COMPACT_EXPOSE;
    } else {
	expose_type |= NORMAL_EXPOSE;
    }
    if (population_p->flags & POPULATION_SATURATED) {
	expose_type |= SATURATED_ICON_EXPOSE;
    } else if (population_p->flags & LABEL_SATURATED) {
	expose_type |= SATURATED_LABEL_EXPOSE;
    } else if (population_p->flags & POPULATION_SELECTED) {
	expose_type |= SELECTED_EXPOSE;
    }
    if (view_p->mouse_event.rubberbanding) {
	expose_type |= RUBBERBAND_EXPOSE;
    }
    if (rfm_is_dark_background(view_p)) {
	expose_type |= DARK_BACKGROUND_EXPOSE;
    }

    erase_text (view_p, population_p, gdk_context, expose_type); 

   
    GdkRectangle label_rect;
    if (!rfm_get_population_label_rect(view_p, population_p, &label_rect)) return;

    // redraw_text_outline
    // (not for saturated labels...)
    // This only applies when we have a background image, really.
    gboolean do_outline = (view_p->flags.type == DESKVIEW_TYPE &&
	    getenv ("RFM_DESKTOP_IMAGE") && 
	    strlen (getenv ("RFM_DESKTOP_IMAGE")) );
    if (do_outline && !(expose_type & SATURATED_EXPOSE)) {
	GdkColor *color=rfm_get_gdk_color (view_p, BACKGROUND_COLOR);
	cairo_set_source_rgb(gdk_context, 
		(double)(color->red)/0xffff, 
		(double)(color->green)/0xffff, 
		(double)(color->blue)/0xffff);
	g_free(color);    
	gint i, j;
	
	gint x = label_rect.x;
	gint y = label_rect.y;
	NOOP("outline for %s\n", population_p->en->path);
	for(i = x; i < x + 3; i++) {
	    for(j = y; j < y + 3; j++) {
		//if(i == x && j == y) continue;
		cairo_move_to (gdk_context, i,j);
		pango_cairo_show_layout (gdk_context, population_p->layout);
		if(population_p->layout2) {
		    gint X=(icon_size >= SMALL_ICON_SIZE)?
			    i:
			    i+population_p->logical_rect.width +2;
		    gint Y=(icon_size >= SMALL_ICON_SIZE)?
			    j + population_p->logical_rect.height:
			    j;
		    cairo_move_to (gdk_context, X, Y);
		    pango_cairo_show_layout (gdk_context, population_p->layout2);
		}
	    }
	}
    }

    // redraw layouts;
    gint X = label_rect.x + 1;
    gint Y = label_rect.y + 1;

    // Redefine text color.
    if (expose_type & SATURATED_EXPOSE) {
	if (expose_type & DARK_BACKGROUND_EXPOSE) {
	    cairo_set_source_rgb(gdk_context, 0.05, 0.05, 0.05);
	} else {
	    cairo_set_source_rgb(gdk_context, 0.95, 0.95, 0.95);
	}
    } else {
	if ((expose_type & SELECTED_EXPOSE) && (expose_type & RUBBERBAND_EXPOSE)){
	    cairo_set_source_rgb(gdk_context, 0.95, 0.95, 0.05);
	} else { 
	    GdkColor *color=
		    rfm_get_gdk_color (view_p, rodent_text_color(population_p));
	    cairo_set_source_rgb(gdk_context, 
		    (double)(color->red)/0xffff, 
		    (double)(color->green)/0xffff,
		    (double)(color->blue)/0xffff);
	    g_free(color);
	}
    }


#define DARK_SAT_COLOR 0.95, 0.95, 0.95
#define LIGHT_SAT_COLOR  0.05, 0.05, 0.05
    if(expose_type & SATURATED_ICON_EXPOSE){
	if (expose_type & SATURATED_LABEL_EXPOSE && !(expose_type & (DETAILS_EXPOSE | COMPACT_EXPOSE))) {
		if(population_p->layout_full)
		    draw_mask(gdk_context,  population_p->layout_full, X, Y, expose_type);
	} else {
		if(population_p->layout)
		    draw_mask(gdk_context,  population_p->layout, X, Y, expose_type);
	}
	if (expose_type & DARK_BACKGROUND_EXPOSE) {
	    cairo_set_source_rgb(gdk_context, DARK_SAT_COLOR);
	} else {
	    cairo_set_source_rgb(gdk_context, LIGHT_SAT_COLOR);
	}
    }

    if (expose_type & SATURATED_LABEL_EXPOSE){
	GdkRectangle rect;
	if (rfm_get_population_label_rect_full(view_p, population_p, &rect)){
	    cairo_move_to (gdk_context, rect.x + 1 , rect.y + 1);
	    pango_cairo_show_layout (gdk_context, population_p->layout_full);
	}
	
    } else {
	cairo_move_to (gdk_context,  X, Y);
	pango_cairo_show_layout (gdk_context, population_p->layout);
    }
 
 /*   if (expose_type & SATURATED_LABEL_EXPOSE && !(expose_type & (DETAILS_EXPOSE | COMPACT_EXPOSE))) {
	if(population_p->layout_full){
	    pango_cairo_show_layout (gdk_context, population_p->layout_full);
	}
    } else {
	if(population_p->layout){
	    pango_cairo_show_layout (gdk_context, population_p->layout);
	}
    }*/

    // Figure out layout2 position
    if (population_p->layout2 && !(expose_type & COMPACT_EXPOSE)){
	if(expose_type & DETAILS_EXPOSE){
	    gint max_tiny_width = NAME_WIDTH(view_p);
	    NOOP(stderr, "name_width=%d\n", max_tiny_width);
	    X = label_rect.x + max_tiny_width +1;
	    Y = label_rect.y + 1;
	    if(expose_type & SATURATED_EXPOSE){
		// mask
		draw_mask(gdk_context,  population_p->layout2, X, Y, expose_type);
		if (expose_type & DARK_BACKGROUND_EXPOSE) {
		    cairo_set_source_rgb(gdk_context, 0.95, 0.95, 0.0);
		} else {
		    cairo_set_source_rgb(gdk_context, 0.0, 0.0, 0.95);
		}
	    }
	} else if (!(expose_type & SATURATED_LABEL_EXPOSE)){
	    X = label_rect.x + 1;
	    Y = label_rect.y + population_p->logical_rect.height + 1;
	    if (expose_type & SATURATED_ICON_EXPOSE){
		draw_mask(gdk_context,  population_p->layout2, X, Y, expose_type);
		if (expose_type & DARK_BACKGROUND_EXPOSE) {
		    cairo_set_source_rgb(gdk_context, DARK_SAT_COLOR);
		} else {
		    cairo_set_source_rgb(gdk_context, LIGHT_SAT_COLOR);
		}
	    }
	}
	if (!(expose_type & DETAILS_EXPOSE) && (expose_type & SELECTED_EXPOSE) 
		&& (expose_type & RUBBERBAND_EXPOSE)){
	    cairo_set_source_rgb(gdk_context, 0.95, 0.95, 0.05);
	} 	
	cairo_move_to (gdk_context,  X, Y);
	if (expose_type & DETAILS_EXPOSE || !(expose_type & SATURATED_LABEL_EXPOSE)){
	    NOOP(stderr, "showing layout2 at %d,%d\n", X, Y);
	    pango_cairo_show_layout (gdk_context, population_p->layout2);
	}
    }
}

static void
insert_layout (view_t * view_p, const population_t * population_p, cairo_t *context) {
    if (!view_p){
	g_warning("insert_layout(): !view_p\n");
	return;
    }
    if (!population_p){
	//g_warning("insert_layout(): !population_p\n");
	return;
    }
    redraw_text (view_p, population_p, context);

}

// readlock should be enabled before entering this function...
static void
insert_pixbuf (view_t * view_p, 
	const population_t * population_p,
	cairo_t *gdk_context) {
    if (!view_p){
	g_warning("insert_pixbuf(): !view_p\n");
	return;
    }
    if (!population_p){
	//g_warning("insert_pixbuf(): !population_p\n");
	return;
    }
    if (!population_p->pixbuf) {
	g_warning("insert_pixbuf(): !population_p->pixbuf\n");
	return;
    }
	
    GdkRectangle item_rect;
    if (!rfm_get_population_rect(view_p, population_p, &item_rect)) {
	g_warning("insert_pixbuf(): !rfm_get_population_rect(view_p, population_p, &item_rect)\n");
	return;
    }

    GdkRectangle icon_rect;
    if (!rfm_get_population_icon_rect(view_p, population_p, &icon_rect)){
	g_warning("insert_pixbuf(): !rfm_get_population_icon_rect(view_p, population_p, &icon_rect)\n");
	return;
    }
	    
    gint icon_size = ICON_SIZE(view_p);
    rodent_clean_rectangle (view_p,
		item_rect.x, 
		item_rect.y,
		item_rect.width,
		icon_size+TEXTSPACING, 
		gdk_context);

    gint cellwidth = CELLWIDTH(view_p);
    gint cellheight = CELLHEIGHT(view_p);
    gint x = (icon_size >= SMALL_ICON_SIZE)?
	    item_rect.x + ((cellwidth - icon_size) / 2):
	    item_rect.x;
 
    // saturated item background 
    // (the color is for the thin margin around the rectangle
    //  and will also determine the rectangle fill color)
    if(population_p->flags & POPULATION_SATURATED) {
		saturated_rectangle(view_p, 
			x,
			item_rect.y,
			icon_size + 2, 
			icon_size + 2, gdk_context, 
			RED);
     } else if (population_p->flags & POPULATION_SELECTED) {
	if(view_p->mouse_event.rubberbanding) {
	        // rubberband box
		saturated_rectangle(view_p, 
			item_rect.x,
			item_rect.y,
			cellwidth, 
			cellheight, 
			gdk_context, 
			RUBBERBAND_COLOR);
	} else {
		// simply select
		saturated_rectangle(view_p, 
			x,
			item_rect.y,
			icon_size + 2, 
			icon_size + 2, gdk_context, 
			MAGENTA);
	}
     } else {
	    rodent_clean_rectangle (view_p,
		icon_rect.x, 
		icon_rect.y,
		icon_rect.width + 5,
		icon_rect.height + 5,
		gdk_context);
    }
    // Special stuff for thumbnails (but not for folders)
    if(population_p->en && !IS_SDIR(population_p->en->type) && 
	    (population_p->flags & POPULATION_THUMBNAILED) ) {
	NOOP ("MIME: %s is image type\n", population_p->en->mimetype);
	// cute shadow
	if (!(population_p->flags & POPULATION_SATURATED) &&
	    !(population_p->flags & POPULATION_SELECTED)
	    ) {
	    cairo_set_source_rgba(gdk_context, 0.0, 0.0, 0.0, 0.35);
	    cairo_rectangle(gdk_context, 
		icon_rect.x+2,
		icon_rect.y+2,
		icon_rect.width + 3,
		icon_rect.height + 3);
	    cairo_fill(gdk_context);
	}
	// black frame
	cairo_set_source_rgb(gdk_context, 0.0, 0.0, 0.0);
	
	cairo_rectangle(gdk_context, 
		icon_rect.x,
		icon_rect.y,
		icon_rect.width + 2,
		icon_rect.height + 2);
	cairo_fill(gdk_context);

	// fill in with white color, for the benefit of transparent previews.
	if (population_p->flags & POPULATION_SELECTED) {
	    cairo_set_source_rgba(gdk_context, 1.0, 1.0, 1.0, 0.450);
	} else {
	    cairo_set_source_rgb(gdk_context, 1.0, 1.0, 1.0);
	}
	cairo_rectangle(gdk_context, 
		icon_rect.x+1,
		icon_rect.y+1,
		icon_rect.width,
		icon_rect.height);
	cairo_fill(gdk_context);
	
    }
    // icon pixbuf
    NOOP("---> inserting %s at %d,%d\n",
	    population_p->en->path, icon_rect.x+1, icon_rect.y+1);


    if (GDK_IS_PIXBUF(population_p->pixbuf)){
		gdk_cairo_set_source_pixbuf(gdk_context, 
				population_p->pixbuf,
				icon_rect.x+1,
				icon_rect.y+1);
		if(population_p->flags & POPULATION_SELECTED) {
			cairo_paint_with_alpha(gdk_context, 0.450);
		} else {
			cairo_paint (gdk_context); 
		}
    }
    
    // Entry dependent pixbuf modifiers
    if (!population_p->en) goto done;

#if 10
    GdkPixbuf *emblem_pixbuf=NULL;
    // Emblems...
    // Module emblems
    if (POPULATION_MODULE(population_p)){
	// ask POPULATION_MODULE for emblem
	const gchar *emblem=(const gchar *)
	    rfm_natural(PLUGIN_DIR, POPULATION_MODULE(population_p), 
		    population_p->en, "get_emblem");
	if (emblem) {
	  emblem_pixbuf =
	    rfm_get_pixbuf (emblem, icon_size/1.8);
	  if (GDK_IS_PIXBUF(emblem_pixbuf)){
	    gdk_cairo_set_source_pixbuf(gdk_context, 
		emblem_pixbuf,
		icon_rect.x+1+icon_rect.width - icon_rect.width/1.8, 
		icon_rect.y+1);
//		icon_rect.y+1+icon_rect.height/1.8);
	    cairo_paint_with_alpha(gdk_context, 0.85);
	  }
	}
    }


    // Up type icon emblems...
    if (IS_UP_TYPE (population_p->en->type)){
	gboolean have_icons_plugin =
	    GPOINTER_TO_INT(rfm_void(MODULE_DIR, "icons", "module_active"));
	if (have_icons_plugin) {
	    if(view_p->flags.preferences & __SHOW_HIDDEN){
	      emblem_pixbuf = 
		  rfm_get_pixbuf ("xffm/emblem_show-hidden", icon_size/3);
	      if (GDK_IS_PIXBUF(emblem_pixbuf)){
		gdk_cairo_set_source_pixbuf(gdk_context, 
		  emblem_pixbuf,
		  icon_rect.x+1, 
		  icon_rect.y+1);
		cairo_paint_with_alpha(gdk_context, 0.85);
	      }
	    }
	    if(view_p->flags.preferences & __SHOW_IMAGES){
	      emblem_pixbuf = 
		  rfm_get_pixbuf ("xffm/emblem_show-previews", icon_size/6);
	      if (GDK_IS_PIXBUF(emblem_pixbuf)){
		  gdk_cairo_set_source_pixbuf(gdk_context, 
		    emblem_pixbuf,
		    icon_rect.x+1, 
		    icon_rect.y+1 + icon_rect.height - icon_rect.width/6);
		  cairo_paint_with_alpha(gdk_context, 0.85);
	      }
	    }
	}

    }


    // Readonly emblem
    gboolean no_write_emblem=FALSE;
    if (IS_LOCAL_TYPE(population_p->en->type)) {
	no_write_emblem = 
	    !IS_UP_TYPE(population_p->en->type) &&
	    !IS_SDIR(population_p->en->type) &&
	    IS_NOWRITE_TYPE(population_p->en->type);
	   // !rfm_write_ok_path(population_p->en->path);
    }
    if (no_write_emblem){
	gint size = icon_size * 0.30;
	emblem_pixbuf =
	    rfm_get_pixbuf("xffm/emblem_readonly", size);	
	if (GDK_IS_PIXBUF(emblem_pixbuf)){
	    gdk_cairo_set_source_pixbuf(gdk_context,
		emblem_pixbuf,	
		icon_rect.x+1 + icon_rect.width - icon_rect.width/3, 
		icon_rect.y+1);
	    cairo_paint_with_alpha(gdk_context, 0.85);
	}
    }

    
    // Bookmark emblem
    if (!IS_UP_TYPE(population_p->en->type) &&
	    rodent_path_has_bookmark(population_p->en->path)){
	emblem_pixbuf =
	    rodent_get_bookmark_emblem(icon_size);
	if (GDK_IS_PIXBUF(emblem_pixbuf)){
	    gdk_cairo_set_source_pixbuf(gdk_context, 
		emblem_pixbuf,
		icon_rect.x+1 + icon_rect.width - icon_rect.width/3, 
		icon_rect.y+1);
	    cairo_paint_with_alpha(gdk_context, 0.85);
	}
    }

    // Pasteboard emblems...
    GdkPixbuf *pasteboard_pixbuf=NULL;
    // CUT emblem
    if (population_p->flags & POPULATION_CUT) {
	pasteboard_pixbuf = rfm_get_pixbuf ("xffm/stock_cut", icon_size/3);
    }
    // COPIED emblem
    else if (population_p->flags & POPULATION_COPIED) {
	pasteboard_pixbuf = rfm_get_pixbuf ("xffm/stock_copy", icon_size/3);
    }
    if (GDK_IS_PIXBUF(pasteboard_pixbuf)) {
	gdk_cairo_set_source_pixbuf(gdk_context, 
	    pasteboard_pixbuf,
	    icon_rect.x+1, 
	    icon_rect.y+1);
	cairo_paint_with_alpha(gdk_context, 0.85);
    }

    // Symlink emblem...
    if (IS_SLNK(population_p->en->type) && 
	    !IS_UP_TYPE(population_p->en->type)) {
	emblem_pixbuf =
	    rfm_get_pixbuf ("xffm/emblem_link", icon_size/2);
      if (GDK_IS_PIXBUF(emblem_pixbuf)) {
	int height = gdk_pixbuf_get_height (population_p->pixbuf) -
	    gdk_pixbuf_get_height (emblem_pixbuf);
	// "SW"

	gdk_cairo_set_source_pixbuf(gdk_context, 
	    emblem_pixbuf,
	    icon_rect.x+1, 
	    icon_rect.y+1+height);
	cairo_paint_with_alpha(gdk_context, 0.85);
      }

    }
#endif
    // Root emblem 
    if (!IS_UP_TYPE(population_p->en->type)
	    && strcmp(population_p->en->path, "/")==0)
    {
	gint size = icon_size * 0.50;
	emblem_pixbuf =
	    rfm_get_pixbuf("xffm/emblem_system", size);	
	if (GDK_IS_PIXBUF(emblem_pixbuf)){
	    gdk_cairo_set_source_pixbuf(gdk_context,
		emblem_pixbuf,	
		icon_rect.x+1 + icon_rect.width/2, 
		icon_rect.y+1 + icon_rect.height/2);
	    cairo_paint_with_alpha(gdk_context, 0.95);
	}
    }
    // Backup file emblem (synchronize with primary-icons.i)
    if (population_p->en->mimetype &&
	    strstr(population_p->en->mimetype, "x-trash"))
    {
	gint size = icon_size * 0.50;
	emblem_pixbuf =
	    rfm_get_pixbuf("application/x-trash", size);	
	if (GDK_IS_PIXBUF(emblem_pixbuf)){
	    gdk_cairo_set_source_pixbuf(gdk_context,
		emblem_pixbuf,	
		icon_rect.x+1 + icon_rect.width/2, 
		icon_rect.y+1 + icon_rect.height/2);
	    cairo_paint_with_alpha(gdk_context, 0.95);
	}
    }

done:
    return;
}

static gint
get_minimum(view_t *view_p, GSList *render_list, gboolean getX, gint min){
    GSList *tmp;
    gint cellwidth = CELLWIDTH(view_p);
    gint cellheight = CELLHEIGHT(view_p);
    for (tmp=render_list; tmp && tmp->data; tmp=tmp->next){
	population_t *population_p=tmp->data;
#if 0
	// hack test
	gboolean OK = FALSE;
	population_t **population_pp = view_p->population_pp;
	for (;population_pp && *population_pp; population_pp++){
	    if (population_p == *population_pp){
		OK = TRUE;
		break;
	    }
	}
	if (!OK) g_error("get_minimum(): invalid population_p!\n");
#endif
	gint c;
	if (getX) {
	    c=population_p->column * cellwidth + MARGIN_LEFT(view_p);
	} else {
	    c=population_p->row * cellheight + MARGIN_TOP(view_p);
	}
	if (c < min) min=c;

    }
    return min;
}

static gint
get_minimum_x(view_t *view_p, GSList *render_list, gint min){
    return get_minimum(view_p, render_list, TRUE, min);
}

static gint
get_minimum_y(view_t *view_p, GSList *render_list, gint min){
    return get_minimum(view_p, render_list, FALSE, min);
}

static void
render_item(view_t *view_p, const population_t *population_p, cairo_t *context){
    insert_pixbuf (view_p, population_p, context);
    insert_layout (view_p, population_p, context);
}

static
void
render_slist(view_t *view_p, GSList *render_list, cairo_t *context){
    GSList *slist;
    for (slist=render_list; slist && slist->data; slist=slist->next){
	population_t *population_p=slist->data;
	render_item(view_p, population_p, context);	
    }
}

typedef struct expose_slist_t {
    view_t *view_p;
    GSList *render_list;
    GdkRectangle area;
} expose_slist_t;

expose_slist_t *mk_expose_slist_p(view_t *view_p, GSList *render_list, GdkRectangle *area){
    expose_slist_t *expose_slist_p =
	(expose_slist_t *)malloc(sizeof(expose_slist_t));
    if (!expose_slist_p) g_error("malloc: %s", strerror(errno));
    memcpy(&(expose_slist_p->area), area, sizeof(GdkRectangle));
    expose_slist_p->view_p = view_p;
    expose_slist_p->render_list = render_list;
    return expose_slist_p;
}

void free_expose_slist_p(expose_slist_t *expose_slist_p) {
    if (!expose_slist_p) return;
    if (expose_slist_p->render_list) {
	g_slist_free(expose_slist_p->render_list);
    }
    g_free(expose_slist_p);

}



static
void *
expose_slist_f(gpointer data){
//expose_slist(view_t *view_p, GSList *render_list, GdkRectangle *area){
    expose_slist_t *expose_slist_p = data;
    view_t *view_p = expose_slist_p->view_p;
    GSList *render_list = expose_slist_p->render_list;
    GdkRectangle *area = &(expose_slist_p->area);

    if (!area){
	g_warning("area != NULL not met (expose_slist)");
	return NULL;
    }
    NOOP (stderr, "expose_slist_f(): rendering x:%d-%d  y:%d-%d\n",area->x,  area->x+area->width, area->y,  area->y+area->height);

    //
    // Apparently the flash issue does not occur on my 4 core, if there
    // is no background image (gridview)
    //
	NOOP(stderr, "rectangle= x: %d-%d y:%d-%d w:%d\n",
		 area->x, 
		 area->x+area->width, 
		 area->y, 
		 area->y+area->height,
		 area->width);
    

    cairo_t *gdk_context =  
	    gdk_cairo_create(gtk_widget_get_window(view_p->widgets.paper));

    
    // clean  expose area
    
    rodent_clean_rectangle (view_p,
		 area->x, 
		 area->y, 
		 area->width,
		 area->height,
		 gdk_context);


    // Render the affected items.
    // Only if something is listed, of course.
    if (render_list) {
	render_slist(view_p, render_list, gdk_context);
    }
	cairo_destroy(gdk_context);
    return NULL;

}


