/*
 * Copyright (C) 2002-2012 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rodent_libs.h"


static void
destroy_population_item (population_t * population_p);
static population_t *
create_population_t (view_t * view_p,
	record_entry_t * en, 
	gint element, 
	const gchar * icon_id, 
	const gchar * label);
static void *
thread_reload_view (view_t *view_p, xfdir_t *new_xfdir_p);
static void *
destroy_view(void *data);

#include "xfdir.i"
#include "rodent_population.i"
#include "rodent_population_threads.i"


gint
rodent_full_reload_view (view_t *view_p, record_entry_t * target_en) {
    NOOP(stderr, "rodent_full_reload_view --> reload_f thread\n");
    /*if (rfm_global_p->self != g_thread_self()) {
	g_warning("rodent_full_reload_view is a main thread callback! Call aborted\n");
	return 0;
    }*/
    widgets_t *widgets_p = &(view_p->widgets);

    print_loading_msg(widgets_p, target_en);
    
    reload_t *reload_p = (reload_t *)malloc(sizeof(reload_t));
    if (!reload_p) g_error("malloc: %s", strerror(errno));
    memset(reload_p, 0, sizeof(reload_t));
    reload_p->view_p = view_p;
    reload_p->target_en = target_en;
    
    NOOP(stderr, "starting load thread for: %s\n",
	    (target_en)? target_en->path: "rodent-root");
    rfm_view_thread_create(view_p, reload_f, reload_p, "reload_f");
    return 1;
}

gint
rodent_get_population_serial (view_t * view_p) {
    g_mutex_lock (view_p->mutexes.population_serial);
    gint population_serial = view_p->flags.population_serial;
    g_mutex_unlock (view_p->mutexes.population_serial);
    return population_serial;
}
void
rodent_update_cut_icons (view_t * view_p) {
    NOOP ("CUT:update_cut_status\n");
    if(!rfm_population_try_read_lock (view_p)) return;
    // find and update record
    population_t **tmp;
    for(tmp = view_p->population_pp; tmp && *tmp; tmp++) {
        population_t *population_p = *tmp;
        gboolean expose = FALSE;
        if(population_p->en && population_p->en->path) {
            int pasteboard_status = rfm_in_pasteboard (view_p, population_p->en);
            if(pasteboard_status) {
		if (pasteboard_status==2) { //cut
		    if (!(population_p->flags & POPULATION_CUT)) {
			expose = TRUE;
		    }
		    population_p->flags |= POPULATION_CUT;
		    population_p->flags &= (POPULATION_COPIED ^ 0xffffffff);

		} else {
		    if (!(population_p->flags & POPULATION_COPIED)) {
			expose = TRUE;
		    }
		    population_p->flags |= POPULATION_COPIED;
		    population_p->flags &= (POPULATION_CUT ^ 0xffffffff);
		}
                NOOP ("CUT: pasteboard: %s\n", population_p->en->path);
            } else {
                NOOP ("CUT: not in pasteboard: %s\n", population_p->en->path);
		if (population_p->flags & (POPULATION_COPIED|POPULATION_CUT)) {
		    expose = TRUE;
		}
		population_p->flags &= (POPULATION_COPIED ^ 0xffffffff);
		population_p->flags &= (POPULATION_CUT ^ 0xffffffff);
            }
            if(expose){
		rodent_redraw_item(view_p, population_p);
	    }
        }
    }

    // release population semaphore
    NOOP ("POP_SEM:  sem_post(&(view_p->population_sem))\n");
    rfm_population_read_unlock (view_p);
}



static void
start_element (GMarkupParseContext * context,
               const gchar * element_name,
               const gchar ** attribute_names, 
	       const gchar ** attribute_values, 
	       gpointer user_data, 
	       GError ** error) 
{
    if(strcmp (element_name, "keybind")) return;
    if (!attribute_names || !attribute_values) return;

    gchar **a = (gchar **)attribute_names;
    gchar **v = (gchar **)attribute_values;

    for(; a && *a && v && *v; a++, v++){
	if (strcasecmp(*a, "function_id")) continue;
	errno = 0;
	glong function_id = strtol(*v,NULL,0);
	if (errno) continue;
	gchar **aa = (gchar **)attribute_names;
	gchar **vv = (gchar **)attribute_values;
	const gchar *key = NULL;
	const gchar *mask = NULL;
	for(; aa && *aa && vv && *vv; aa++, vv++){
	    if (strcasecmp(*aa, "mask") && strcasecmp(*aa, "key")) continue;
	    if (strcasecmp(*aa, "mask")==0) mask = *vv; else key = *vv;
	}
	if (!key || !mask) continue;
 	rfm_complex(MODULE_DIR, "callbacks", 
		    GINT_TO_POINTER(function_id),
		    (void *)key, (void *)mask,
		    "set_menu_callback_keybind");
   }
    return;
}

static void
glib_parser (const gchar * mimefile) {
    GMarkupParseContext *context;
    GMarkupParser parser = {
        start_element,
        NULL,
        NULL,                   /*text_fun, */
        NULL,
        NULL
    };

    NOOP(stderr, "glib_parser(icon-module): parsing %s\n", mimefile);
    context = g_markup_parse_context_new (&parser, 0, NULL, NULL);
    FILE *f = fopen (mimefile, "r");
    if(!f) {
        g_warning ("cannot open %s\n", mimefile);
        return;
    }
    size_t l;
    gchar line[81];
    GError *error = NULL;
    while(!feof (f) && (l = fread (line, 1, 80, f)) != 0) {
        line[l] = 0;
        g_markup_parse_context_parse (context, line, l, &error);
    }
    fclose (f);
    g_markup_parse_context_free (context);
}


void
rodent_load_keybindings(void){
    gchar *keybindings_file = g_build_filename(KEYBINDINGS_FILE, NULL);
    if (!g_file_test(keybindings_file, G_FILE_TEST_EXISTS)) {
	g_free(keybindings_file);
	return;
    }
    static time_t mtime=0;
    struct stat st;
    stat(keybindings_file, &st);
    if (st.st_mtime != mtime){
	//reload
	glib_parser(keybindings_file);
	mtime = st.st_mtime;
    }

    g_free(keybindings_file);
    return;
}


void 
rodent_icontheme_test(void){
    // Icon theme switch:
    static gchar *use_gtk_theme_icons = NULL;
    if (!use_gtk_theme_icons){
	if (getenv("RFM_USE_GTK_ICON_THEME")){
	    use_gtk_theme_icons = 
	       g_strdup(getenv("RFM_USE_GTK_ICON_THEME"));
	} else {
	    use_gtk_theme_icons=g_strdup("");
	}
    }
    const gchar *instant_gtk_theme_icons =
	(getenv("RFM_USE_GTK_ICON_THEME"))?
	   getenv("RFM_USE_GTK_ICON_THEME"):"";

    if (strcmp(use_gtk_theme_icons,instant_gtk_theme_icons) != 0){
	g_free(use_gtk_theme_icons);
	use_gtk_theme_icons = g_strdup(instant_gtk_theme_icons);
	rfm_view_thread_create(NULL, update_pixbuf_hash_f, NULL, "update_pixbuf_hash_f");
    }
}

gchar *rodent_get_tabname(const gchar *path){

    gchar *basename = g_path_get_basename (path);
    gchar *dirname = g_path_get_dirname (path);
    gchar *parentname = g_path_get_basename (dirname);
    gchar *tabname;
    if(strcmp (parentname, ".") == 0 || strcmp (basename, "/") == 0) {
        tabname = g_strdup (basename);
    } else {
        tabname = g_build_filename (parentname, basename, NULL);
    }

    gchar *utf_tabname=rfm_utf_string(tabname);
    g_free(tabname);
    g_free (basename);
    g_free (dirname);
    g_free (parentname);
    return utf_tabname;
}

//*** end progress bar stuff ***/
static void *
set_view_icon_f (gpointer data) {
    view_t * view_p = data;

    NOOP (">> set_iconview_title\n");
    if (!rfm_main_context_check(view_p, TRUE)) return NULL;
    record_entry_t *en = view_p->en;
    gchar *icon_id = rfm_get_entry_icon_id(view_p, en, FALSE);

    if (en && rodent_path_has_bookmark(en->path)) {
	gchar *g = g_strconcat(icon_id, "/composite/places_user-bookmarks", NULL);
	g_free(icon_id);
	icon_id = g;
    } 
    
    NOOP("icon resolved as %s\n", icon_id);

    set_application_icon (&(view_p->widgets), icon_id);

    // Set tab icon
    GdkPixbuf *pb = rfm_get_pixbuf (icon_id, 20);
    g_free(icon_id);
    GtkWidget *image = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image);

    GList *children = gtk_container_get_children (GTK_CONTAINER (view_p->widgets.page_icon));
    if(children) {
        gtk_container_remove (GTK_CONTAINER (view_p->widgets.page_icon), GTK_WIDGET (children->data));
    }
    gtk_container_add (GTK_CONTAINER (view_p->widgets.page_icon), image);
    g_list_free (children);
    GtkImage *menu_image = g_object_get_data(G_OBJECT(view_p->widgets.page_label), "menu_image");
    if (menu_image) {
	gtk_image_set_from_pixbuf(menu_image, pb);
    }

    return NULL;
}

static void *
set_view_title_f (gpointer data) {
    view_t * view_p = data;

    NOOP (">> set_iconview_title\n");
    if (!rfm_main_context_check(view_p, TRUE)) return FALSE;
    record_entry_t *en = view_p->en;
    const gchar *path;

    if(en && en->path && strlen (en->path)) {
        if (en->pseudo_path) path = en->pseudo_path;
	else path = en->path;
    } else {
        path = g_get_host_name ();
    }

    if(view_p->module) {
        if(view_p->module && rfm_natural (PLUGIN_DIR, view_p->module, view_p->en, "window_title")) {
            path = rfm_natural (PLUGIN_DIR, view_p->module, view_p->en, "window_title");
        } else {
	    NOOP("module has no window_title function\n");
	}
    }
    set_icon_name (rfm_global_p->window, path);
    gchar *utf_tabname=rodent_get_tabname(path);

    gchar *chopped_path=g_strdup(path);
    rfm_chop_excess (chopped_path);
    gchar *utf_path=rfm_utf_string(chopped_path);
    g_free(chopped_path);

    NOOP(stderr, "setting bold title for %s\n", utf_tabname);
    gchar *markup = g_strconcat("<span weight=\"normal\">", utf_tabname, "</span>", NULL);
    gtk_label_set_markup (GTK_LABEL (view_p->widgets.page_label), markup);
    g_free (markup);
    GtkLabel *menu_label = g_object_get_data(G_OBJECT(view_p->widgets.page_label), "menu_label");
    if (menu_label) {
	markup = g_strconcat("<span weight=\"normal\">", utf_path, "</span>", NULL);
	gtk_label_set_markup(menu_label, utf_path);
	g_free (markup);
    }
 
    g_free (utf_tabname);

    return NULL;
}

void
rodent_set_view_title (view_t * view_p) {
    if(view_p->flags.type == DESKVIEW_TYPE) return;
    rfm_context_function(set_view_title_f, view_p);
    return;
}

void
rodent_set_view_icon (view_t * view_p) {
    if(view_p->flags.type == DESKVIEW_TYPE) return;
    rfm_context_function(set_view_icon_f, view_p);
    return;
}


gint rodent_trigger_reload(view_t *view_p){
    TRACE("rodent_trigger_reload()...\n");
    if (g_thread_self() == rfm_global_p->self) {
	record_entry_t *target_en = NULL;
	if (view_p->en) target_en = rfm_copy_entry(view_p->en);
	return (rodent_full_reload_view (view_p, target_en));
    }
    // Called from a non-main thread...
    return xfdir_monitor_control_greenlight(&(view_p->widgets));

}
gint
rodent_refresh (widgets_t * widgets_p, record_entry_t * en) {
    NOOP ("rodent_refresh(%s)...\n", (en) ? en->path : "en==NULL");
    if (en && en->st) NOOP ("en->st->st_uid  %d\n", en->st->st_uid );
    g_return_val_if_fail (widgets_p != NULL, 0);
    view_t *view_p = widgets_p->view_p;
    //view_p->redraw_pixmap = TRUE;
    gint retval;
    if(en) {
	if (en->path && g_path_is_absolute(en->path)){
	    gboolean exists;
	    if (IS_LOCAL_TYPE(en->type)) exists = g_file_test(en->path, G_FILE_TEST_IS_DIR);
	    else exists = rfm_g_file_test_with_wait(en->path, G_FILE_TEST_IS_DIR);
	    if (!exists) {
		rfm_show_text(widgets_p);
		rfm_diagnostics(widgets_p, "xffm/stock_dialog-error", NULL);
		rfm_diagnostics(widgets_p, "xffm_tag/stderr", en->path, ": ", 
			strerror(ENOENT), "\n",NULL);
		return 0;
	    }
	}
        UNSET_UP_TYPE (en->type);
        if (en->st && view_p->en && view_p->en->st) NOOP("source inode=%d target inode=%d\n",
	    (gint)view_p->en->st->st_ino,
	    (gint)en->st->st_ino);
    }
    // The purpose of the following lines is to check whether the target 
    // to refresh is the same as the current view, but path differs.
    // XXX Only for local files (modules may trip here until verified).
    if (!rfm_is_valid_view(view_p)) return FALSE;
    if (en && IS_LOCAL_TYPE(en->type) && !en->module){
	gboolean valid_entry = (en->st != NULL);
	gboolean valid_view = view_p->en && view_p->en->st;
	if (valid_view && valid_entry) {
	    gboolean same_inode = view_p->en->st->st_ino == en->st->st_ino &&
			       view_p->en->st->st_dev == en->st->st_dev;
	    gboolean redundant_condition= strcmp(view_p->en->path, en->path)!=0;
	    if (same_inode && redundant_condition){
		// this condition also is true if a reload is attempted while the
		// original load has not completed.
		//rfm_diagnostics(widgets_p, "xffm/stock_dialog-warning", NULL);
		//rfm_diagnostics(widgets_p, "xffm_tag/stderr", 
		//	strerror(EXDEV),"\n",NULL);
		g_warning("rodent_refresh cancelled upon redundant_condition\n");
		return 0;
	    }
	}
    }
    TRACE( "rodent_refresh calling rodent_full_reload_view()\n");
    retval = rodent_full_reload_view ((gpointer) view_p, en);

    return retval;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
///////////////////////////////////  xfdir  //////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

/**
 * SECTION:xfdir
 * @short_description: xfdir structure is a location to place the contents of 
 * the rodent view. The primary xfdir functions allow for the loading and
 * monitoring of local files on a POSIX compliant system.
 * 
 * @section_id:	xfdir
 * @stability: Current
 * @include: primary.h
 *
 **/


/**
 * xfdir_exit_monitor:
 * @xfdir_p:  pointer to an xfdir_t structure
 *
 * Schedule monitor to exit as soon as possible. This function should 
 * be accessed from threads other than the monitor thread. For the monitor thread,
 * use the static exit_monitor()  or rfm_increment_monitor() from outside.
 **/
void
xfdir_exit_monitor (view_t * view_p) {
    view_p->flags.monitor_id++;
    NOOP("xfdir monitor: xfdir_exit_monitor->%d\n", view_p->flags.monitor_id);
}


/**
 * xfdir_count_files:
 * @path: directory path
 * Returns: file count within directory, including hidden files
 *
 * This function is non-recursive.
 **/
gint
xfdir_count_files (const gchar * path) {
    if(!rfm_g_file_test (path, G_FILE_TEST_IS_DIR))
        return 0;
    DIR *directory;
    directory = opendir (path);
    if(!directory)
        return 0;
    gint count = 0;
    struct dirent *d;
    while((d = readdir (directory)) != NULL) {
        if(strcmp (d->d_name, ".") == 0)
            continue;
        if(strcmp (d->d_name, "..") == 0)
            continue;
        count++;
    }
    closedir (directory);
    return count;
}

/**
 * xfdir_count_hidden_files:
 * @path: directory path
 * Returns: hidden file count within directory
 *
 * This function is non-recursive.
 **/
gint
xfdir_count_hidden_files (const gchar * path) {
    if(!rfm_g_file_test (path, G_FILE_TEST_IS_DIR))
        return 0;
    DIR *directory;
    directory = opendir (path);
    if(!directory)
        return 0;
    gint count = 0;
    struct dirent *d;
    while((d = readdir (directory)) != NULL) {
        if(strcmp (d->d_name, ".") == 0)
            continue;
        if(strcmp (d->d_name, "..") == 0)
            continue;
        if(d->d_name[0] != '.')
            continue;
        count++;
    }
    closedir (directory);
    return count;
}

// This is obsolete now...
void
xfdir_register_popup(view_t *view_p, GtkWidget *popup){
    if (!view_p || !GTK_IS_WIDGET(popup)){
	g_warning("xfdir_register_popup(): !view_p || !GTK_IS_WIDGET(popup)");
	return;
    }
    g_object_set_data(G_OBJECT(view_p->widgets.paper), 
		"private_popup_widget", popup);
}

gboolean
xfdir_monitor_control_greenlight(widgets_t *widgets_p){
    if (!widgets_p){
	g_warning("xfdir_monitor_control_greenlight(): widgets_p is NULL\n");
	return FALSE;
    }
    view_t *view_p = widgets_p->view_p;
    if (!view_p) {
	NOOP("xfdir_monitor_control_greenlight(): view_p is NULL\n");
	return FALSE;
    }
    NOOP(stderr, "signalling with xfdir_monitor_control_greenlight\n");
    // Just signal here. Monitor thread may even be dead for all that matters.

    g_mutex_lock(view_p->mutexes.monitor_control);
    gboolean active_monitor = view_p->flags.active_monitor;
    view_p->flags.monitor_go = TRUE;
    g_cond_broadcast(view_p->mutexes.monitor_signal);
    g_mutex_unlock(view_p->mutexes.monitor_control);
    return active_monitor;
}

static void *
destroy_view(void *data){
    if (rfm_global_p->self == g_thread_self()) g_error("rfm_destroy_view is a thread function.\n");
    view_t *view_p=data;

    TRACE("disconnecting g_signal handler %ld\n", view_p->signal_handlers[0]);
    if (view_p->signal_handlers[0]) {
	g_signal_handler_disconnect(G_OBJECT(rfm_global_p->window),
	     view_p->signal_handlers[0]);
    }
    	
    view_p->widgets.status=NULL;
    view_p->widgets.paper=NULL;
   // Cleanup go history.
    GList *tmp;
    for(tmp = view_p->go_list; tmp; tmp = tmp->next) {
	rfm_destroy_entry ((record_entry_t *) (tmp->data));
    }
    g_list_free (view_p->go_list);
    // Cleanup selection list.
    GSList *t;
    for(t = view_p->selection_list; t && t->data; t = t->next) {
	rfm_destroy_entry ((record_entry_t *) (t->data));
    }
    g_slist_free (view_p->selection_list);
    for(t = view_p->f_list; t && t->data; t = t->next) {
	rfm_destroy_entry ((record_entry_t *) (t->data));
    }
    g_slist_free (view_p->f_list);
    // Cleanup reselect list.
    for(t = view_p->reselect_list; t && t->data; t = t->next) {
	g_free(t->data);
    }
    g_slist_free (view_p->reselect_list);
    g_free(view_p->lp_command);


    rfm_mutex_free(view_p->mutexes.reload_mutex);
    rfm_mutex_free(view_p->mutexes.status_mutex);

    rfm_mutex_free(view_p->mutexes.monitor_control );
    rfm_cond_free(view_p->mutexes.monitor_signal);

    rfm_mutex_free(view_p->mutexes.population_serial );
    rfm_rw_lock_clear(&(view_p->mutexes.monitor_lock)); 
    rfm_rw_lock_clear(&(view_p->mutexes.population_rwlock));
    rfm_rw_lock_clear(&(view_p->mutexes.view_lock));

    g_free (view_p->deepest_dir);
    g_free(view_p->widgets.notebook);
    rfm_destroy_entry(view_p->en);
    g_free(view_p);
    return NULL;
}


view_t *
rodent_new_view(void){
    view_t *view_p = (view_t *) malloc (sizeof (view_t));
    if (!view_p) g_error("malloc: %s", strerror(errno));
    memset (view_p, 0, sizeof (view_t));

    // initialize the mutex collection 

    rfm_mutex_init(view_p->mutexes.reload_mutex);
    rfm_mutex_init(view_p->mutexes.status_mutex);


    rfm_mutex_init(view_p->mutexes.monitor_control);
    rfm_cond_init(view_p->mutexes.monitor_signal);

    rfm_mutex_init(view_p->mutexes.population_serial);
	
    rfm_rw_lock_init(&(view_p->mutexes.monitor_lock)); 
    rfm_rw_lock_init(&(view_p->mutexes.population_rwlock));
    rfm_rw_lock_init(&(view_p->mutexes.view_lock));

    view_p->mouse_event.boxX = -1;
    view_p->mouse_event.boxY = -1;

    view_p->flags.preferences = __SHOW_IMAGES;
    view_p->flags.sortcolumn = GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID;
    return view_p;
}

void *
rodent_set_toggle_buttons(void *data){
    if (rfm_global_p->self != g_thread_self()){
	rfm_context_function(rodent_set_toggle_buttons, data);
	return NULL;
    }
    view_t *view_p = data;
    GtkWidget *button;
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "show_hidden_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.preferences & __SHOW_HIDDEN);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "case_insensitive_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.preferences & __CASE_INSENSITIVE);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "show_previews_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.preferences & __SHOW_IMAGES);
    }
    //sort stuff //////////////////////////////////////////////////////////////
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Name_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.sortcolumn == NAME_SORT);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Type_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.sortcolumn == TYPE_SORT);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Date_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.sortcolumn == DATE_SORT);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Size_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.sortcolumn == SIZE_SORT);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Owner_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.sortcolumn == OWNER_SORT);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Group_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.sortcolumn == GROUP_SORT);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Mode_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), view_p->flags.sortcolumn == MODE_SORT);
    }
    //  ascending/descending
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Ascending_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), SHOWS_ASCENDING(view_p->flags.preferences));
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Descending_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), !SHOWS_ASCENDING(view_p->flags.preferences));
    }
    // iconsize buttons  /////////////////////////////////////////////////////////
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "List_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), ICON_SIZE_ID(view_p) == LIST_ICON_SIZE);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Tiny_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), ICON_SIZE_ID(view_p) == TINY_ICON_SIZE);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Normal_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), ICON_SIZE_ID(view_p) == SMALL_ICON_SIZE);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Big_buttonBig_buttonBig_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), ICON_SIZE_ID(view_p) == MEDIUM_ICON_SIZE);
    }
    button =
	g_object_get_data(G_OBJECT(rfm_global_p->window), "Huge_button");
    if (button){
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(button), ICON_SIZE_ID(view_p) == BIG_ICON_SIZE);
    }

    return NULL;
}
