#ifdef COPYRIGHT_INFORMATION
#include "gplv3.h"
#endif
/*
 * Copyright (C) 2002-2012 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// All this runs in non main threads.
static gboolean
transfer_operations(view_t *view_p, xfdir_t *new_xfdir_p, const gchar *module); 

typedef struct reload_t {
    const view_t *view_p;
    record_entry_t *target_en;
    xfdir_t *xfdir_p;
}reload_t;

typedef struct heartbeat_t{
    const view_t *view_p;
    const record_entry_t *target_en;
    xfdir_t *xfdir_p;
    GMutex *xfdir_mutex;
    GCond *xfdir_signal;
    gint heartbeat;
    view_preferences_t *view_preferences_p;
    GThread *thread;
} heartbeat_t;

#define resaturate(X) rfm_context_function (resaturate_f, X);

static void *
resaturate_f(gpointer data){
    view_t *view_p = data;
    if (!rfm_main_context_check(view_p, FALSE)) return NULL;
    const population_t *population_p = 
	rodent_find_in_population (view_p, view_p->mouse_event.current_mouseX, view_p->mouse_event.current_mouseY);
    if (population_p) {
	    rodent_saturate_item (view_p, population_p);	
    }
    return NULL;
}

static xfdir_t *
get_root_xfdir (view_t * view_p) {
    GSList *plugin_list = rfm_find_plugins ();
    GSList *list;
    gint max_elements = 0;
    for(list = plugin_list; list; list = list->next) {
	if (rfm_is_root_plugin ((gchar *) list->data)) {
	    max_elements++;
	} else {
	    NOOP(stderr, "%s is not a root plugin\n", (gchar *) list->data);
	}
    }
    // extra elements: home directory and filesystem icon.
    max_elements += 2;  
    xfdir_t *xfdir_p = (xfdir_t *)malloc(sizeof(xfdir_t));
    if (!xfdir_p) g_error("malloc: %s:", strerror(errno));
    memset (xfdir_p, 0, sizeof(xfdir_t));

    xfdir_p->pathc = max_elements;	
    xfdir_p->gl = (dir_t *)malloc(xfdir_p->pathc*sizeof(dir_t));
    if (!xfdir_p->gl) g_error("malloc: %s", strerror(errno));
    memset(xfdir_p->gl,0,xfdir_p->pathc*sizeof(dir_t));
    
    

    gint item=0;
    /* create filesystem root icon */
    xfdir_p->gl[item].en=rfm_stat_entry("/", 0);
    xfdir_p->gl[item].en->path = g_strdup("/");
    xfdir_p->gl[item].en->tag = g_strdup(_("This is the root of the filesystem"));
    xfdir_p->gl[item].pathv = g_strdup(_("Root Directory"));
    SET_ROOT_TYPE(xfdir_p->gl[item].en->type);
    /* create homedir icon */
    item++;
    xfdir_p->gl[item].en=rfm_stat_entry(g_get_home_dir(), 0);
    xfdir_p->gl[item].en->path = g_strdup(g_get_home_dir());
    xfdir_p->gl[item].pathv = g_strdup(_("Home Directory"));
    SET_ROOT_TYPE(xfdir_p->gl[item].en->type);

    item++;

    NOOP ("ROOTS: now plugins...\n");
    for(list = plugin_list; list; list = list->next) {
	if (!rfm_is_root_plugin ((gchar *) list->data))continue;
	const gchar *module_name = list->data;
        NOOP ("ROOTS: creating population item for module %s\n", module_name);
	xfdir_p->gl[item].en=rfm_mk_entry(0);
	xfdir_p->gl[item].en->st = NULL;
	xfdir_p->gl[item].en->module = module_name; 
	gchar *module_label = 
	    rfm_void (PLUGIN_DIR, (void *)module_name, "module_label");
   
	xfdir_p->gl[item].en->path= module_label;    
	xfdir_p->gl[item].pathv = g_strdup(module_label);
	SET_ROOT_TYPE(xfdir_p->gl[item].en->type);
        item++;
    }
    rfm_cursor_reset(rfm_global_p->window);
    return xfdir_p;
}


static void *
direct_loadxfdir(gpointer data){
    NOOP(stderr, "direct_loadxfdir\n");
    reload_t *reload_p = data;
    view_preferences_t *view_preferences_p = 
	rfm_get_view_preferences (reload_p->view_p->flags.type,
		(record_entry_t *)reload_p->target_en);
    xfdir_t *xfdir_p = xfdir_get ((view_t *)reload_p->view_p, 
	    view_preferences_p,
	    (record_entry_t *)reload_p->target_en,
	    NULL);
	    
    if(xfdir_p==NULL) {
	NOOP ("load_xfdir_structure: xfdir_p is NULL!\n");
    }
    return xfdir_p;
}

static void *
heartbeat_loadxfdir(gpointer data){
    heartbeat_t *heartbeat_p = data;
    xfdir_t *xfdir_p = NULL;

    NOOP( "at heartbeat_loadxfdir\n");
    xfdir_p = xfdir_get ((view_t *)heartbeat_p->view_p, 
	    heartbeat_p->view_preferences_p,
	    (record_entry_t *)heartbeat_p->target_en,
	    &(heartbeat_p->heartbeat));
    NOOP( "at heartbeat_loadxfdir got xfdir_p\n");
	    
    g_mutex_lock(heartbeat_p->xfdir_mutex);
    heartbeat_p->xfdir_p = xfdir_p;
    g_cond_signal(heartbeat_p->xfdir_signal);
    g_mutex_unlock(heartbeat_p->xfdir_mutex);
    if(xfdir_p==NULL) {
	NOOP ("load_xfdir_structure: xfdir_p is NULL!\n");
    }
    return xfdir_p;

}

static 
void *wait_on_thread(gpointer data){
    heartbeat_t *heartbeat_p = data;
    xfdir_t *xfdir_p = g_thread_join(heartbeat_p->thread);

    rfm_mutex_free(heartbeat_p->xfdir_mutex);
    rfm_cond_free(heartbeat_p->xfdir_signal);
    g_free(heartbeat_p->view_preferences_p);
    g_free (heartbeat_p);
    return xfdir_p;
}


static
xfdir_t *
load_xfdir_with_timeout (reload_t *reload_p, gboolean with_heartbeat) {
    TRACE( "load_xfdir_with_timeout with_heartbeat %d\n",with_heartbeat);
    if (!with_heartbeat){
	return direct_loadxfdir(reload_p);

    }
    heartbeat_t *heartbeat_p = (heartbeat_t *)malloc(sizeof(heartbeat_t));
    if (!heartbeat_p) g_error("malloc heartbeat_p: %s\n",strerror(errno));
    memset(heartbeat_p, 0, sizeof(heartbeat_t));

    // Get user defined preferences for target entry
    heartbeat_p->view_preferences_p = 
	rfm_get_view_preferences (reload_p->view_p->flags.type,
		(record_entry_t *)reload_p->target_en);

#ifdef DEBUG
	if(reload_p->view_p->flags.preferences & __SHOW_HIDDEN){
	    NOOP( "load_xfdir_with_timeout: shows hidden...\n");
	} else {
	    NOOP( "load_xfdir_with_timeout: does *not* show hidden...\n");
	} 
#endif

    rfm_mutex_init(heartbeat_p->xfdir_mutex);
    rfm_cond_init(heartbeat_p->xfdir_signal);
    heartbeat_p->view_p = reload_p->view_p;
    heartbeat_p->target_en = reload_p->target_en;

    g_mutex_lock(heartbeat_p->xfdir_mutex);
    NOOP( "Creating wait thread for heartbeat_loadxfdir\n");
    heartbeat_p->thread =
	rfm_thread_create ("heartbeat_loadxfdir", heartbeat_loadxfdir, heartbeat_p, TRUE);
    if (!heartbeat_p->thread){
	g_mutex_unlock(heartbeat_p->xfdir_mutex);
	return NULL;
    }
#define RFM_LOAD_TIMEOUT 5
    if (!heartbeat_p->xfdir_p) {
	gint load_timeout = RFM_LOAD_TIMEOUT;
	const gchar *p = getenv ("RFM_LOAD_TIMEOUT");
	if(p && strlen (p)) {
	    errno=0;
	    long value = strtol(p, NULL, 0);
	    if (errno == 0){
		load_timeout = value;
	    }
	}


	// loop will be broken when heartbeat stops.
	// since this is read only, mutex can be skipped.
	while (heartbeat_p->heartbeat >= 0){
	    NOOP("*heartbeat records = %d\n", heartbeat_p->heartbeat);
	    rfm_threadwait();
	}
	// Now that heartbeat has stopped. Wait for the timeout condition.
	if (!rfm_cond_timed_wait(heartbeat_p->xfdir_signal, heartbeat_p->xfdir_mutex, load_timeout))
	{
	    g_mutex_unlock(heartbeat_p->xfdir_mutex);
	    TRACE( "dead heartbeat\n");
	    // Dead heartbeat:
	    // Fire off a wait and cleanup thread.
	    rfm_view_thread_create((view_t *)heartbeat_p->view_p, wait_on_thread, heartbeat_p, "wait_on_thread");
	    return NULL;
	}
    }
    g_mutex_unlock(heartbeat_p->xfdir_mutex);
    xfdir_t *xfdir_p = wait_on_thread(heartbeat_p);
    return xfdir_p;
}


static void
exit_helper_threads(view_t *view_p){
    // Signal monitor to exit.
    xfdir_exit_monitor (view_p);
    // If any thumbnailer is active, signal it to exit:
    if (view_p->flags.thumbnailer_active) {
	view_p->flags.thumbnailer_exit=TRUE;    
    } 
    // FIXME: signal exit to tips threads here (normal or preview)
    // to avoid race collision. This is not yet coded.
    // XXX: Is this still valid?
}


static void
timeout_message(widgets_t *widgets_p, const gchar *path){
    // ETIMEDOUT       Connection timed out (POSIX.1)
    // EDEADLK         Resource deadlock avoided (POSIX.1)
    gchar *msg = g_strdup_printf(_("Unable to load: %s..."), path);
    rfm_threaded_show_text(widgets_p);
    gchar *m = g_strconcat( strerror(ETIMEDOUT), ": ", msg, 
	    " (", strerror(EDEADLK),")\n", NULL);
    rfm_threaded_diagnostics(widgets_p, "xffm/stock_dialog-error", m);
    g_free(msg);
}

static void
error_trap(xfdir_t *new_xfdir){
    if (new_xfdir->pathc > 0) return;
    g_warning("No items were loaded into xfdir structure!");
    new_xfdir->pathc = 1;
    new_xfdir->gl = (dir_t *) malloc (sizeof (dir_t));
    if(!new_xfdir->gl) g_error ("malloc: %s", strerror(errno));
    memset (new_xfdir->gl, 0, sizeof (dir_t));
    // root item:
    new_xfdir->gl[0].en = NULL;
    new_xfdir->gl[0].pathv = g_strdup (g_get_host_name ());
    return;
}

static
void
set_deepest_dir (view_t * view_p) {
    if(view_p->en && view_p->en->path && rfm_g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
        if(!view_p->deepest_dir || !strstr (view_p->deepest_dir, view_p->en->path)) {
            g_free (view_p->deepest_dir);
            view_p->deepest_dir = g_strdup_printf ("%s/.", view_p->en->path);
        }
    } else {
        //g_warning("cannot set deepest dir");
    }
}



static void
transfer_open(view_t *view_p, xfdir_t *new_xfdir){
    NOOP(stderr, "transfer_open...\n");
    // Error trap: if nothing is loaded, then at least put in
    // the localhost icon.
    error_trap(new_xfdir);
    // Set the entry preferences in the view.
    record_entry_t *en = view_p->en;
    view_preferences_t *view_preferences_p = 
	rfm_get_view_preferences (view_p->flags.type, en);
    rfm_set_view_preferences (view_p, view_preferences_p);
    g_free(view_preferences_p);
#ifdef DEBUG
	if(view_p->flags.preferences & __SHOW_HIDDEN){
	    NOOP( "transfer_open: shows hidden...\n");
	} else {
	    NOOP( "transfer_open: does *not* show hidden...\n");
	}
#endif


    // This will set the layout parameters to the correct
    // values and then set the paper widget accordingly.
    rfm_layout_configure(view_p, new_xfdir->pathc);
      
    // Non null entry scenario
    if (en) view_p->module = en->module;
    else view_p->module = NULL;

    widgets_t *widgets_p = &(view_p->widgets);
    g_free (widgets_p->workdir);
    if (en && IS_LOCAL_TYPE(en->type)) widgets_p->workdir = g_strdup (en->path);
    else widgets_p->workdir = g_strdup(g_get_home_dir());
    // set the dtfile quick dir buttons (or menu items in our case)
    if(en && !en->module) {
	if(!en->path || 
		!IS_SDIR(en->type)) {
	    g_warning ("Invalid entry sent to xfdir_get_local");
	} else {
	    set_deepest_dir (view_p);
	}
    }
    return;
}

static gboolean
want_monitor(widgets_t *widgets_p, record_entry_t *en){
    if (!en) return FALSE; 
    // Local files monitor:
    if (en->module){
	// Plugin requesting a file monitor thread:
	if (rfm_void(PLUGIN_DIR, en->module, "module_monitor")) return TRUE;
    } else {
	// Local files monitor:
	if (IS_LOCAL_TYPE(en->type)) return TRUE;
	// Non local types with monitor:
	else if (IS_MONITOR_TYPE(en->type))return TRUE;
    } 
    if (g_path_is_absolute(en->path)) {
	// warning 
	rfm_threaded_diagnostics (widgets_p, "xffm/stock_disconnect", NULL);
	rfm_threaded_diagnostics (widgets_p, "xffm-tag/red", 
		g_strconcat(_("Could not initialize monitoring"), 
		    " (", en->path,
		    ": ", _("Remote Connection..."),")\n", NULL));
    }
    return FALSE;
}
static void 
save_environment_desktop_dir(view_t *view_p, record_entry_t *target_en){
    // save the default desktop dir to the environment...
    if(view_p->flags.type != DESKVIEW_TYPE) {
	return;
    }
    g_free (view_p->desktop_dir);
    if (!target_en) {
	view_p->desktop_dir = g_strdup("");
    } else if (target_en->module) {
	if (rfm_void (PLUGIN_DIR, target_en->module, "module_name")) {
	    view_p->desktop_dir = g_strdup_printf ("module:%s",
		(gchar *)rfm_void (PLUGIN_DIR, target_en->module, "module_name"));
	} else {
	    view_p->desktop_dir = g_strdup (target_en->module);
	}
    } else {
	view_p->desktop_dir = g_strdup (target_en->path);
    }
    rfm_setenv ( "RFM_DESKTOP_DIR", view_p->desktop_dir, TRUE);
    // save configuration too
    rfm_rational (MODULE_DIR, "settings", (void *) "RFM_DESKTOP_DIR", (void *) view_p->desktop_dir, "mcs_set_var");
    NOOP(" setting RFM_DESKTOP_DIR to %s\n",view_p->desktop_dir);
    return;
}


// This requires GDK MUTEX and could be sent to thread-queue
// 
static void
transfer_close(view_t *view_p, xfdir_t *new_xfdir_p){
    record_entry_t *en = view_p->en;
    widgets_t *widgets_p = &(view_p->widgets);
    if (en && en->tag){
	rfm_threaded_status (widgets_p, "xffm/stock_dialog-info", g_strdup(en->tag));
    } else if (!en) {
	rfm_threaded_status (widgets_p, "xffm/device_computer", g_strdup(_("Installed Plugins")));
    }


    // start background directory monitor thread, or else free
    // xdfir memory block.
    
    if (want_monitor(widgets_p, en)) {
	NOOP("xfdir_start_monitor...\n");
	g_mutex_lock(view_p->mutexes.monitor_control);
	    view_p->flags.active_monitor = TRUE;
	g_mutex_unlock(view_p->mutexes.monitor_control);
	xfdir_start_monitor (view_p, new_xfdir_p);
    } else {
	// Free xfdir memory block.	
	NOOP("Not starting xfdir monitor...\n");
	g_mutex_lock(view_p->mutexes.monitor_control);
	view_p->flags.active_monitor = FALSE;
	g_mutex_unlock(view_p->mutexes.monitor_control);
	xfdir_free_data (new_xfdir_p);
	g_free (new_xfdir_p);

    }
    if (view_p->flags.type == DESKVIEW_TYPE) {
	// Save desktop directory for navigational desktop
	save_environment_desktop_dir(view_p, en);
    } else {
	rfm_context_function(rodent_set_scroll_position, view_p);
        rodent_set_view_icon (view_p);
        rodent_set_view_title (view_p);
	rodent_set_toggle_buttons(view_p);
 	rfm_update_status_line (view_p);
    }
    // resaturate any icon which was previously saturated...
    
    resaturate(view_p);
}

// This should only be done after the write lock is active...
static void
increase_population_serial (view_t * view_p) {
    g_mutex_lock (view_p->mutexes.population_serial);
    view_p->flags.population_serial++;
    g_mutex_unlock (view_p->mutexes.population_serial);
    TRACE ("increase_population_serial: %d\n",
	    view_p->flags.population_serial);

}

static record_entry_t *
init_view (view_t * view_p, record_entry_t * en) {
    TRACE ("rodent_population: init_view\n");

    increase_population_serial (view_p);
    view_p->mouse_event.boxX = -1, view_p->mouse_event.boxY = -1;
    view_p->mouse_event.old_X = -1, view_p->mouse_event.old_Y = -1;
#if 0
    // destroy selection ... not really...
    if(view_p->selection_list) {
	GSList *tmp=view_p->selection_list;
	for (;tmp && tmp->data; tmp=tmp->next){
	    record_entry_t *en=tmp->data;
	    rfm_destroy_entry(en);
	}
        g_slist_free (view_p->selection_list);
        view_p->selection_list = NULL;
    }
#endif
    view_p->mouse_event.mouseX = -1;
    view_p->mouse_event.mouseY = -1;
    view_p->widgets.rename = NULL;
    view_p->mouse_event.label_p = NULL;
    view_p->mouse_event.doing_drag_p = NULL;
    view_p->mouse_event.saturated_p = NULL;
 // nope:   view_p->mouse_event.selected_p = NULL;
    view_p->en = en;

  
    return en;
}


void
rodent_destroy_population (population_t **population_pp) {
    if(population_pp) return;
    population_t **tmp;
    NOOP ("rodent_population: rodent_destroy_population\n");
    for(tmp = population_pp; tmp && *tmp; tmp++) {
        population_t *population_p = *tmp;
        destroy_population_item (population_p);
        population_p = NULL;
    }
    g_free (population_pp);
}

// This is the reload done by user driven callback (also in its own thread).

static void 
update_f(view_t *view_p, record_entry_t *target_en, xfdir_t *new_xfdir) {
    if (!new_xfdir){
	g_error("new_xfdir cannot be null here.");
    }

    // Set lock for icontheme tests:
    rfm_rw_lock_reader_lock(&(rfm_global_p->icon_theme_lock));
    // Lock is now in place, icontheme thread will wait.
    
    if (!rfm_population_write_lock (view_p)) {
	g_warning ("update_f: !rfm_population_write_lock (view_p)");
	return;
    }

    NOOP(stderr, "+++  update_f...\n");
    population_t **old_population_pp = view_p->population_pp;
    record_entry_t *old_en = view_p->en;
    const gchar *module = (old_en)?old_en->module:NULL;

    init_view (view_p, target_en);
    NOOP( "init_view done...\n");
    transfer_open(view_p, new_xfdir);
    NOOP( "transfer_open done...\n");
    if (!transfer_operations(view_p, new_xfdir, module)){
	g_warning("This is bad. transfer_operations failed\n");
    }
    NOOP( "transfer_operations done...\n");


   
    rodent_destroy_population(old_population_pp);
    rfm_destroy_entry(old_en);
    // This function will do the scroll bar setting.
    transfer_close(view_p, new_xfdir);
    rfm_population_write_unlock (view_p);
    rodent_redraw_items(view_p);

    NOOP("rfm_fireup_previews\n");	
    rfm_fireup_previews (view_p);	    
    // Lift lock. Icon theme checks may now proceed as usual.	
    rfm_rw_lock_reader_unlock(&(rfm_global_p->icon_theme_lock));

    NOOP("hard reload thread done...\n");
}

extern gboolean noexpose;
// This is a thread function...
static void *
reload_f (gpointer data) {
    // Hold until the setup RW lock is available.
    rfm_rw_lock_reader_lock (&(rfm_global_p->setup_lock));

    reload_t *reload_p = data;
    view_t *view_p = (view_t *)reload_p->view_p;
    rfm_rw_lock_reader_unlock (&(rfm_global_p->setup_lock));

    // This mutex will protect against multiple function calls
    // simultaneous on the same view.
    g_mutex_lock(view_p->mutexes.reload_mutex);
    widgets_t *widgets_p = &(view_p->widgets);
    record_entry_t *record_entry_p  = (record_entry_t *)reload_p->target_en;
    xfdir_t *new_xfdir = NULL;

    rodent_set_expose_hold(view_p, TRUE);

    // Step 2.
    // 2. Get new xfdir. This requires no extra locks nor mutexes...
    // XXX: add roots should work with a xfdir get to simplify this...
    gboolean aborted_load = FALSE;
    if (record_entry_p == NULL){
	// We could wait for monitor exited condition to continue...
	new_xfdir = get_root_xfdir(view_p);
    } else {
	if (IS_LOCAL_TYPE(record_entry_p->type)){
	    // This is a non-heartbeat controlled load. Will block.
	    new_xfdir = load_xfdir_with_timeout(reload_p, FALSE);
	    //new_xfdir = load_xfdir_with_timeout(reload_p, TRUE);
	} else {
	    // This is a heartbeat controlled load.
	    if (reload_p->target_en->module && rfm_void(PLUGIN_DIR, reload_p->target_en->module, "blocking_load")) {
		new_xfdir = load_xfdir_with_timeout(reload_p, FALSE);
	    } else {
		new_xfdir = load_xfdir_with_timeout(reload_p, TRUE);
	    }
	}

	// On timeout, null is returned.
	if (!new_xfdir){
	    timeout_message(widgets_p, record_entry_p->path);
	    // If this is not the initial load, then abort reload.
	    rfm_destroy_entry(record_entry_p);
	    if(view_p->go_list && g_list_last (view_p->go_list)){
		rodent_expose_all (view_p);
		aborted_load = TRUE;
	    } else {
		reload_p->target_en = record_entry_p = NULL;
		new_xfdir = get_root_xfdir(view_p);
	    }
	} 
    }
    TRACE( "reload_f has loaded xfdir structure...\n");
    rodent_set_expose_hold(view_p, FALSE);
    if (!aborted_load){
	// By now we have a new_xfdir, or valid NULL, so we can do a transfer.
	// target_en will substitute view_p->en downstream.
	exit_helper_threads(view_p);
	update_f(view_p, reload_p->target_en, new_xfdir);
	//sem_post(rfm_global_p->setup_semaphore);
    } else {
	// Reset layout status so that a simple expose will put previous
	// population on the paper.
	TRACE( "rodent_full_reload_view(), reload_f aborted call.\n");
    }
    rfm_threaded_cursor_reset(rfm_global_p->window);
    g_free(reload_p);
    g_mutex_unlock(view_p->mutexes.reload_mutex);
	NOOP(stderr, "reload_f () is done...\n");

    return GINT_TO_POINTER(!aborted_load);
}

static population_t **
mk_grid_elements (view_t * view_p, int grid_elements) {
    NOOP (">> graphics_mk_grid_elements\n");
    population_t **population_pp;
    /* we use (grid_elements+1) to use a NULL pointer as an
     * end of list marker */
    NOOP (stderr, "mk_grid_elements(): grid_elements=%d (elements)\n", grid_elements);
    //increase_population_serial (view_p);     
    population_pp = (population_t **) malloc ((grid_elements + 1) * sizeof (population_t *));
    memset ((void *)(population_pp), 0, (grid_elements + 1) * sizeof (population_t *));
    population_pp[grid_elements] = NULL;
    return population_pp;
}
//

static void
transfer_population (view_t * view_p, xfdir_t *new_xfdir_p, const gchar *module) {
     // Construct hash table for reloaded population_pp;
    // (skip up element)
    // (hashed items are dir_t *)
    GHashTable *new_hash = g_hash_table_new(g_str_hash, g_str_equal);
    gint i;
    NOOP ("transfer(0x%x): pathc = %d\n", GPOINTER_TO_INT(g_thread_self()),
	    (gint)new_xfdir_p->pathc);
    for(i = 0; i < new_xfdir_p->pathc; i++) {
        if(!new_xfdir_p->gl[i].en) continue;
        //NOOP ("transfer(): new_hash: %s\n", new_xfdir_p->gl[i].en->path);
	g_hash_table_insert(new_hash, new_xfdir_p->gl[i].en->path, (new_xfdir_p->gl)+i);
    }
    NOOP(stderr, "t_population loop 1 done...\n");

    gboolean full_reload = (!view_p->en ||
	    (!module && view_p->en->module ) ||
	    (module && !view_p->en->module ) ||
	    (module && view_p->en->module && strcmp(module, view_p->en->module ) ));


    GHashTable *old_hash = g_hash_table_new(g_str_hash, g_str_equal);
    population_t **population_pp = view_p->population_pp;
    // Construct hash table for current view population_pp.
    // (skip up element)
    // (hashed items are population_t *)
    GSList *remove_list = NULL;
    // this is buggy if more than one dummy item (so do not skip):
    // if (population_pp && *population_pp) population_pp++;
    if (!full_reload) for(; population_pp && *population_pp; population_pp++) {
	population_t *population_p = *population_pp;
	if (!population_p->en || !population_p->en->path) continue;
	g_hash_table_insert(old_hash, population_p->en->path, population_p);
	// Construct list of remove items.
	if (!g_hash_table_lookup(new_hash, population_p->en->path)){
            NOOP ("transfer(): adding to remove list: %s\n", population_p->en->path);
	    remove_list = g_slist_prepend(remove_list, population_p);
	}

    }
    NOOP(stderr, "t_population loop 2 done...\n");

    // create new population_pp
    population_pp = mk_grid_elements (view_p, new_xfdir_p->pathc);
    // Transfer transferable items and create new items.
//    if (view_p->population_pp) {population_pp[0] = view_p->population_pp[0]; }
    
    // If iconsize id has changed, we must also get new icons and 
    // new pango layouts!
    // If only iconsize not changed, we can skip new icons.
    gint icon_size = ICON_SIZE(view_p);
    gint icon_size_id = ICON_SIZE_ID(view_p);
    for(i = 0; i < new_xfdir_p->pathc; i++) {
	if ( new_xfdir_p->gl[i].en && !new_xfdir_p->gl[i].en->path ){
            g_warning("transfer_population(): this is bad....\n");
	    continue;
	}
	population_t *population_p = NULL;
	if (new_xfdir_p->gl[i].en){
	    population_p = g_hash_table_lookup(old_hash, new_xfdir_p->gl[i].en->path);
	}
	if (population_p){
            population_pp[i] = population_p;
	    NOOP(stderr, "updating population record for %s \n",
		    (new_xfdir_p->gl[i].en)?new_xfdir_p->gl[i].en->path:"ROOT");
            /*  update the stat record */
            if(population_p->en->st && new_xfdir_p->gl[i].en->st) {

		// remove any previous popup preview image, including the thumbnail file.
		// if not requested or out of date.
		gboolean out_of_date = memcmp(population_p->en->st, new_xfdir_p->gl[i].en->st, sizeof(struct stat)); 
		gboolean zap_preview = !(view_p->flags.preferences & __SHOW_IMAGES);
		if (zap_preview || out_of_date) {
		    if (out_of_date) {
			memcpy (population_pp[i]->en->st, new_xfdir_p->gl[i].en->st, sizeof (struct stat));
		    }
		    // remove any previous popup preview image, including the thumbnail file.
		    // (see xfdir.i:674)
		    if (population_p->preview_pixbuf) {
			NOOP (stderr, "removing preview pixbuf at transfer_population\n"); 
			population_p->preview_pixbuf = NULL;
			gchar *thumbnail = 
			    rfm_get_thumbnail_path (population_p->en->path, PREVIEW_IMAGE_SIZE);
			if (unlink(thumbnail));
			g_free(thumbnail);
			// We might get a new preview here.
		    }
		    if (zap_preview){
			population_p->pixbuf = 
			    rfm_get_pixbuf (population_p->icon_id, icon_size);
			population_p->flags &= (POPULATION_THUMBNAILED ^ 0xffffffff);
			population_p->flags &= (POPULATION_PREVIEW_DONE ^ 0xffffffff);
			population_p->flags &= (POPULATION_PREVIEW_BUSY ^ 0xffffffff);
		    }
		}
		// update the icon if necessary
		gchar *icon_id = 
		    rfm_get_entry_icon_id(view_p, new_xfdir_p->gl[i].en, TRUE);
		// We add the condition for not showing images, since this is necessary
		// to reflect the change when the user switches the toggle.
		if ((view_p->flags.preferences & __INVALIDATE_ICONS) || strcmp(icon_id, population_p->icon_id)){
		    g_free(population_p->icon_id);
		    population_p->icon_id = icon_id;
		    population_p->pixbuf = 
			rfm_get_pixbuf (population_p->icon_id, icon_size);
		    if (view_p->flags.preferences & __INVALIDATE_ICONS) {
			population_p->flags &= (POPULATION_THUMBNAILED ^ 0xffffffff);
			population_p->flags &= (POPULATION_PREVIEW_DONE ^ 0xffffffff);
			population_p->flags &= (POPULATION_PREVIEW_BUSY ^ 0xffffffff);
		    }
		} else {
		    g_free(icon_id);
		}
	    }
#if 10
	    if (population_p->icon_size_id != icon_size_id) {
		population_p->icon_size_id = icon_size_id;
		rodent_layout_cleanup(population_p);
		// XXX if I only clean out icons when iconsize changes,
		//     things break. Somehow the following 
		//     test does not do the trick.
		//     (no big deal):
		// } if (population_p->icon_size != icon_size){
		//
		population_p->pixbufW = population_p->icon_size;
		population_p->pixbufH = population_p->icon_size;
		NOOP(stderr, "updating iconsize to: %d\n",population_p->icon_size);
		population_p->pixbuf = 
			    rfm_get_pixbuf (population_p->icon_id, 
				    icon_size);
		// This will create its own thread:
		if (view_p->flags.preferences & __SHOW_IMAGES) {
		    update_preview(view_p, population_p);
		}
	    }
#endif
	    
	} else {
	    // create a new population item.
            NOOP (stderr, "transfer(): Creating new population item for %s\n",
		    (new_xfdir_p->gl[i].en)?new_xfdir_p->gl[i].en->path:"ROOT");
	    // create mimetype icon (no magic)
	    // bugfree test for doing magic... if stat record is valid, do magic.
	    gboolean do_magic = TRUE;
	    if (new_xfdir_p->gl[i].en == NULL ||
		    new_xfdir_p->gl[i].en->st == NULL || 
		    (new_xfdir_p->gl[i].en->st->st_mtime == 0 &&
		    new_xfdir_p->gl[i].en->st->st_blksize == 0)){
		do_magic = FALSE;
	    }
	    gchar *icon_id = 
		rfm_get_entry_icon_id(view_p, new_xfdir_p->gl[i].en, do_magic);
            population_pp[i] = 
		create_population_t (view_p, new_xfdir_p->gl[i].en, i, icon_id, new_xfdir_p->gl[i].pathv);
	    g_free(icon_id);

            NOOP ("transfer(): Creating new mimetype icon for %s\n",
		    (new_xfdir_p->gl[i].en)?new_xfdir_p->gl[i].en->path:"ROOT");    

	    if (!GDK_IS_PIXBUF(population_pp[i]->pixbuf)){
		NOOP("transfer_population(): pixbuf creation failed for %s\n", population_pp[i]->icon_id);
		g_free(population_pp[i]->icon_id);
		population_pp[i]->icon_id = g_strdup("xffm/stock_file");
		population_pp[i]->pixbuf = 
		    rfm_get_pixbuf ("xffm/stock_file", icon_size);
	    }
	}
    }
    NOOP(stderr, "transfer_population loop 3 done...\n");
    view_p->flags.preferences &= (__INVALIDATE_ICONS ^ 0xffffffff);
    // That done, update the population selection list. 

    GSList *tmp;


    // Time to do the switch to the new population pp.
    // old population is now obsolete. Pointer array
    // is no longer necessary (we have the remove list).
    population_t **obsolete_population_pp = view_p->population_pp;
    view_p->population_pp = population_pp;
    g_free(obsolete_population_pp);

    // Remove obsolete population items
    for (tmp = remove_list; tmp && tmp->data; tmp = tmp->next){
	population_t *population_p = tmp->data;
        NOOP ("transfer(): destroying  population item: %s\n", population_p->en->path);
	destroy_population_item (population_p);
    }


    NOOP(stderr, "transfer_population loop 5 done...\n");

    // Clean up temporary hash tables and remove list.
    g_hash_table_destroy(old_hash);
    g_hash_table_destroy(new_hash);
    g_slist_free(remove_list);

    rfm_layout_pango_layout_setup(view_p);

    return ;
}


static gboolean
transfer_operations(view_t *view_p, xfdir_t *new_xfdir_p, const gchar *module){
    /* make new population_pp and transfer old population_pp elements,
     * creating new elements for those not in old population_pp
     * and eliminating those in old population_pp no longer present
     * */

    // Write lock on population...
    //if (!rfm_population_write_lock (view_p)){
//	return FALSE;
    //}
    // Requires write lock at least:
    transfer_population (view_p, new_xfdir_p, module);


    NOOP(stderr, "transfer_population done...\n");
    // recalculate x,y
    // get x,y coordinates for each population_p
    // or correct row, column
    //
    // Requires read lock at least:
    rodent_recalc_population_geometry (view_p);
    //rfm_population_write_unlock (view_p);
   
    if(view_p->en && view_p->en->st){
        memcpy (view_p->en->st, new_xfdir_p->en->st, sizeof (struct stat));
    }

    return TRUE;
}

// This is the reload done by the monitor thread.

static void *
thread_reload_view (view_t *view_p, xfdir_t *new_xfdir_p) {

    TRACE( "+++   thread_reload_view...\n");

    g_mutex_lock(view_p->mutexes.status_mutex);
    gboolean status = view_p->flags.status;
    g_mutex_unlock(view_p->mutexes.status_mutex);
    if(status == STATUS_EXIT) {
	NOOP(stderr, "view_p->flags.status == STATUS_EXIT\n");
	return NULL;
    }
        
    //g_mutex_lock(view_p->mutexes.reload_mutex);

    rfm_rw_lock_reader_lock(&(rfm_global_p->icon_theme_lock));


    if(!new_xfdir_p) {
        g_error ("short_rodent_thread_reload_view: !new_xfdir_p");
    }
    NOOP (stderr, "starting short_rodent_reload_view...\n");
    /*  monitor in wait state  with view_p->flags.refresh */

    NOOP ("POP_SEM: reload_view lock\n");
    // population semaphore is ours and population serial is valid...
    // Requires read lock at least:
    if (!rfm_population_write_lock (view_p)) {
	g_warning ("update_f: !rfm_population_write_lock (view_p)");
	return NULL;
    }

    // This will set the layout parameters to the correct
    // values and then set the paper widget accordingly.
    rfm_layout_configure(view_p, new_xfdir_p->pathc);

    if (transfer_operations(view_p, new_xfdir_p, (view_p->en)?view_p->en->module:NULL)){ 
	// This requires GDK MUTEX and could be sent to thread-queue
	NOOP( "short_reload_view: reload is done\n");
	resaturate(view_p);
	NOOP ("population_sem: reload_view(): released!\n");
    }
    rfm_population_write_unlock (view_p);
    rodent_redraw_items(view_p);

    NOOP("softreload: rfm_fireup_previews\n");	
    rfm_fireup_previews (view_p);
    rfm_rw_lock_reader_unlock(&(rfm_global_p->icon_theme_lock));

    //g_mutex_unlock(view_p->mutexes.reload_mutex);	
    // this is dumb: sem_post(rfm_global_p->setup_semaphore);
    // This semaphore holds this function...
    return GINT_TO_POINTER(1);

}

static
GdkPixbuf *get_population_pixbuf(view_t * view_p,
	population_t *population_p,
	const gchar * icon_id){
    if (!population_p || !icon_id){
	g_warning("get_population_pixbuf() !population_p || !icon_id\n");
	return NULL;
    }

    if (population_p->en == NULL) {
	return rfm_get_pixbuf (icon_id, ICON_SIZE(view_p));
    }


    GdkPixbuf *pixbuf = rfm_get_pixbuf (icon_id, ICON_SIZE(view_p));
    if (GDK_IS_PIXBUF(pixbuf)) {
	//g_free(population_p->icon_id);
	//population_p->icon_id = g_strdup(icon_id);
	return pixbuf;
    }
    NOOP("cannot get pixbuf for icon=%s\n", icon_id);
    icon_id = "xffm/stock_file/composite/stock_close";
    pixbuf = get_population_pixbuf(view_p, population_p, "xffm/stock_file/composite/stock_close");
    g_free(population_p->icon_id);
    population_p->icon_id = g_strdup(icon_id);
    return pixbuf;
}


/**
 *
 * create_population_t: create a population element
 *
 * @view_p:
 * @en: the entry to assign to the population element.
 * @element: the element serial id number 
 * @icon_id: the mime_icon identifier. If NULL, then icon is obtained from entry information.
 * @label: the text which is to appear under the icon. If NULL, then the text is inferred from entry information.
 * 
 * **/

static void
set_population_grid_row (view_t * view_p, population_t *population_p, gint element){
    gint grid_rows = GRID_ROWS(view_p);
    gint grid_columns = GRID_COLUMNS(view_p);
    if (view_p->flags.type == ICONVIEW_TYPE) {
	if(grid_columns) {
	    population_p->row = element / grid_columns;
	    population_p->column = element % grid_columns;
	} else {
	    population_p->row = 0;
	    population_p->column = element;
	}
    } else { // transposed matrix for desktop
	if(grid_rows) {
	    population_p->row = element % grid_rows;
	    population_p->column = element / grid_rows;
	} else {
	    population_p->row = element;
	    population_p->column = 0;
	}
    }
}

static population_t *
create_population_t (view_t * view_p,
	record_entry_t * en, 
	gint element, 
	const gchar * icon_id, 
	const gchar * label) 
{
    GdkPixbuf *pixbuf = NULL;
    population_t *population_p;

    population_p = (population_t *) malloc (sizeof (population_t));
    if (!population_p) g_error("malloc: %s", strerror(errno));
    memset (population_p, 0, sizeof (population_t));
    population_p->view_p = view_p;
    population_p->icon_size = ICON_SIZE(view_p);
    population_p->icon_size_id = ICON_SIZE_ID(view_p);
	
    set_population_grid_row(view_p, population_p, element);

    ///////////////////////////////////////

    if(en) {
        population_p->en = rfm_copy_entry (en);
	population_p->module = en->module;
    } else {
        population_p->en = NULL;
	label = g_get_host_name();
    }
    if (icon_id == NULL) {
        population_p->icon_id = rfm_get_entry_icon_id(view_p, population_p->en, TRUE); 
    } else {
	population_p->icon_id = g_strdup(icon_id);
    }


    population_p->flags = 0;
    population_p->pixbuf = get_population_pixbuf(view_p, population_p, population_p->icon_id);

    gint pasteboard_status = rfm_in_pasteboard (view_p, population_p->en);
    NOOP("rfm_in_pasteboard done\n"); 
    if(pasteboard_status == 2) {
	population_p->flags |= POPULATION_CUT;
    } else if (pasteboard_status == 1) {
	population_p->flags |= POPULATION_COPIED;
    }

    if(GDK_IS_PIXBUF(pixbuf)) {
        population_p->pixbufW = gdk_pixbuf_get_width ((const GdkPixbuf *)pixbuf);
        population_p->pixbufH = gdk_pixbuf_get_height ((const GdkPixbuf *)pixbuf);
    } else {
        population_p->pixbufW = population_p->icon_size;
        population_p->pixbufH = population_p->icon_size;
    }


    population_p->label = g_strdup(label);

    return population_p;
}
//////////////////////////   janitor    /////////////////////////////
//

static void
destroy_population_item (population_t * population_p) {
    if(!population_p) return;
    /* objects that need to be unreffed: 
     * (tagged as referrence in types.h) */

    if(population_p->preview_pixbuf){
        g_object_unref (population_p->preview_pixbuf);
	population_p->preview_pixbuf = NULL;
    }

    // This is mutex protected:
    rodent_layout_cleanup(population_p);
    rfm_destroy_entry (population_p->en);
    g_free(population_p->icon_id);
    g_free(population_p->label);
    g_free (population_p);
}


// This is a thread function, must have GDK mutex set for gtk commands...
gpointer
janitor(gpointer data){
    NOOP("starting janitor...\n");
    // This is a stop light for initial setup:
    rfm_rw_lock_reader_lock (&(rfm_global_p->setup_lock));
    rfm_rw_lock_reader_unlock (&(rfm_global_p->setup_lock));

    //GtkWidget *window=data;
    do {
	g_mutex_lock(rfm_global_p->status_mutex);
	gint status = rfm_global_p->status;
	TRACE("janitor at window wait...exit=%d\n", (status == STATUS_EXIT));
	g_cond_wait(rfm_global_p->janitor_signal, 
		rfm_global_p->status_mutex);
	status = rfm_global_p->status;
	if (status == STATUS_EXIT) {
	    g_cond_broadcast(rfm_global_p->status_signal);
	}
	g_mutex_unlock(rfm_global_p->status_mutex);
	
	TRACE("janitor woken.\n");

	rfm_rw_lock_reader_lock (&(rfm_global_p->view_list_lock));
	NOOP(stderr, "Janitor got the view_list_lock\n");
	GSList *view_list = NULL;
	GSList *list = rfm_global_p->window_view_list;
	for (; list && list->data; list=list->next){
	    view_list = g_slist_prepend(view_list, list->data);    
	}
	rfm_rw_lock_reader_unlock (&(rfm_global_p->view_list_lock));

	for (list = view_list; list && list->data; list=list->next){
	    view_t *view_p = list->data;
	    gint view_status;
	    if (status == STATUS_EXIT) view_status = status;
	    else {
		g_mutex_lock(view_p->mutexes.status_mutex);
		view_status = view_p->flags.status;
		g_mutex_unlock(view_p->mutexes.status_mutex);
	    }
	    if (view_status == STATUS_EXIT) {
		g_cond_broadcast(view_p->mutexes.monitor_signal);
		// View needs to be destroyed.
		// Is the view thread locked?
		if(!rfm_rw_lock_writer_trylock(&(view_p->mutexes.view_lock))){
		    // Go on to the next view to cleanup
		    TRACE( "janitor: will not destroy view 0x%x (still in use)...\n",
			GPOINTER_TO_INT(view_p));
#ifdef DEBUG_TRACE
		    // This makes trace very verbose
		    // rfm_thread_unreference(NULL, NULL);
#endif
		    
		    continue;
		}
		TRACE( "janitor() will now destroy view 0x%x\n", 
		    GPOINTER_TO_INT(view_p));
		rfm_global_p->window_view_list =
		    g_slist_remove(rfm_global_p->window_view_list, view_p);
		rfm_rw_lock_writer_unlock(&(view_p->mutexes.view_lock));
		// Cleanup population 
		if (rfm_population_write_lock (view_p)){
		    rodent_destroy_population (view_p->population_pp);
		    rfm_population_write_unlock (view_p);
		} else {
		    NOOP(stderr, "ABSURD: no write lock for view 0x%x, \n", 
			    GPOINTER_TO_INT(view_p));
		}
		view_p->population_pp = NULL;
		// finish up.
		destroy_view(view_p);
		continue;
	    } 
	}
	g_slist_free(view_list);

	rfm_rw_lock_reader_lock (&(rfm_global_p->view_list_lock));
	gint active_views = g_slist_length(rfm_global_p->window_view_list);
	rfm_rw_lock_reader_unlock (&(rfm_global_p->view_list_lock));
	if(!active_views && status == STATUS_EXIT)break;
	if (!active_views) rfm_global_p->status = STATUS_EXIT;
		    
    } while (TRUE);
    TRACE( "Janitor exiting now.\n");
    g_main_loop_quit(rfm_global_p->main_loop);
    return NULL;
}

static gboolean updating_pixbuf_hash = FALSE;
static void *
update_pixbuf_hash_f(void *data){
    if (updating_pixbuf_hash) return NULL;
    updating_pixbuf_hash=TRUE;
    NOOP("---->   time to update pixbuf hash!\n");
    rfm_change_icontheme();
    //sleep(1);
    updating_pixbuf_hash=FALSE;
    return NULL;
}

