/*
 * Copyright (C) 2002-2012 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "rodent_libs.h"
#include "rodent_popup_threads.i"

////////////////////////////////////////////////////////////////////////////////////////
//             rodent_create_popup_bythread guts (threaded functions) 
//
//              everything in this file *must* be gdk mutex protected
////////////////////////////////////////////////////////////////////////////////////////

// menu creation options:
// types: submenu, menuitem, separator
// options:
//        submenu: parent_id, id, icon_id, string, 
//        menuitem: parent_id, id, icon_id, string, menu_callback_t
//        separator: parent_id, id

GtkWidget *
rodent_thread_add_menu_separator (GtkWidget * parent, gchar *text) {
    popup_controller();
    return thread_add_menu_separator (parent, text);
}

GtkWidget *
rodent_thread_add_submenu (GtkWidget * parent, const gchar * label, const gchar * name, const gchar * iconfile) 
{
    popup_controller();
    return thread_add_submenu (parent, label, name, iconfile);
}

GtkWidget *
rodent_thread_mk_pixmap_menu (const gchar * icon_id, GtkWidget * parent, gint caso) {
    popup_controller();
    return thread_mk_pixmap_menu (parent, icon_id, caso);
}

void
rodent_thread_multimenu_make(GtkAccelGroup * accel_group, menu_definitions_t *in_p){
    if (!in_p || !in_p->parent_id){
	g_error("thread_multimenu_make: incorrect function call.\n");
    }
    popup_controller();
    thread_multimenu_make(accel_group, in_p);
    return ;
}

gpointer
rodent_create_popup_bythread(gpointer data){
    if (!rfm_global_p || rfm_global_p->self == g_thread_self()){
	g_warning("rodent_create_popup_bythread: only to be called from non main thread\n");
    }
    static gsize initialized = 0;
    if (g_once_init_enter(&initialized)){
	rfm_rec_mutex_lock (&(rfm_global_p->popup_create_mutex));
	GtkWidget *popup = thread_mk_popup_menu();
	g_cond_broadcast(rfm_global_p->popup_ready);
	rfm_rec_mutex_unlock (&(rfm_global_p->popup_create_mutex));
	g_once_init_leave(&initialized, 1);
	return popup;
    }
    return rfm_get_widget(MAIN_POPUP_MENU_ID);
 }


