/*
 * Copyright 2005-2012 Edscott Wilson Garcia 
 * license: GPL v.3
 */
#define NON_RELEASE

#define __RFM_MAIN_C__
#include  "config.h"



#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <glob.h>
#include <limits.h>
#include <memory.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>

#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <gmodule.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include "rodent_libs.h"

#ifdef ICONVIEW_BINARY
# include "gridview_lib.h"
#endif
#ifdef DESKVIEW_BINARY
# include "deskview_lib.h"
#endif

#define FORCE_CORE
#ifdef FORCE_CORE
# include <sys/time.h>
# include <sys/resource.h>
#endif

extern gchar *rfm_iconpath;

static void
finishit (int sig) {
    TRACE ("got signal %d (SIGUSR1=%d,SIGUSR2=%d)", sig, SIGUSR1, SIGUSR2);
    fflush (NULL);
    if(sig == SIGUSR1) {
        return;
    } else if(sig == SIGUSR2) {
        return;
    } else {
        g_warning ("Rodent: signal %d received. \n", sig);
        //cleanup_tmpfiles();
        rfm_void (MODULE_DIR, "settings", "mcs_shm_stop");
        exit (1);
    }
}

static void
signal_connections (void
    ) {
#ifdef HAVE_SIGACTION
    struct sigaction act;
    act.sa_handler = finishit;
    sigemptyset (&act.sa_mask);
# ifdef SA_RESTART
    act.sa_flags = SA_RESTART;
# else
    act.sa_flags = 0;
# endif

//      sigaction(SIGHUP,&act,NULL);
//      sigaction(SIGTERM,&act,NULL);
# ifndef DEBUG
//      sigaction(SIGINT,&act,NULL);
//      sigaction(SIGQUIT,&act,NULL);
//      sigaction(SIGABRT,&act,NULL);
//      sigaction(SIGBUS,&act,NULL);
#  ifndef FORCE_CORE
    sigaction (SIGSEGV, &act, NULL);
#  endif
//      sigaction(SIGFPE,&act,NULL);

    sigaction (SIGUSR1, &act, NULL);
    sigaction (SIGUSR2, &act, NULL);
# endif
#else
//      signal(SIGHUP, finishit);
//      signal(SIGTERM, finishit);
# ifndef DEBUG
//      signal(SIGINT, finishit);
//      signal(SIGQUIT, finishit);
//      signal(SIGABRT, finishit);
//      signal(SIGBUS, finishit);
#  ifndef FORCE_CORE
    signal (SIGSEGV, finishit);
#  endif
//      signal(SIGFPE, finishit);
    signal (SIGUSR1, finishit);
    signal (SIGUSR2, finishit);
# endif
#endif
}

static void
rfm_global_init (int argc, char *argv[]) {
    TRACE("rfm_global_init...\n");
    if(rfm_global_p) return;
    rfm_global_p = (rfm_global_t *) malloc (sizeof (rfm_global_t));
    memset (rfm_global_p, 0, sizeof (rfm_global_t));
    rfm_global_p->argv = (char **)malloc ((argc + 3) * sizeof (char *));
    memset (rfm_global_p->argv, 0, (argc + 3) * sizeof (char *));
    rfm_global_p->argv[0] = g_path_get_basename (argv[0]);
    int i;
    for(i = 1; i < argc; i++) {
        rfm_global_p->argv[i] = g_strdup (argv[i]);
    }
    rfm_global_p->argc = argc;
    // main thread...
    rfm_global_p->self = g_thread_self();
    rfm_global_p->pid = getpid();
    rfm_rw_lock_init(&(rfm_global_p->setup_lock));
    rfm_rw_lock_init(&(rfm_global_p->icon_theme_lock));
    rfm_rw_lock_init(&(rfm_global_p->view_list_lock));
    rfm_rw_lock_init(&(rfm_global_p->window_lock));
    rfm_cond_init(rfm_global_p->janitor_signal);
    
    // Popup control
    rfm_mutex_init(rfm_global_p->status_mutex);
    rfm_cond_init(rfm_global_p->status_signal);
    rfm_mutex_init(rfm_global_p->popup_mutex);
    rfm_cond_init(rfm_global_p->popup_ready);

    rfm_rec_mutex_init(&(rfm_global_p->popup_create_mutex));
    

    TRACE("rfm_global_init(): self=0x%x\n",
	    GPOINTER_TO_INT(rfm_global_p->self));    
}

static gboolean
version_query (int argc, char *argv[]
    ) {
    int i;
    for(i = 0; i < argc; i++) {
        if(strcmp (argv[i], "--version") == 0) {
            fprintf (stdout, "%c[31m\tRodent uses the following programs:\n", 0x1B);
            //fprintf (stdout, "%c[0m",0x1B);
            //fprintf(stderr, "\tDBH version-%s, ", DBH_VERSION);
            gchar **pp;
#if THIS_IS_LINUX
            gchar *shell = g_strdup_printf ("%s --version", getenv ("SHELL"));
            FILE *pipef;
            gchar *commands[] = { "cp --version", "mv --version", "ln --version",
                "rm --version", "mount --version", "grep --version",
                "fgr --V",
                "sudo -V", shell, "", NULL
            };
            for(pp = commands; pp && *pp; pp++) {
                pipef = popen (*pp, "r");
                fprintf (stdout, "%c[35m-------------------------------------------------------------------\n", 0x1B);
                if(pipef) {
                    char line[2048];
                    memset (line, 0, 2048);
                    while(fgets (line, 2047, pipef) && !feof (pipef)) {
                        fprintf (stdout, "%c[32m%s", 0x1B, line);
                    }
                    pclose (pipef);
                }
            }
	    g_free(shell);
#else
            gchar *commands[] = { "cp", "mv", "ln", "rm", "mount", "grep","fgr", "sudo", getenv ("SHELL"),  NULL};
            for(pp = commands; pp && *pp; pp++) {
		gchar *path=g_find_program_in_path(*pp);
                fprintf (stdout, "%c[32m%s", 0x1B, path);
		g_free(path);
	    }

#endif
            fprintf (stdout, "%c[31m\tThis is %s %s (%s build %s)\n", 0x1B, PACKAGE_NAME, TAG, VERSION, BUILD);
            fprintf (stdout, "%c[31m\t%s\n", 0x1B, COPYRIGHT);
            fprintf (stdout, "%c[31m\tBuilt with GTK+-%d.%d.%d,linked with GTK+-%d.%d.%d.\n",
                     0x1B, GTK_MAJOR_VERSION, GTK_MINOR_VERSION, GTK_MICRO_VERSION,
                     gtk_major_version, gtk_minor_version, gtk_micro_version);
            fprintf (stdout, "%c[31m\tSend bug reports to %s\n", 0x1B, PACKAGE_BUGREPORT);
            fprintf (stdout, "%c[0m\n", 0x1B);
            exit (0);
        }

    }
    return FALSE;
}

static gboolean enable_desktop = FALSE;
static gboolean synchronize_X = FALSE;

static void
process_arg (int *argc, char ***argv) {
    int i;
    int out_argc = 0;
    static char **out_argv;

    out_argv = (char **)malloc ((*argc + 2) * sizeof (char *));
    out_argv[*argc - 1] = out_argv[*argc] = NULL;
    for(i = 0; i < *argc; i++) {
	if (strcmp(*((*argv) + i), "--sync")==0) {
	    synchronize_X = TRUE;
	    continue;
	}
        out_argv[out_argc] = *((*argv) + i);
        TRACE ("out_argcv[%d]=%s\n", out_argc, out_argv[out_argc]);
        out_argc++;
    }
    /* do variable substitution for $HOME or ~ */
    if(strcmp (out_argv[out_argc - 1], "$HOME") == 0)
        out_argv[out_argc - 1] = g_strdup(g_get_home_dir ());
    if(strcmp (out_argv[out_argc - 1], "~") == 0)
        out_argv[out_argc - 1] = g_strdup(g_get_home_dir ());
    if (rfm_g_file_test(out_argv[out_argc - 1], G_FILE_TEST_IS_DIR)){
	// change to absolute path
	if (!g_path_is_absolute (out_argv[out_argc - 1])){
	    gchar *original= g_get_current_dir();
	    if (chdir(out_argv[out_argc - 1]) < 0){
		g_warning("chdir(%s): %s", out_argv[out_argc - 1], strerror(errno));
	    }
	    out_argv[out_argc - 1] = g_get_current_dir();
	    if (chdir(original) < 0){
		g_warning("chdir(%s): %s", original, strerror(errno));
	    }
	    g_free(original);
	}
    }
    /* if no directory specified, use current WD */
 /*   if (out_argc < *argc + 1 && strstr(out_argv[0],"rodent-forked")==NULL)  {
        // USE EXTRA POINTER INITIALLY ALLOCATED TO NULL 
        out_argv[out_argc] = g_get_current_dir ();
        out_argc++;
    }*/
  /*  if(!rfm_g_file_test (out_argv[out_argc - 1], G_FILE_TEST_IS_DIR)
	    && strstr(out_argv[0],"rodent-forked")==NULL) {
        // USE EXTRA POINTER INITIALLY ALLOCATED TO NULL 
        out_argv[out_argc] = g_get_current_dir ();
        out_argc++;
    }*/
    *argv = out_argv;
    *argc = out_argc;
}

int
main (int argc, char *argv[]) {
    NOOP ("Rodent: program name is = %s\n", argv[0]);

    // In order for SSH_ASKPASS to work (for sshfs or for executing
    // ssh commands at the lpterminal) we must detach the tty. 
    // This interferes with stepwise debugging with gdb.
    // The other caveat with gdb is that rfm_g_file_test_with_wait()
    // will detect the timeout under stepwise gdb.
    // was: if(strstr (argv[0], "rodent-forked") != NULL)
    // is now:
    /* do the following before allowing any dynamic library
     * to be loaded to memory (except glib)... */
    rfm_sanity_check (argc, argv, LIBRFM_SERIAL);
    if(version_query (argc, argv)) exit (1);
#if defined DEBUG_TRACE
    g_warning("DEBUG_TRACE enabled: sudo/ssh password dialog will not operate!\n");
#else
    if(strstr (argv[0], "rodent-getpass") == NULL)
    {

        if(fork ()){
	    // This is kinky...
	    // A shared semaphore would be more elegant...
	    gboolean initial = FALSE;
	    gchar *setup_file = g_build_filename(SETUP_FILE, NULL);
	    if (!g_file_test(setup_file, G_FILE_TEST_EXISTS)) {
		initial = TRUE;
		fprintf(stdout, "Initial set up of rodent disk based hashtables. Please wait."); fflush(stdout);
	    } else {
		if (unlink(setup_file));
	    }
	    do {
		if (initial) fprintf(stdout, "."); fflush(stdout);
		sleep(1);
	    } while(!g_file_test(setup_file, G_FILE_TEST_EXISTS));
	    g_free(setup_file);
	    _exit (123);
	}
	setsid();
    } 
#endif

   /* g_thread_supported is geeky. returns true if g_threads are already 
     * initialized, not whether they are supported on the current platform! */
    XInitThreads ();
    rfm_thread_init (NULL);


    // this is a must. Nothing interesting in stdin anyways.
    int fd = fileno (stdin);
    if (fd >= 0){
	FILE *nullfd = fopen ("/dev/null", "rb");
	dup2 (fileno (nullfd), fd);
    }

    /* ignore hangups? */
    (void)signal (SIGHUP, SIG_IGN);

    gchar *g = g_strdup_printf ("%s/%s", g_get_user_cache_dir (), RC_DIR);
    g_mkdir_with_parents (g, 0700);
    g_free (g);
    g = g_strdup_printf ("%s/%s", g_get_user_data_dir (), RC_DIR);
    g_mkdir_with_parents (g, 0700);
    g_free (g);
    g = g_strdup_printf ("%s/%s", g_get_user_config_dir (), RC_DIR);
    g_mkdir_with_parents (g, 0700);
    g_free (g);

#ifdef CORE
    struct rlimit rlim;
    if (!strstr(argv[0], "getpass")) {
	g_warning("Enabling core dumps...");
    }
    rlim.rlim_cur = RLIM_INFINITY;
    rlim.rlim_max = RLIM_INFINITY;
    setrlimit (RLIMIT_CORE, &rlim);
#endif

    process_arg (&argc, &argv);
    if(!g_module_supported ()) {
        g_error ("g_module_supported() != TRUE\n");
        exit (1);
    }
    signal_connections ();

    rfm_global_init (argc, argv);
    rfm_rw_lock_writer_lock (&(rfm_global_p->setup_lock));

    /***************************************************/
    /* start loading required dynamic libraries here... */
#ifdef ENABLE_NLS
    /* this binds rfm domain: */
    bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);

    bindtextdomain ("librfm", PACKAGE_LOCALE_DIR);
    bindtextdomain ("rfm-gui", PACKAGE_LOCALE_DIR);
# ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset ("librfm", "UTF-8");
    bind_textdomain_codeset ("rfm-gui", "UTF-8");
# endif
# ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    TRACE ("Rodent: binding %s, at %s\n", GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
# endif
#endif
 
    setlocale (LC_ALL, "");

    gtk_init (&argc, &argv);

    if (synchronize_X) {
	    TRACE("Synchronizing X events!!!!\n");
	    Display *display=gdk_x11_display_get_xdisplay(gdk_display_get_default());
	    XSynchronize(display, True);
    }

    // set gtk properties
    GtkSettings *settings = gtk_settings_get_default();
    if (settings){
	/* make sure the type is realized */
	g_type_class_unref (g_type_class_ref (GTK_TYPE_IMAGE_MENU_ITEM));
	g_type_class_unref (g_type_class_ref (GTK_TYPE_BUTTON));
	g_object_set(G_OBJECT(settings), 
	    "gtk-button-images", TRUE,
	    "gtk-menu-images", TRUE,
	    NULL);
    }

    /* read in shm settings: this will also pull in icon module... */
    NOOP("Rodent: Loading settings dialog module\n");
    if (!rfm_void (MODULE_DIR, "settings", "mcs_shm_start")){
	NOOP(stderr, "Setting default or preset envvars\n");
	gint i;
	for(i = 0; i < RFM_OPTIONS; i++) {
	    gboolean default_value =
		environ_v[i].env_string && 
		strlen(environ_v[i].env_string);
	    gboolean preset_value =
		getenv(environ_v[i].env_var) &&
		strlen(getenv(environ_v[i].env_var));
	    if (preset_value) {
		// Do nothing: use preset value over default value
	    } else if (default_value){
		// use default value
		rfm_setenv(g_strdup(environ_v[i].env_var),
			g_strdup(environ_v[i].env_string), FALSE);
		NOOP(stderr, "setting %s -> %s (%d)\n",
			environ_v[i].env_var, environ_v[i].env_string, FALSE);
	    } else {
		// unset envvar
		rfm_setenv(g_strdup(environ_v[i].env_var), NULL, FALSE);
	    }
	}
	if(getenv ("EDITOR") && strlen (getenv ("EDITOR"))) {
	    gchar *editor=g_path_get_basename(getenv ("EDITOR"));
	    if (strchr(editor, ' ')) *strchr(editor, ' ') = 0;
	    gchar *fullpath=g_find_program_in_path(editor);
	    // Don't show the EDITOR_CMD if it does not exist
	    if (fullpath) {
		// if nano or vi, then use terminal emulator
		if (strcmp(editor, "vi")==0 || strcmp(editor, "nano")==0){
		    const gchar *t=getenv("TERMINAL_CMD");
		    gchar *term = g_find_program_in_path(t);
		    if (term) g_free(term);
		    else {
			t=NULL;
			for (i=0; terminals_v[i]; i++){
			    // sakura is broken... 
			    if (strstr(terminals_v[i], "sakura")) continue;
			    term = g_find_program_in_path(terminals_v[i]);
			    if (term){
				t=terminals_v[i];
				g_free(term);
				break;
			    }
			}
		    }
		    if (!t || !strlen(t)) t="xterm";
		    gchar *b=g_strdup_printf("%s %s %s",
			    t, rfm_term_exec_option(t), editor);
		    g_free(editor);
		    editor = b;
		}
		g_free(fullpath);
		rfm_setenv("EDITOR", editor, FALSE);
	    } else {
		g_free(editor);
	    }
	}	
    }
    // Check if icon module was available to load
    if(!rfm_void(MODULE_DIR, "icons", "module_active")){
        TRACE ("Rodent: Cannot load icon theme module: proceeding with Rodent basic and gtk built-in icons.\n");
    }
    if(strstr (argv[0], "-getpass")) {

        gchar *p;
	const gchar *sudo_command=getenv("RFM_ASKPASS_COMMAND");

	gchar *string=g_strdup_printf("%s", _("Enter password"));
	if (sudo_command && strlen(sudo_command)) {
	    gchar *p = (gchar *)sudo_command;
	    gchar *markup=g_strdup("");
	    for (;p && *p; p++){
		gchar *g;
		if (*p == '&') g = g_strdup_printf("%s&amp;", markup);
		else g = g_strdup_printf("%s%c", markup, *p );
		g_free(markup);
		markup = g;
	    }
	    g_free(string);
	    string=g_strdup_printf("%s\n%s:", markup, _("Enter password"));
	    g_free(markup);
	}
	
        p = rfm_get_password (NULL, string);

	g_free(string);
	if (p && strlen(p)) {
	    fprintf (stdout, "%s\n", p);
	    memset(p, 0, strlen(p));
	} else {
	    // No password, either cancel or close, then
	    // send interrupt signal to parent
	    pid_t parent = getppid();
	    kill(parent, SIGINT);
	}
	g_free(p);

	exit(0);
    }


    rfm_init_env ();            /* allocate memory for dynamic environment variables */
    rfm_setenv ("PWD", (gchar *) GETWD, FALSE);


    if(strstr (argv[0], "rodent-desk") != NULL){
        enable_desktop = FALSE;
    } else if(getenv ("RFM_ENABLE_DESKTOP") && strlen (getenv ("RFM_ENABLE_DESKTOP"))){
        enable_desktop = GPOINTER_TO_INT(rfm_void(MODULE_DIR, "settings", "localhost_check"));
    }

    NOOP ("enable_desktop=%d\n", enable_desktop);
    if(enable_desktop) {
	Display *display=gdk_x11_display_get_xdisplay(
		gdk_display_get_default());
        GError *error = NULL;
        Atom selection_atom = XInternAtom (display, "RODENT_DESK_ATOM", True);
        Window xid = (selection_atom != None)?
	    XGetSelectionOwner (display, selection_atom):
	    None;
        NOOP ("enabling desktop=%d atom=%d xid=%d\n", 
		enable_desktop, (gint)selection_atom, (gint)xid);
        if(selection_atom != None && xid != None) 
	{
            TRACE ("rodent-desk already running\n");
        } else if(!g_spawn_command_line_async ("rodent-desk", &error)) {
            g_warning ("%s", error->message);
        }
    } else {
        NOOP ("Auto deskview disabled\n");
    }
#if 0
    if(strstr (argv[0], "rfm-run") || strstr (argv[0], "rfm-find")) {
        smc_connect (argc, argv, SmRestartNever);
    } else {
# ifdef ICONVIEW_BINARY
        smc_connect (argc, argv, SmRestartIfRunning);
# endif
# ifdef DESKVIEW_BINARY
        smc_connect (argc, argv, SmRestartImmediately);
# endif

    }
#endif


#ifndef PACKAGE_URL
# define PACKAGE_URL "http://xffm.org"
#endif

#ifdef ICONVIEW_BINARY
    TRACE("Rodent: create_gridview ()\n");
    rfm_global_p->window = create_gridview ();
#endif

#ifdef DESKVIEW_BINARY
    TRACE("Rodent: create_desktop ()\n");
    rfm_global_p->window = create_desktop ();
#endif

    SETWD();
    {
	gchar *setup_file = g_build_filename(SETUP_FILE, NULL);
	FILE *f=fopen(setup_file, "w");
	if (f) fclose(f);
	g_free(setup_file);
    }
    // flush all pending gtk stuff from the main event loop
    // (which is not yet up and running)
    while (gtk_events_pending()) gtk_main_iteration();
    gdk_flush();
    // Now we can release threads which want to do stuff with the gtk widgets.
    rfm_rw_lock_writer_unlock (&(rfm_global_p->setup_lock));
    TRACE("Rodent: wait on setup semaphore\n");
    NOOP("Starting janitor thread...\n");
    // This thread will exit only after window is destroyed in this
    // thread. Therefore, window nor view destruction should wait for 
    // janitor_thread termination.
    rfm_thread_create ("janitor", janitor, rfm_global_p->window, FALSE);
 
    TRACE("Rodent: entering event loop\n");
    rfm_global_p->main_loop = g_main_loop_new (NULL, FALSE);
    rfm_global_p->self_context = g_main_context_default();
    g_main_loop_run(rfm_global_p->main_loop);
    //gtk_main ();
    TRACE("Rodent: exited event loop\n");

    // Free shared memory reference to shm settings 
    rfm_void (MODULE_DIR, "settings", "mcs_shm_stop");
    

    return 0;
}
